/* --------------------------------------------------------------------------
 *	Graphic framebuffer text management
 * --------------------------------------------------------------------------
 *
 *	Copyright (C) 1997		Emmanuel Marty
 *	Copyright (C) 1997		The GGI project
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2, or (at your option)
 *	any later version.
 * 
 *	You should have received a copy of the GNU General Public License
 *	along with this program; see the file COPYING.  If not, write to
 *	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * --------------------------------------------------------------------------
 *
 *	$Log: gx_text16.inc,v $
 *	Revision 1.2  1997/12/24 18:20:50  emarty
 *	textmode emulation now works
 *	
 *	Revision 1.1  1997/12/20 13:03:56  emarty
 *	major: initial checkin - does nothing for now.
 *	
 */

/*
** Global variables - incompatible with multihead, but the only
** solution for current KGI kernel. This file will go away when
** the EvStack kernel is here and implements real graphic framebuffer
** consoles, so this is not an issue.
*/

static struct kgi_font fbtext_font = {
			0,0,			/* height, width	*/
			0,0, 			/* dx, dy		*/
			0,			/* base line		*/
			NULL,			/* info			*/
			NULL			/* data			*/
};

static mem_vaddr	fbtext_graphmem;	/* pointer to vidmem	*/
static uint		fbtext_linesize;	/* size of line (bytes) */
static uint		fbtext_charsize;	/* size of char (bytes)	*/
static uint		fbtext_hdelta;		/* hori char delta	*/
static uint		fbtext_ldelta;		/* pixel line delta	*/
static uint		fbtext_vdelta;		/* vert char delta	*/

/*
** text16 palette
*/

static ggi_color fbtext_palette[18] = {
						/* Normal		*/

		{ 0x0000, 0x0000, 0x0000 },	/* COLOR_BLACK		*/
		{ 0x7F00, 0x0000, 0x0000 },	/* COLOR_RED		*/
		{ 0x0000, 0x7F00, 0x0000 },	/* COLOR_GREEN		*/
		{ 0x7F00, 0x7F00, 0x0000 },	/* COLOR_YELLOW		*/
		{ 0x0000, 0x0000, 0x7F00 },	/* COLOR_BLUE		*/
		{ 0x7F00, 0x0000, 0x7F00 },	/* COLOR_MAGENTA	*/
		{ 0x0000, 0x7F00, 0x007F },	/* COLOR_CYAN		*/
		{ 0x7F00, 0x7F00, 0x7F00 },	/* COLOR_WHITE		*/

						/* Highlight		*/

		{ 0x0000, 0x0000, 0x0000 },	/* COLOR_BLACK		*/
		{ 0xFF00, 0x0000, 0x0000 },	/* COLOR_RED		*/
		{ 0x0000, 0xFF00, 0x0000 },	/* COLOR_GREEN		*/
		{ 0xFF00, 0xFF00, 0x0000 },	/* COLOR_YELLOW		*/
		{ 0x0000, 0x0000, 0xFF00 },	/* COLOR_BLUE		*/
		{ 0xFF00, 0x0000, 0xFF00 },	/* COLOR_MAGENTA	*/
		{ 0x0000, 0xFF00, 0x00FF },	/* COLOR_CYAN		*/
		{ 0xFF00, 0xFF00, 0xFF00 },	/* COLOR_WHITE		*/

						/* Cursor colors	*/

		{ 0x7F00, 0x7F00, 0x7F00 },	/* color 0		*/
		{ 0x7F00, 0x7F00, 0x7F00 },	/* color 1		*/
};

static ggi_clut fbtext_clut = {
		18,
		fbtext_palette
};

/*
** graphic framebuffer text operations emulation for current kernel
*/

static void fbtext_move_mem(kgi_pixmap *fb, uint dst, uint src, uint cnt) {
	register ushort *pbuf = (void *) fb->pbuf;

	memmove(pbuf + dst, pbuf + src, cnt + cnt);
}

static void fbtext_set_mem(kgi_pixmap *fb, uint src, ggi_pixel val, uint cnt) {
	register ushort *pbuf = ((ushort *) fb->pbuf) + src;

	while (cnt--) {

		*(pbuf++) = val;
	}
}

static void fbtext_invert_mem(kgi_pixmap *fb, uint start, uint end) {
	register ushort *pix = ((ushort *) fb->pbuf) + start;

	while (start++ <= end) {

		*pix =     ((*pix) & 0x88FF) 
			| (((*pix)<<4) & 0x7000) | (((*pix>>4)) & 0x0700);
		pix++;
	}
}

static void fbtext_update_mem(kgi_pixmap *fb, uint pos, ggi_pixel val) {
	*(((ushort*) fb->pbuf) + pos) = val;
}

static void fbtext_flush_mem(kgi_pixmap *fb) {
}

/*
**	character line rendering function
**	needs speed, implemented in assembler
**
**	only x86 assembler here, but since the MediaGX is an x86
**	processor, it doesn't matter.
*/

#ifdef __i386__

static void fbtext_renderchars (ushort *buf, ushort *pbuf, char *gpos,
				uint32 h, uint32 xpos, uint32 ldelta,
				uint32 xmax, uint32 cnt, char *fnt) {

__asm__ __volatile__ (

"	movl	%7, %%eax\n"
"	orl	%%eax,%%eax\n"
"	je	9f\n"

"11:\n"
"	movl	%2,%%edi\n"
"	movl	%%edi,%%esi\n"
"	addl	fbtext_hdelta,%%esi\n"
"	movl	%%esi,%2\n"

"	movl	%0,%%esi\n"
"	movw	(%%esi),%%ax\n"
"	addl	$2,%0\n"
"	movl	%1,%%esi\n"
"	movw	%%ax,(%%esi)\n"
"	addl	$2,%1\n"

"	movzbl	%%al,%%esi\n"
"	imull	fbtext_charsize,%%esi\n"
"	addl	%8,%%esi\n"

"	mov	%%ah,%%al\n"
"	shrb	$4,%%al\n"
"	andb	$0x07,%%al\n"
"	mov	%%ah,%%cl\n"
"	andb	$0x80,%%cl\n"
"	orb	%%cl,%%al\n"
"	andb	$0x0F,%%ah\n"

"	movl	%3,%%ecx\n"
"	movl	fbtext_ldelta,%%ebx\n"

"0:\n"
"	movb	(%%esi), %%dl\n"
"	incl	%%esi\n"

"	movb	%%al, %%dh\n"
"	addb	%%dl, %%dl\n"
"	jnc	1f\n"
"	movb	%%ah, %%dh\n"
"1:\n"
"	movb	%%dh, (%%edi)\n"
"	incl	%%edi\n"

"	movb	%%al, %%dh\n"
"	addb	%%dl, %%dl\n"
"	jnc	2f\n"
"	movb	%%ah, %%dh\n"
"2:\n"
"	movb	%%dh, (%%edi)\n"
"	incl	%%edi\n"

"	movb	%%al, %%dh\n"
"	addb	%%dl, %%dl\n"
"	jnc	3f\n"
"	movb	%%ah, %%dh\n"
"3:\n"
"	movb	%%dh, (%%edi)\n"
"	incl	%%edi\n"

"	movb	%%al, %%dh\n"
"	addb	%%dl, %%dl\n"
"	jnc	4f\n"
"	movb	%%ah, %%dh\n"
"4:\n"
"	movb	%%dh, (%%edi)\n"
"	incl	%%edi\n"

"	movb	%%al, %%dh\n"
"	addb	%%dl, %%dl\n"
"	jnc	5f\n"
"	movb	%%ah, %%dh\n"
"5:\n"
"	movb	%%dh, (%%edi)\n"
"	incl	%%edi\n"

"	movb	%%al, %%dh\n"
"	addb	%%dl, %%dl\n"
"	jnc	6f\n"
"	movb	%%ah, %%dh\n"
"6:\n"
"	movb	%%dh, (%%edi)\n"
"	incl	%%edi\n"

"	movb	%%al, %%dh\n"
"	addb	%%dl, %%dl\n"
"	jnc	7f\n"
"	movb	%%ah, %%dh\n"
"7:\n"
"	movb	%%dh, (%%edi)\n"
"	incl	%%edi\n"

"	movb	%%al, %%dh\n"
"	addb	%%dl, %%dl\n"
"	jnc	8f\n"
"	movb	%%ah, %%dh\n"
"8:\n"
"	movb	%%dh, (%%edi)\n"
"	incl	%%edi\n"

"	addl	%%ebx, %%edi\n"
"	decl	%%ecx\n"
"	jnz	0b\n"

"	incl	%4\n"
"	movl	%6,%%eax\n" 
"	cmpl	%%eax,%4\n"
"	jb	10f\n"
"	movl	$0,%4\n"
"	movl	%5,%%ebx\n"
"	addl	%%ebx,%2\n"

"10:\n"
"	decl	%7\n"
"	jne	11b\n"

"9:\n"

 : /* output */
 : /* input */ "g"(buf), "g"(pbuf), "g"(gpos), "g"(h), "g"(xpos), "g"(ldelta),
		"g"(xmax), "g"(cnt), "g"(fnt)
 : /* clobbered */ "cc","%eax","%ebx","%ecx","%edx","%esi","%edi");
}

#else
	#error Need to implement non-i386 fbtext char rendering
#endif

static void fbtext_copyto_mem(kgi_pixmap *fb, void *buf, uint start, uint cnt) {
	uint32	h;
	ushort	*pbuf = ((ushort *) fb->pbuf) + start;

	if ( (h = fbtext_font.height) ) {
		uint32	xpos = start % fb->size.x;

		fbtext_renderchars (
			buf,
			pbuf,
			((char *) (fbtext_graphmem)) +
				(start / fb->size.x) * fbtext_vdelta +
				xpos * fbtext_hdelta,
			h,
			xpos,
			fbtext_vdelta-(fb->size.x*fbtext_hdelta),
			fb->size.x,
			cnt,
			fbtext_font.data
		);
	}
	else {
		memcpy(pbuf, buf, cnt + cnt);
	}

}

static void fbtext_copyfrom_mem(kgi_pixmap *fb, void *buf,
				uint start, uint cnt) {

	memcpy(buf, ((ushort *) fb->pbuf) + start, cnt + cnt);
}

#ifdef __i386__
#define	fbtext_move_io		fbtext_move_mem
#define	fbtext_set_io		fbtext_set_mem
#define	fbtext_invert_io	fbtext_invert_mem
#define	fbtext_update_io	fbtext_update_mem
#define	fbtext_flush_io		fbtext_flush_mem
#define	fbtext_copyfrom_io	fbtext_copyfrom_mem
#define	fbtext_copyto_io	fbtext_copyto_mem
#else
	#error Need to implement non-i386 fbtext_..._io framebuffer operations
#endif

static ggi_pixel fbtext_atop_color(kgi_pixmap *fb, ggi_attr attr) {
	register ggi_pixel
	result = (attr & ATTR_REVERSE)
			? (((attr & 0x0007) << 8) | ((attr & 0x0700) << 4))
			: (((attr & 0x0007) << 12) | (attr & 0x0700));

	if (attr & ATTR_BLINK) {

		result |= 0x8000;
	}

	if (attr & ATTR_BOLD) {

		result |= 0x0800;
	}

	return result;
}

static ggi_attr fbtext_ptoa_color(kgi_pixmap *fb, ggi_pixel val) {
	return	  ((val & 0x8000) ? ATTR_BLINK : 0)
		| ((val & 0x0800) ? ATTR_BOLD  : 0)
		| ((val & 0x7000) >> 12) | (val & 0x0F00);
}

static ggi_pixel fbtext_atop_mono(kgi_pixmap *fb, ggi_attr attr) {
	register ggi_pixel result;

	if (attr & ATTR_UNDERLINE) {

		result = 0x0100;	/* MDA can't do reverse underline :-( */

	} else {

		result = (attr & ATTR_REVERSE) ? 0x7000 : 0x0700;
	}

	if (attr & ATTR_BLINK) {

		result |= 0x8000;
	}

	if (attr & ATTR_BOLD) {

		result |= 0x0800;
	}

	return result;
}

static ggi_attr fbtext_ptoa_mono(struct kgi_pixmap *fb, ggi_pixel val) {
	register ggi_attr result;

	if ((val & 0x7700) == 0x0100) {

		result = ATTR_UNDERLINE;

	} else {

		result = (val & 0x7000) ? ATTR_REVERSE : ATTR_NORMAL;
	}

	if (val & 0x8000) {

		result |= ATTR_BLINK;
	}

	if (val & 0x0800) {

		result |= ATTR_BOLD;
	}

	return result;
}

/*
**	font postion to pixel mapping and vice versa. This cares about 256 
**	character fonts only (at least for now).
*/

static ggi_pixel fbtext_ftop(kgi_pixmap *fb, uint pos) {
	return pos & 0x00FF;
}

static uint fbtext_ptof(kgi_pixmap *fb, ggi_pixel val) {
	return val & 0x00FF;
}

/*	We have color and mono operations, both for I/O and normal memory; so
**	there are four framebuffer operations to provide.
*/

#define	FBTEXT_FB_OPS(mode, mem)\
	fbtext_##mode##_##mem = {\
		fbtext_move_##mem,\
		fbtext_set_##mem,\
		fbtext_invert_##mem,\
		fbtext_update_##mem,\
		fbtext_flush_##mem,\
		fbtext_copyfrom_##mem,\
		fbtext_copyto_##mem,\
		\
		fbtext_atop_##mode,\
		fbtext_ptoa_##mode,\
		\
		fbtext_ftop,\
		fbtext_ptof\
	}

kgi_fb_ops FBTEXT_FB_OPS(mono,  io);
kgi_fb_ops FBTEXT_FB_OPS(mono,  mem);
kgi_fb_ops FBTEXT_FB_OPS(color, io);
kgi_fb_ops FBTEXT_FB_OPS(color, mem);

/*
** text emulation for chipset driver
*/

static void fbtext_set_font(struct kgi_display *dpy, struct kgi_font *font,
				uint from, uint to) {
	struct	chipset_private *private = KGIM_PRIVATE (dpy, chipset);
	uint32	*sprite, j;
	uint	tmp;

/*
**	store font information
*/

	memcpy (&fbtext_font, font, sizeof (struct kgi_font));

/*
**	!! work around a bug in KGI - width and height are reversed
*/

/*	tmp = fbtext_font.height;		*/
	tmp = 8;				/* only 8 pixels supported */
	fbtext_font.height = fbtext_font.width;
	fbtext_font.width = tmp;

/*
**	compute necessary values for font emulation routines
*/

	fbtext_hdelta	= fbtext_font.width;
	fbtext_ldelta	= fbtext_linesize - 8;
	fbtext_vdelta	= fbtext_font.height * fbtext_linesize;
	fbtext_charsize = (fbtext_font.dx >> 3) * fbtext_font.dy;

/*
**	init cursor data
*/

	sprite = (uint32 *) private->spritemem;

	for (j=0; j < (fbtext_font.height - 2); j++) {
		*sprite ++ = 0xFFFF0000;
		*sprite ++ = 0xFFFF0000;
	}

	*sprite ++ = 0x00FFFF00;
	*sprite ++ = 0xFFFF0000;
	*sprite ++ = 0x00FFFF00;
	*sprite ++ = 0xFFFF0000;

	for (j=fbtext_font.height; j<32; j++) {
		*sprite ++ = 0xFFFF0000;
		*sprite ++ = 0xFFFF0000;
	}
}

/*
**	set text palette
*/

static void fbtext_set_textclut (struct kgi_display *dpy) {
	if (dpy->set_clut)
		(dpy->set_clut) (dpy, &fbtext_clut, 0, 18);
}
