/*
** --------------------------------------------------------------------------
**	Hercules I/O
** --------------------------------------------------------------------------
**
**	Copyright (C) 1995 by Jochen Hoenicke,
**			      [jochen.hoenicke@informatik.uni-oldenburg.de]
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2, or (at your option)
**	any later version.
**
**	You should have received a copy of the GNU General Public License
**	along with this program; see the file COPYING.  If not, write to
**	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
**
** ---------------------------------------------------------------------------
*/

#undef DONTSET

/*
**	Kernel includes
*/

#include <asm/io.h>

/*
**     GGI includes
*/

#include <ggi/kgi.h>

static inline unsigned char inbCRT(unsigned char reg)
	{ outb_p(reg, CRTIm); return inb_p(CRTDm); }

#ifndef DONTSET 
static inline void outbCTRL(unsigned char value)
	{ outb_p(value, CTRLm); }

static inline void outbCONF(unsigned char value)
	{ outb_p(value, CONFm); }

static inline void outbCRT(unsigned char value, unsigned char reg) 
	{ outb_p(reg, CRTIm); outb_p(value, CRTDm); }
#else
#warning DONTSET: Resister setting disabled
static inline void outbCTRL(unsigned char value) {}
static inline void outbCONF(unsigned char value) {}
static inline void outbCRT(unsigned char value, unsigned char reg) {}
#endif /* DONTSET */

/*
**	clock I/O functions
*/

void CLK_out8(struct kgi_display *dpy, uint8 value, uint8 reg)
{
	/* The HGC has only one clock */
	ASSERT(value == 0);
}

uint8 CLK_in8(struct kgi_display *dpy, uint8 reg)
{
	return 0;
}

#define	NoCRTRegs	0x10

struct statetable
{
   unsigned char CTRL, CONF;
   unsigned char CRT[NoCRTRegs];
};

static void restore_CRT(unsigned char *buf)
{
	register unsigned char i;

        for(i = 0; i < NoCRTRegs; i++) {
                outbCRT(*(buf++), i);
        }
}

static void restore_state(struct statetable *state)
{
#ifndef DONTSET
        outbCTRL(state->CTRL & ~CTRL_ENB_CRT);
        outbCONF(state->CONF);
        restore_CRT(state->CRT);
        outbCTRL(state->CTRL);
#endif
}

#if defined(DEBUG_IO)||defined(DEBUG_STATE)
static void dump_state(struct statetable *state)
{ 
        int i;
        printk("CTRL: %.2x\n",state->CTRL);
        printk("CONF: %.2x",state->CONF);
        
        for(i = 0; i < NoCRTRegs; i++) {
                if(!(i & 0xf)) printk("\nCRT:%.2x: ",i); 
                printk("%.2x, ",state->CRT[i]); }
        printk("\n");
}
#endif

