/* --------------------------------------------------------------------------
** S3 86c765 (Trio64V+) chipset driver
** --------------------------------------------------------------------------
**
** Copyright (C)	1996-1997	Steffen Seeger
**			1997		Jon M. Taylor
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2, or (at your option)
** any later version.
**
** You should have received a copy of the GNU General Public License
** along with this program; see the file COPYING.  If not, write to
** the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
**
** ---------------------------------------------------------------------------
** 
** $Log: 86c765.c,v $
** Revision 1.8  1997/12/31 16:21:56  seeger
** Made the AEON_PCI #define an module-load option
**
** Revision 1.7  1997/12/14 21:55:45  becka
** uint/sint namespace fixed ... please test ...
**
** Revision 1.6  1997/11/08 15:12:29  becka
** Namespace cleanup ... probably broke all and everything ...
**
** Revision 1.5  1997/10/14 03:31:44  taylorj
** Minor: more formatting cleanups
**
** Revision 1.4  1997/10/10 02:32:40  taylorj
** Minor: Many small cleanups.
**
** Revision 1.3  1997/09/12 07:28:43  taylorj
** Minor: Reformatted source to be consistent with GGI coding style guidelines,
** modified all debugging output to use the DEBUGx macros, and added CVS
** revision logging to headers where needed.
**
** ---------------------------------------------------------------------------
*/

#include <ggi/maintainers.h>
#define	MAINTAINER	Jon_Taylor
#define	DRIVER_NAME	"S3 86c765 (Trio64) chipset driver"
#define	DRIVER_REV	"$Revision: 1.8 $"

#define DEBUG_LEVEL	0

#include <ggi/system.h>
#include <ggi/module.h>

#define	S3_86C765
#define	TRIO64PLUS
#define	HARDWARE_TEXT_CURSOR	0

#include "S3/Trio.h"
#define	IO_VGA_BASE	0x03B0
#define	IO_VGA_SIZE	0x30
#define	IO_ACCEL_BASE	0x02E8
#define	IO_ACCEL_SIZE	4

#define	TEXT_LEN	(64 KB)

struct chipset_private
{
	/* Internal stuff */
	mem_vaddr	mem_vga;
	mem_vaddr	mem_accel;

	ggi_uint		frame;
	ggi_uint		split;
	ggi_uint		origin, org_xdot, org_ydot, org_offs;
	ggi_uint		fb_size;

	uint32		ptr_old, cur_old;
	uint16		ptr_and, cur_and;
	uint16		ptr_xor, cur_xor;

	/* Exported stuff */
	struct kgi_mmio_region
	
	lram,		/* on-board RAM		little endian	*/
	lpdt,		/* pixel data transfer	little endian	*/
	lpfd,		/* pixel formatter data	little endian	*/
	lpfm,		/* pixel formatter data	little endian	*/
	
	bram,		/* on-board RAM		big endian	*/
	bpdt,		/* pixel data transfer	big endian	*/
	bpfd,		/* pixel formatter data	big endian	*/
	bpfm;		/* pixel formatter mask	big endian	*/

	/* Administrative stuff	*/
	struct io_region	io_vga_region;
	struct io_region	io_accel_region;
	struct mem_region	mem_accel_region;
	struct mem_region	mem_law_region;

	struct kgi_chipset	chipset_info;
};

#define GET_PRIVATE \
	struct chipset_private *private = KGIM_PRIVATE(dpy, chipset)

#define	IO_VGA			(private->io_vga_region.baseptr-IO_VGA_BASE)
#define	IO_ACCEL		(private->io_accel_region.baseptr-IO_ACCEL_BASE)
#define	MEM_LAW			(private->mem_law_region.baseptr)

#define	MEM_ACCEL		(private->mem_accel)
#define	MEM_VGA			(private->mem_vga)

#define	IO_VGA_REGION		(private->io_vga_region)
#define	IO_ACCEL_REGION		(private->io_accel_region)
#define	MEM_ACCEL_REGION	(private->mem_accel_region)
#define	MEM_LAW_REGION		(private->mem_law_region)

#define	CHIPSET_INFO		(private->chipset_info)

#define	TEXTMODE_MMIO_REGIONS	NULL
#define	GRAPHMODE_MMIO_REGIONS	NULL

#define	NoSEQRegs	0x05
#define	NoCRTRegs	0x80
#define	NoGRCRegs	0x09
#define	NoATCRegs	0x15


#define	BUGGY_MMIO
#include "s3_mmio.inc"
#include "s3_text16.inc"

/* Chipset info */
static struct kgi_chipset s3_765_info =
{
	"S3 Incorporated",		/* Manufacturer			*/
	"Trio64",			/* Model			*/
	GM_TEXT16 | /* GM_4BIT | */	/* Available display modes	*/
	GM_8BIT | GM_15BIT | GM_16BIT |
	GM_24BIT | GM_32BIT,
	0,				/* Flags			*/
	{ 2048, 2048 },			/* Maximum resolution		*/
	0,				/* Available RAM (autodetected)	*/
	{ 10000000, 95000000 }		/* DCLK limits			*/
};

/* This table holds the default vaules for the Trio64 VGA+ registers.	*/
/* A copy of it will be filled in as appropriate during mode		*/
/* programming in kgim_chipset_set_mode().				*/
static struct statetable default_state =
{
	0x23,							/* MISC	*/
	0x00,							/* FCTL	*/
	{
		0x03, 0x00, 0x0F, 0x20, 0x02			/* SEQ	*/
	},
	{
		0x00, 0x00, 0x00, 0x80, 0x00, 0x00, 0x00, 0x10,	/* CRT	*/
		0x00, 0x40, 0x00, 0x1F, 0x00, 0x00, 0xFF, 0xFF,
		0x00, 0x00, 0x00, 0x00, 0x1F, 0x00, 0x00, 0xE3,	/* 0x1x */
		0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x2x */
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x01, 0x00, 0x00, 0x10, 0x00, 0x00, 0x00,	/* 0x3x */
		0x48, 0xA5, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x4x */
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x38, 0x00, 0x00, 0x00,	/* 0x5x */
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x6x */
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x7x */
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
	},
	{
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* GRC	*/
		0xFF
	},
	{
		0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,	/* ATC	*/
		0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,
		0x84, 0x00, 0x0F, 0x00, 0x00
	}
};

/* VGA I/O register space */
static const struct io_region s3_io_vga = 
{
	IO_NULL,			/* IO region start (virtual)	*/
	IO_VGA_BASE,			/* IO region start (physical)	*/
	IO_VGA_SIZE,			/* IO region size		*/
	0xFFFF,				/* Decode all address bits	*/
	"S3 VGA chipset IO"		/* IO region name		*/
};

/* S3 accelerator I/O register space */
static const struct io_region s3_io_accel = 
{
	IO_NULL,			/* IO region start (virtual)	*/
	IO_ACCEL_BASE,			/* IO region start (physical)	*/
	IO_ACCEL_SIZE,			/* IO region size		*/
	0x03FF,				/* Decode 10 address bits only	*/
	"S3 accelerator I/O"		/* IO region name		*/
};

/* S3 MMIO (old style) register space */
static const struct mem_region s3_mem_accel = 
{
	MEM_NULL,			/* Mem region start (virtual)	*/
	0xA0000,			/* Mem region start (physical)	*/
	0x10000,			/* Mem region size		*/
	-1,				/* Decode all address bits	*/
	"S3 accelerator mem"		/* Mem region name		*/
};

/* Trio64 linear address window space */
static const struct mem_region s3_764_mem_law = 
{
	MEM_NULL,			/* Mem region start (virtual)	*/
	0,				/* Mem region start (physical)	*/
	0x0080000,			/* Mem region size		*/
	-1,				/* Decode all address bits	*/
	"Trio64 LAW"			/* Mem region name		*/
};

/* Trio64V+ linear address window space */
static const struct mem_region s3_765_mem_law = 
{
	MEM_NULL,			/* Mem region start (virtual)	*/
	0,				/* Mem region start (physical)	*/
	0x04000000,			/* Mem region size		*/
	-1,				/* Decode all address bits	*/
	"Trio64V+ LAW"			/* Mem region name		*/ 
};

/* Probe the hardware for default values or use hardcoded defaults */
static void s3_examine_chipset(struct kgi_display *dpy)
{
	#define COPYDEFAULT(X) (default_state.CRT[X] = CRT_in8(dpy, X))
	register uint8 foo;
	GET_PRIVATE;
	
	foo = COPYDEFAULT(0x36);
	switch(foo & CR36_MEMSIZE_MASK) 
	{

	case CR36_1MB:
		CHIPSET_INFO.ramsize = 1 MB;
		MEM_LAW_REGION.size = 1 MB;
		default_state.CRT[0x58] |= CR58_LAW_1MB;
		break;

	case CR36_2MB:
		CHIPSET_INFO.ramsize = 2 MB;
		MEM_LAW_REGION.size = 2 MB;
		default_state.CRT[0x58] |= CR58_LAW_2MB;
		break;

	case CR36_4MB:
		CHIPSET_INFO.ramsize = 4 MB;
		MEM_LAW_REGION.size = 4 MB;
		default_state.CRT[0x58] |= CR58_LAW_4MB;
		break;

	default:
		CHIPSET_INFO.ramsize = 1 MB;
		MEM_LAW_REGION.size = 1 MB;
		default_state.CRT[0x58] |= CR58_LAW_1MB;
		NOTICE1("WARNING: Unknown RAM size, assuming 1MB!");
	}
	
	MEM_LAW_REGION.size = CHIPSET_INFO.ramsize;

#if GRAPH_BASE
#error need default register values!
#else

	COPYDEFAULT(CRT_CONFIG2);
	default_state.CRT[0x3A] |= 
		(CRT_in8(dpy, 0x3A) & (CR3A_PCIMASK | CR3A_REFRESHMASK));
	default_state.CRT[0x40] |= (CRT_in8(dpy, 0x40) & 0xf0);
	default_state.CRT[0x58] |= (CRT_in8(dpy, 0x58) & CR58_VL_T1_LATCH);
	COPYDEFAULT(CRT_MEMCONTROL3);
	COPYDEFAULT(CRT_MEMCONTROL4);
	COPYDEFAULT(0x62);
	default_state.CRT[0x66] |= (CRT_in8(dpy, 0x66) & CR66_PCI_DISCONNECT);
	COPYDEFAULT(CRT_CONFIG3);

	default_state.CRT[0x3A] = 0x01;

	DEBUG0("default values: %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x",
	       default_state.CRT[CRT_CONFIG2],
	       default_state.CRT[0x3a],
	       default_state.CRT[0x40],
	       default_state.CRT[0x58],
	       default_state.CRT[CRT_MEMCONTROL3],
	       default_state.CRT[CRT_MEMCONTROL4],
	       default_state.CRT[0x62],
	       default_state.CRT[0x66],
	       default_state.CRT[CRT_CONFIG3]);

#endif

	TRACE2(dump_state(&default_state));

	private->ptr_and	= 0x00FF;
	private->ptr_xor	= 0x6F00;
	private->cur_and	= 0x00FF;
	private->cur_xor	= 0x7F00;

	private->lram.next	= NULL;
	private->lram.name	= "Video RAM";
	private->lram.type	= MMAP_TYPE_MMIO | MMAP_PUBLIC | MMAP_FRAMEBUFFER;
	private->lram.prot	= MMIO_RDWR | MMIO_SHARED;
	private->lram.psize	= private->lram.size = CHIPSET_INFO.ramsize;
	private->lram.pstart	= MEM_LAW_REGION.base;

#undef COPYDEFAULT
}


/* Verify the presence of a Trio64V+ (S3 86c765) */
static int s3_verify_86c765(struct kgi_display *dpy)
{
	GET_PRIVATE;
	mem_addr base0 = 0;
	struct kgim_options_pci *pci = KGIM_OPTIONS(dpy, pci);
	struct kgim_options_misc *misc = KGIM_OPTIONS(dpy, misc);

	DEBUG2("testing for S3 Vision 86c765, dev = %.8x", pci->dev);

	memset(private, 0, sizeof(*private));
	private->mem_vga   = MEM_NULL;
	private->mem_accel = MEM_NULL;

	IO_VGA_REGION      = s3_io_vga;
	IO_ACCEL_REGION    = s3_io_accel;
	MEM_ACCEL_REGION   = s3_mem_accel;
	MEM_LAW_REGION     = s3_765_mem_law;
	CHIPSET_INFO       = s3_765_info;

	if (pci->dev != PCICFG_NULL) 
	{
		/*	We got passed a PCI device number.  Verify that the 
		**	device is present and is a Trio64V+.  If it is not the
		**	primary display device, the PCI BIOS may have put it to
		**	sleep, so it may need to be woken up.
		*/

		uint16 cmd;
		#define IO_ENABLED      (cmd & PCI_COMMAND_IO)
		#define MEM_ENABLED     (cmd & PCI_COMMAND_MEMORY)
		#define ENABLED         (cmd & (PCI_COMMAND_IO | \
						PCI_COMMAND_MEMORY))

		/*	First, we need to make sure we are indeed looking at a
		**	Trio64V+ before we can do anything.
		*/
		if ((pcicfg_in16(pci->dev + PCI_VENDOR_ID) != 
		     PCI_VENDOR_ID_S3)
		    || ((pcicfg_in16(pci->dev + PCI_DEVICE_ID)) !=
			PCI_DEVICE_ID_S3_TRIO)) 
		{

			NOTICE1("PCI device is not a Trio64V+");
			return -E(CHIPSET, UNKNOWN);                    
		}

		/*	Now we check if the specified LAW window is available,
		**	so we don't end up having to restore any hardware
		**	settings in case it isn't. 
		*/
		cmd   = pcicfg_in16(pci->dev + PCI_COMMAND);
		pcicfg_out16(PCI_COMMAND_IO | PCI_COMMAND_MEMORY,
			     pci->dev + PCI_COMMAND);
		
		base0 = pcicfg_in32(pci->dev + PCI_BASE_ADDRESS_0);

		MEM_LAW_REGION.base = (misc->law_base 
			? misc->law_base : base0) & 0xFF800000;
		MEM_LAW_REGION.size = 64 MB;
		if ((MEM_LAW_REGION.base == 0) ||
			mem_check_region(&MEM_LAW_REGION)) {

			ERROR("new mem_law region busy.");
			return -EBUSY;
		}

		if (KGIM_OPTIONS(dpy, chipset) && (strcmp("AEON-MSP",
			KGIM_OPTIONS(dpy, chipset)->chipset) == 0)) {

			DEBUG0("AEON-MSP option given, enabling chipset");
			pcicfg_out16(cmd | PCI_COMMAND_MEMORY | PCI_COMMAND_IO,
				pci->dev + PCI_COMMAND);
			cmd |= PCI_COMMAND_MEMORY | PCI_COMMAND_IO;
		}

		/*	Ok, there _is_ a PCI Trio64V+.  This one might still
		**	be sleeping, so we may have to wake it up. If it is
		**	not sleeping, it might be the primary display or it
		**	might already be handled by another driver.
		*/
			
		if (ENABLED) 
		{
			
			/* !!!	remember for restore ?
			**	was_primary = TRUE;
			*/

			/*	It is active, so it may be the primary 
			**	display adapter, but it may also be handled
			** 	by another driver already.  So, we first make
			** 	sure no other driver is already handling it.
			*/
			
			/* 	Check if the PCI linear address window for
			**	this display adapter is available.
			*/
			MEM_LAW_REGION.base = base0;
			MEM_LAW_REGION.size = 0x1000; /* 4K */
			if (base0 && MEM_ENABLED && 
				mem_check_region(&MEM_LAW_REGION)) 
			{
				
				ERROR("mem_law region busy.");
				return -EBUSY;
			}

			/*	Check if the VGA and accel I/O ports are
			** 	available.
			*/
			if (IO_ENABLED && 
				(io_check_region(& IO_VGA_REGION) ||
				io_check_region(& IO_ACCEL_REGION))) {

				ERROR("io_vga or io_accel region busy.");
				return -EBUSY;
			}

			/*	Take control of the VGA and accel I/O port
			**  	regions.
			*/
			io_claim_region(&IO_VGA_REGION);
			io_claim_region(&IO_ACCEL_REGION);

			/*	Now that we have taken control of the card, we
			**	need to	enable S3 'new MMIO' mode, which 
			**	allows us to relocate the LAW anywhere in the
			**	PCI address space and gives us access to 
			**	memory-mapped registers within that LAW.  
			*/

			/* Unlock access to the S3 extended VGA registers
			*/
			s3_unlock(dpy);

			/* Switch to MMIO-only mode
			*/
			CRT_out8(dpy, CRT_in8(dpy,0x58) | CR58_LAW_4MB | 
				CR58_LAW_ENABLE, 0x58);
			CRT_out8(dpy, CR53_MMIO_RELOC_MMIO, 0x53);
			
			/* Disable the card's I/O port access
			*/
#if 0
			NOTICE1("I/O ports switched off!");
			SEQ_out8(dpy, SR09_MMIO_ONLY, 0x09);
			default_state.SEQ[0x09] |= SR09_MMIO_ONLY;
#endif
			/* Release our lock on the I/O port regions, since
			** we do not need them in MMIO mode.
			*/
			io_free_region(& IO_ACCEL_REGION);
			io_free_region(& IO_VGA_REGION);

		} else /* if (ENABLED) */
		{ 
			
			/*	The display adapter is present but disabled.
			** 	We need to wake it up.
			*/

			DEBUG1("Waking up sleeping Trio64V+.");

			pcicfg_out16(PCI_COMMAND_IO | PCI_COMMAND_MEMORY,
				pci->dev + PCI_COMMAND);
			io_out8(0x01, 0x3c3);
		}

		/*	Everything should be up and running now, so we can
		** 	claim the card's LAW for ourselves.
		*/
		
		MEM_LAW_REGION.size = 0x4000000; /* 64 MB */
		MEM_LAW_REGION.base = (misc->law_base
			? misc->law_base : base0) & ~(MEM_LAW_REGION.size - 1);

		DEBUG1("mem_law_region @ %.8x, size %x", MEM_LAW_REGION.base,
			MEM_LAW_REGION.size);

		mem_claim_region(&MEM_LAW_REGION);
		pcicfg_out32(MEM_LAW_REGION.base, pci->dev +
			PCI_BASE_ADDRESS_0);

		/*	Setting SR08 doesn't seem to be enough. So we just
		**	disable it via PCI configuration space (again).
		*/
		pcicfg_out16(PCI_COMMAND_MEMORY, pci->dev + PCI_COMMAND);

		MEM_VGA   = MEM_LAW + 0x1008000;
		MEM_ACCEL = MEM_LAW + 0x1000000;

		s3_unlock(dpy);

		#if DEBUG_LEVEL > 0
			
			/*	Verify it again
			*/

			if ((CRT_in8(dpy, 0x30) != CR30_USE_CR2x) ||
				(CRT_in8(dpy, 0x2D) != CR2D_TRIO) ||
				(CRT_in8(dpy, 0x2E) != CR2E_TRIO64)) {

				ERROR("!! 80c765 MMIO verification failed !!");
				MEM_LAW = mem_free_region(& MEM_LAW_REGION);

				io_free_region(& IO_ACCEL_REGION);
				io_free_region(& IO_VGA_REGION);
				return -E(CHIPSET, UNKNOWN);
			}
		#endif

		/* Update softcopy registers to reflect the MMIO settings we 
		** just made.
		*/
		default_state.CRT[0x58] |= CR58_LAW_ENABLE | CR58_LAW_4MB;
		default_state.CRT[0x53] = CR53_MMIO_RELOC_MMIO;
		default_state.CRT[0x59] = (MEM_LAW_REGION.base >> 24) & 0xFF;
		default_state.CRT[0x5A] = (MEM_LAW_REGION.base >> 16) & 0xFF;

		/* Now we need to determine the card's features.
		*/
		s3_examine_chipset(dpy);

		return EOK;

	} else /* if (pci->dev != PCICFG_NULL)	*/
	{
		
		/*	We didn't get passed a PCI device number, so we need
		** 	to probe for a Trio64V+ ourselves.
		*/

		uint8 CR53, CR58, CR59, CR5A;

		/*	Check if anyone else has locked the VGA and accel I/O
		** 	ports.  If they have, we are out of luck.
		*/
		if (io_check_region(& IO_VGA_REGION) ||
			io_check_region(& IO_ACCEL_REGION)) {

			ERROR("io_vga or io_accel region busy.");
			return -EBUSY;
		}
		
		/*	Take control of the VGA and accel I/O port
		**  	regions.
		*/
		io_claim_region(& IO_ACCEL_REGION);
		io_claim_region(& IO_VGA_REGION);

		/* Unlock access to the S3 extended VGA registers
		*/
		s3_unlock(dpy);

		/*	Probe the primary display's I/O ports to see if it is
		** 	a Trio64V+.
		*/
		if ((CRT_in8(dpy, 0x30) != CR30_USE_CR2x) ||
			(CRT_in8(dpy, 0x2D) != CR2D_TRIO) ||
			(CRT_in8(dpy, 0x2E) != CR2E_TRIO64)) 
		{

			NOTICE2("S3 Vision 80c765 I/O-verification failed.");
			io_free_region(& IO_ACCEL_REGION);
			io_free_region(& IO_VGA_REGION);
			return -E(CHIPSET, UNKNOWN);
		}

		/*	The primary display adapter is a Trio64V+.  Now we 
		**	need to make sure that we can operate it in new
		**	MMIO mode.
		*/
		CR53 = CRT_in8(dpy, 0x53);
		CR58 = CRT_in8(dpy, 0x58);
		CR59 = CRT_in8(dpy, 0x59);
		CR5A = CRT_in8(dpy, 0x5A);

		base0 = (CR59 << 24) | (CR5A << 16);
		MEM_LAW_REGION.size = 0x4000000; /* 64 MB */
		MEM_LAW_REGION.base = (misc->law_base 
			? misc->law_base : base0) & ~(MEM_LAW_REGION.size - 1);
		if (!MEM_LAW_REGION.base || mem_check_region(&MEM_LAW_REGION))
		{

			ERROR("(new) mem_law region busy.");
			io_free_region(& IO_ACCEL_REGION);
			io_free_region(& IO_VGA_REGION);
			return -EBUSY;                 
		}

		default_state.CRT[0x53] = CR53_MMIO_RELOC_MMIO;
		default_state.CRT[0x58] |= CR58_LAW_ENABLE | CR58_LAW_4MB;
		default_state.CRT[0x59] = (MEM_LAW_REGION.base >> 24) & 0xFF;
		default_state.CRT[0x5A] = (MEM_LAW_REGION.base >> 16) & 0xFF;

		CRT_out8(dpy, default_state.CRT[0x59], 0x59);
		CRT_out8(dpy, default_state.CRT[0x5A], 0x5A);
		CRT_out8(dpy, CRT_in8(dpy, 0x58) | CR58_LAW_4MB |
			CR58_LAW_ENABLE, 0x58);
		CRT_out8(dpy, CR53_MMIO_RELOC_MMIO, 0x53);

		mem_claim_region(& MEM_LAW_REGION);
		MEM_VGA   = MEM_LAW + 0x1008000;
		MEM_ACCEL = MEM_LAW + 0x1000000;

		/*	It should be responding to new MMIO and I/O, so we
		**	release the io_regions and move to MMIO only mode.
		*/
		#if DEBUG_LEVEL > 0

			if ((CRT_in8(dpy, 0x30) != CR30_USE_CR2x) ||
				(CRT_in8(dpy, 0x2D) != CR2D_TRIO) ||
				(CRT_in8(dpy, 0x2E) != CR2E_TRIO64)) {

				ERROR("!! 86c765 MMIO verification failed !!");
				MEM_LAW = mem_free_region(& MEM_LAW_REGION);

				CRT_out8(dpy, CR53, 0x53);
				CRT_out8(dpy, CR58, 0x58);
				CRT_out8(dpy, CR59, 0x59);
				CRT_out8(dpy, CR5A, 0x5A);
				
				io_free_region(& IO_ACCEL_REGION);
				io_free_region(& IO_VGA_REGION);
				return -E(CHIPSET, UNKNOWN);
			}
		#endif
#if 0
		NOTICE1("I/O ports switched off!");
		SEQ_out8(dpy, SR09_MMIO_ONLY, 0x09);
		default_state.SEQ[0x09] |= SR09_MMIO_ONLY;
#endif
		/* ???	Disabling I/O using SR08 does not seem to work as
		** ???	intended. Thus we use 'internal ' PCICFG space.
		*/
		mem_out8(PCI_COMMAND_MEMORY, MEM_VGA + 4);

		io_free_region(& IO_ACCEL_REGION);
		io_free_region(& IO_VGA_REGION);
	}

	DEBUG2("%s rev %i detected.", CHIPSET_INFO.model, \
		CRT_in8(dpy, CRT_CHIP_REV));
	s3_examine_chipset(dpy);

	return EOK;
}



/* KGI module interface */

int kgim_chipset_init(struct kgi_card *card, struct kgi_display *dpy)
{
	ggi_uint foo;
	struct chipset_private *private;

	NOTICE(DRIVER_NAME " rev " DRIVER_REV);

	private = kgi_malloc_small(sizeof(*private), GFP_KERNEL);
	if (private == NULL) 
	{

		NOTICE2("ERROR: failed to allocate private data!");
		return -ENOMEM;
	}

	memset(private, 0, sizeof(*private));
	KGIM_PRIVATE(dpy, chipset) = private;
	card->chipset = &CHIPSET_INFO;

	if (s3_verify_86c765(dpy)) 
	{

		ERROR("failed to detect Trio64V+!");
		kgi_free_small(private);
		KGIM_PRIVATE(dpy, chipset) = NULL;
		card->chipset = NULL;
		return -E(CHIPSET, UNKNOWN);
	}

	/* Select PLL to drive DCLK */
	foo = MISC_in8(dpy);
	if ((foo & MISC_CLOCK_MASK) != MISC_CLOCK_MASK) 
	{

		MISC_out8(dpy, (foo & ~MISC_CLOCK_MASK) | MISC_EXT_CLOCK);
	}

	for (foo = 0; foo < 10000000; foo++);

	return EOK;
}

void kgim_chipset_done(struct kgi_display *dpy)
{
	GET_PRIVATE;

	s3_lock(dpy);	/* !!! Check and clean up! */

	if (IO_VGA_REGION.baseptr) 
	{

		io_free_region(&IO_VGA_REGION);
	}
	
	if (IO_ACCEL_REGION.baseptr) 
	{

		io_free_region(&IO_ACCEL_REGION);
	}
	
	if (MEM_ACCEL_REGION.baseptr) 
	{

		mem_free_region(&MEM_ACCEL_REGION);
	}
	
	if (MEM_LAW_REGION.baseptr) 
	{

		mem_free_region(&MEM_LAW_REGION);
	}

	kgi_free_small(private);
	KGIM_PRIVATE(dpy, chipset) = NULL;
	NOTICE(DRIVER_NAME " rev " DRIVER_REV " removed.");
}

static void adjust_timing(struct kgi_mode *mode)
{
	register int foo = (mode->request.graphtype == GT_TEXT16) 
				? mode->request.text.x : 8;

	#define ADJUST(X) mode->x.X -= (mode->x.X % foo)

	ADJUST(width);
	ADJUST(blankstart);
	ADJUST(syncstart);
	ADJUST(syncend);
	ADJUST(blankend);

	if (mode->x.total % foo) 
	{

		mode->x.total += foo - (mode->x.total % foo);
	}

	if (((mode->x.syncend / foo) & 0x3F) ==
		((mode->x.syncstart / foo) & 0x3F)) 
	{

		mode->x.syncend -= foo;
	}

	if (((mode->x.blankend / foo) & 0x7F) ==
		((mode->x.blankstart / foo) & 0x7F)) 
	{

		mode->x.blankend -= foo;
	}

	if ((mode->y.syncend & 0x0F) == (mode->y.syncstart & 0x0F)) 
	{

		mode->y.syncend--;
	}

	if ((mode->y.blankend & 0xFF) == (mode->y.blankstart & 0xFF)) 
	{

		mode->y.blankend--;
	}

	#undef ADJUST
}

int kgim_chipset_check_mode (struct kgi_display *dpy, struct kgi_mode *mode,
	enum kgi_tm_cmd cmd)
{
	GET_PRIVATE;
	int Bpp = 1;

	if (mode->flags & FL_MON_INTERLACE) 
	{

		NOTICE1("Interlaced modes not supported yet\n");
		return -E(CHIPSET, UNKNOWN);
	}

	switch(mode->request.graphtype) 
	{

	case GT_32BIT:	
		Bpp +=4;	/* (4+2+1+1)/2 = 4 Bpp	*/
	case GT_15BIT:
	case GT_16BIT:	
		Bpp +=2;	/*  (2+1+1)/2 =   2 Bpp	*/
	case GT_8BIT:
		Bpp +=1;	/*    (1+1)/2 =   1 Bpp	*/
	case GT_4BIT:  		/*        1/2 = 0.5 Bpp	*/
	case GT_TEXT16:	
		break;

	default:
		DEBUG1("graphtype %i not supported", mode->request.graphtype);
		return -E(CHIPSET, UNKNOWN);
	}

	if (mode->lclk.mul != 1) 
	{

		NOTICE1("Invalid lclk multiplier");
 		return -E(CHIPSET, UNKNOWN);
	}

	switch (mode->lclk.div) 
	{

	case 1:
	case 2:
	case 4:
	case 8:
	case 16:
	case 32:
		break;

	default:
		DEBUG1("invalid CLK to LCLK divider %i", mode->lclk.div);
		return -E(CHIPSET, UNKNOWN);
	}

	#define	XVIRTUAL	(mode->request.virt.x)
	#define	YVIRTUAL	(mode->request.virt.y)

	switch (cmd) 
	{

	case CMD_PROPOSE:
		
		if (mode->request.graphtype == GT_TEXT16) 
		{

			int textx = mode->request.text.x;
			int texty = mode->request.text.y;

			mode->magnify.x = mode->magnify.y = 1;

			if ((textx < 7) || (9 < textx)) 
			{

				DEBUG1("ERROR: Invalid char width %i", textx);
				return -EINVAL;
			}

			if (!(XVIRTUAL || YVIRTUAL)) 
			{

				NOTICE1("ERROR: Need x or y virtual!");
				return -EINVAL;
			}

			if (!XVIRTUAL)
			{

				XVIRTUAL = TEXT_LEN * textx / 
					(YVIRTUAL / texty);
				DEBUG1("Setting xvirtual = %i", XVIRTUAL);
			}

			if (!YVIRTUAL) 
			{

				YVIRTUAL = TEXT_LEN * texty / 
					(XVIRTUAL / textx);
				DEBUG1("Setting yvirtual = %i", YVIRTUAL);
			}

			if (((XVIRTUAL / textx) * (YVIRTUAL / texty)) > 
			    TEXT_LEN) 
			{

				NOTICE1("ERROR: Not enough text memory");
				return -ENOMEM;
			}

			mode->request.visible.x -=
				(mode->request.visible.x % textx);
			mode->request.visible.y -=
				(mode->request.visible.y % texty);

		} else /* mode->request.graphtype == GT_TEXT16 */
		{

			mode->magnify.x = mode->magnify.y = 1;

			if (XVIRTUAL <= 640) 
			{
				
				XVIRTUAL = 640;
			}
			else if (XVIRTUAL <= 800)
			{
				
				XVIRTUAL = 800;
			}
			else if (XVIRTUAL <= 1024)
			{	
				
				XVIRTUAL = 1024;
			}
			else if (XVIRTUAL <= 1152)
			{
				
				XVIRTUAL = 1152;
			}
			else if (XVIRTUAL <= 1280)
			{
				
				XVIRTUAL = 1280;
			}
			else if (XVIRTUAL <= 1600)
			{
				
				XVIRTUAL = 1600;
			}
			else if (XVIRTUAL <= 2048)
			{
				
				XVIRTUAL = 2048;
			}
			else 
			{

				DEBUG1("Too many horizontal pixels (%i)", 
				       XVIRTUAL);
				return -E(CHIPSET, UNKNOWN);
			}

			if ((XVIRTUAL* YVIRTUAL* Bpp / 2) >
				CHIPSET_INFO.ramsize) 
			{

				DEBUG1("not enough memory for %ix%i@%ibpp",
					XVIRTUAL, YVIRTUAL, Bpp * 4);
				return -ENOMEM;
			}

			mode->request.visible.x &= 0x7FFFFFF8;
			mode->request.visible.y &= 0x7FFFFFF8;
		}

		mode->x.width = mode->request.visible.x;
		mode->y.width = mode->request.visible.y;
		return EOK;

	case CMD_LOWER:

		if (mode->dclk < CHIPSET_INFO.dclk.min) 
		{

			DEBUG1("ERROR: DCLK too low (%i)\n", mode->dclk);
			return -E(CHIPSET, UNKNOWN);
		}

		mode->dclk = (mode->dclk > CHIPSET_INFO.dclk.max)
			? CHIPSET_INFO.dclk.max : mode->dclk;

		adjust_timing(mode);
		return EOK;

	case CMD_RAISE:

		if (mode->dclk > CHIPSET_INFO.dclk.max) 
		{

			DEBUG1("ERROR: DCLK too high (%i)\n", mode->dclk);
			return -E(CHIPSET, UNKNOWN);
		}

		mode->dclk = (mode->dclk < CHIPSET_INFO.dclk.min)
			? CHIPSET_INFO.dclk.min : mode->dclk;

		adjust_timing(mode);
		return EOK;

	case CMD_CHECK:

		if (mode->request.graphtype == GT_TEXT16) 
		{

			if ((mode->magnify.x != 1) || 
			    (mode->magnify.y != 1) || 
			    (mode->request.text.x < 7) ||
			    (mode->request.text.x > 9) ||
			    (mode->request.text.y < 1) ||
			    (mode->request.text.y > 32) ||
			    (((XVIRTUAL/mode->request.text.x) *
			      (YVIRTUAL/mode->request.text.y)) > TEXT_LEN) ||
			    (mode->dclk > CHIPSET_INFO.dclk.max) ||
			    (mode->dclk < CHIPSET_INFO.dclk.min)) 
			{

				return -E(CHIPSET, UNKNOWN);
			}

			mode->flags |= FL_CHIP_COLORTEXT;
			mode->fb = MEM_LAW_REGION.baseptr;
			mode->fb_size = TEXT_LEN;
			mode->fsize = (8 * mode->request.virt.x *
				mode->request.virt.y) /
				(mode->request.text.x * mode->request.text.y);
			mode->io = &text16_color_io;
			mode->mem = &text16_color_mem;

		} else 
		{

			switch (XVIRTUAL) 
			{

			case 640:
			case 800:
			case 1024:
			case 1152:
			case 1280:
			case 1600:
			case 2048:
				break;

			default:
				return -E(CHIPSET, UNKNOWN);
			}

			if (((XVIRTUAL * YVIRTUAL * Bpp / 2) >
			     CHIPSET_INFO.ramsize) || 
			    (mode->magnify.x != 1) ||
			    (mode->magnify.y != 1) ||
			    (mode->clk.mul !=1) || 
			    (mode->clk.div != 1) || 
			    (mode->dclk < CHIPSET_INFO.dclk.min) || 
			    (CHIPSET_INFO.dclk.max<mode->dclk)) 
			{

				return -E(CHIPSET, UNKNOWN);
			}

			mode->fb	= MEM_LAW_REGION.baseptr;
			mode->fb_size	= CHIPSET_INFO.ramsize;
			mode->fsize	= mode->request.virt.x * 
				mode->request.virt.y * Bpp / 2;
			mode->io	= NULL;
			mode->mem	= NULL;
		}

#define TM(X) (((unsigned) CHIPSET_INFO.maxdots.x < mode->x.X)\
|| ((unsigned) CHIPSET_INFO.maxdots.y < mode->y.X))

		if (TM(width) || TM(blankstart) || TM(syncstart) || 
			TM(syncend) || TM(blankend) || TM(total)) 
		{

			return -E(CHIPSET, UNKNOWN);
		}
#undef TM

		adjust_timing(mode);
		mode->mmio = & private->lram;

		return EOK;

	default:
		INTERNAL_ERROR;
		return -E(CHIPSET, UNKNOWN);
	}

#undef XVIRTUAL
#undef YVIRTUAL
}

void kgim_chipset_set_mode (struct kgi_display *dpy, struct kgi_mode *mode)
{
	IS_CRITICAL;
	GET_PRIVATE;
	struct statetable setup = default_state;
	register int foo;
	int DCLKdiv;
	int sr8, sr15, sr18, invert_vclk = 0;

	/* Set up bits per pixel and logical screen width */
	switch(mode->request.graphtype) 
	{

	case GT_TEXT16:
		foo = (2 * mode->request.virt.x / mode->request.text.x) / 4;
		break;

	case GT_4BIT:
		setup.CRT[0x50] |= CR50_4BPP;
		foo = mode->request.virt.x >> 4;
		break;

	case GT_8BIT:
		setup.CRT[0x3A] |= CR3A_NO4BPPMODE;
		setup.CRT[0x50] |= CR50_8BPP;
		foo = mode->request.virt.x >> 3;
		break;

	case GT_15BIT:
	case GT_16BIT:
		setup.CRT[0x3A] |= CR3A_NO4BPPMODE;
		setup.CRT[0x50] |= CR50_16BPP;
		setup.CRT[0x43] = 0x09;
		foo = mode->request.virt.x >> 2;
		break;

	case GT_32BIT:
		setup.CRT[0x3A] |= CR3A_NO4BPPMODE;
		setup.CRT[0x50] |= CR50_32BPP;
		foo = mode->request.virt.x >> 1;
		break;
	default:
		INTERNAL_ERROR;
		return;
	}

	setup.CRT[CRT_LOGICALWIDTH]  |= foo & 0xFF;
	setup.CRT[0x51] |= (foo & 0x100) ? CR51_LOGICALWIDTH_B8 : 0;
	setup.CRT[0x51] |= (foo & 0x200) ? CR51_LOGICALWIDTH_B9 : 0;

	if (mode->request.graphtype == GT_TEXT16) 
	{

		setup.CRT[0x17] = CR17_CGA_ADDR_WRAP | CR17_WORDMODE |
			CR17_ENABLE_SYNC | CR17_CGA_BANKING | 
			CR17_HGC_BANKING;
		setup.CRT[0x31] |= CR31_HISPEEDTXT;

		setup.SEQ[0x02] = SR02_PLANE0 | SR02_PLANE1;

		setup.GRC[0x05] |= GR5_ODD_EVEN;
		setup.GRC[0x06] |= GR6_MAP_B8_32 | GR6_CHAIN_OE;

		setup.CRT[0x34] |= CR34_ENB_FIFO; 

		switch((DCLKdiv = mode->request.text.x)) 
		{

		case 7: 
			setup.CRT[0x32] |= 0x01; 
			break;
			
		case 8: 
			setup.SEQ[0x01] |= SR01_8DOT_CHCLK; 
			break;
			
		case 9: 
			break;
			
		default:
			INTERNAL_ERROR;
			return;
		}

		setup.CRT[0x09] |= mode->request.text.y - 1;
		setup.ATC[0x13] = (mode->request.text.x == 9) ? 0x08 : 0x00;

	} else /* mode->request.graphtype == GT_TEXT */
	{

		DCLKdiv = 8;

		setup.SEQ[0x01] |= SR01_8DOT_CHCLK;
		setup.SEQ[0x02] |= SR02_PLANE_MASK; /* ??? neccessary? */
		setup.ATC[0x10] |= AR10_GRAPHICS;
		setup.GRC[0x05] |= GR5_8BIT_COLOR; /* ??? neccessary? */
		setup.GRC[0x06] |= GR6_GRAPHMODE | GR6_MAP_A0_64 /*??? needed? */;
		setup.CRT[0x31] |= CR31_ENHANCED;
		setup.CRT[0x58] |= CR58_LAW_ENABLE;
		setup.CRT[0x09] |= 0;
			/* !!! TM->ymagnify-1 & CR09_CHARHEIGHT_MASK; */
		setup.CRT[0x09] |= (mode->flags & FL_CHIP_DOUBLESCAN)
			? CR09_DOUBLESCAN : 0;

		switch(mode->request.virt.x) 
		{
 
		case  640:
			setup.CRT[0x50] |= CR50_640;
			break;
		case  800:
			setup.CRT[0x50] |= CR50_800;
			break;
		case 1024:
			setup.CRT[0x50] |= CR50_1024;
			break;
		case 1152:
			setup.CRT[0x50] |= CR50_1152;
			break;
		case 1280:
			setup.CRT[0x50] |= CR50_1280;
			break;
		case 1600:
			setup.CRT[0x50] |= CR50_1600;
			break;
		case 2048:
			setup.CRT[0x50] |= CR50_2048;
			setup.CRT[0x31] |= CR31_2PAGE_SCRN;
			break;
		default:
			INTERNALERROR;
		}

		foo = 0;
		switch(mode->request.graphtype) 
		{

		case GT_TEXT16:
		case GT_4BIT:
		case GT_8BIT:
			if (mode->dclk < 67500000) 
			{

				foo = CR67_DAC_MODE0;
			} else 
			{

				foo = CR67_DAC_MODE8;
				invert_vclk = 2;
			}
			break;
			
		case GT_15BIT:
			foo = CR67_DAC_MODE9;
			setup.CRT[0x33] |= 0x08;
			break;
			
		case GT_16BIT:
			foo = CR67_DAC_MODEA;
			setup.CRT[0x33] |= 0x08;
			break;

		case GT_24BIT:
		case GT_32BIT:
			foo = 0xd0;
			break;

		default:
			INTERNAL_ERROR;
			return;
		}

		setup.CRT[0x67] = foo | invert_vclk ;
	}

	/* 	Set up horizontal timing 
	*/

	/* Horizontal display end 
	*/
	foo = (mode->x.width / DCLKdiv) - 1;
	setup.CRT[0x01] |= foo & 0xFF;
	setup.CRT[0x5D] |= (foo & 0x100) ? CR5D_HDISPLAYEND_B8 : 0;

	/* Horizontal blank start 
	*/
	foo = (mode->x.blankstart / DCLKdiv);
	setup.CRT[0x02] |= foo & 0xFF;
	setup.CRT[0x5D] |= (foo & 0x100) ? CR5D_HBLANKSTART_B8 : 0;

	/* Horizontal sync start 
	*/
	foo = (mode->x.syncstart / DCLKdiv);
	setup.CRT[0x04] |= foo & 0xFF;
	setup.CRT[0x5D] |= (foo & 0x100) ? CR5D_HSYNCSTART_B8 : 0;

	/* Horizontal sync end 
	*/
	foo = (mode->x.syncend / DCLKdiv);
	setup.CRT[0x05] |= foo & CR05_HSE_MASK;
	setup.CRT[0x5D] |= ((foo-(mode->x.syncstart / DCLKdiv)) > 0x20)
		? CR5D_HSYNCEND_B5 : 0;

	/* Horizontal blank end 
	*/
	foo = (mode->x.blankend / DCLKdiv);
	setup.CRT[0x03] |= foo & CR03_HBE_MASK;
	setup.CRT[0x05] |= (foo & 0x20) ? CR05_HBLANKEND_B5 : 0;
	setup.CRT[0x5D] |= ((foo-(mode->x.blankstart / DCLKdiv)) > 0x40)
		? CR5D_HBLANKEND_B6 : 0;

	/* Horizontal total
	*/
	foo = (mode->x.total / DCLKdiv) - 5;
	setup.CRT[0x00] |= foo & 0xFF;
	setup.CRT[0x5D] |= (foo & 0x100) ? CR5D_HTOTAL_B8 : 0;

	/* 	Set up vertical timing 
	*/

	/* Vertical display end 
	 */
	foo = mode->y.width - 1;
	setup.CRT[0x12] |= foo & 0xFF;
	setup.CRT[0x07] |= (foo & 0x100) ? CR07_VDISPLAYEND_B8  : 0;
	setup.CRT[0x07] |= (foo & 0x200) ? CR07_VDISPLAYEND_B9  : 0;
	setup.CRT[0x5E] |= (foo & 0x400) ? CR5E_VDISPLAYEND_B10 : 0;

	/* Vertical blank start 
	*/
	foo = mode->y.blankstart - 1;
	setup.CRT[0x15] |= foo & 0xFF;
	setup.CRT[0x07] |= (foo & 0x100) ? CR07_VBLANKSTART_B8  : 0;
	setup.CRT[0x09] |= (foo & 0x200) ? CR09_VBLANKSTART_B9  : 0;
	setup.CRT[0x5E] |= (foo & 0x400) ? CR5E_VBLANKSTART_B10 : 0;

	/* Vertical sync start 
	*/
	foo = mode->y.syncstart;
	setup.CRT[0x10] |= foo & 0xFF;
	setup.CRT[0x07] |= (foo & 0x100) ? CR07_VSYNCSTART_B8  : 0;
	setup.CRT[0x07] |= (foo & 0x200) ? CR07_VSYNCSTART_B9  : 0;
	setup.CRT[0x5E] |= (foo & 0x400) ? CR5E_VSYNCSTART_B10 : 0;

	/* Vertical sync end
	*/
	foo = mode->y.syncend;
	setup.CRT[0x11] |= foo & CR11_VSYNCEND_MASK;

	/* Vertical blank end
	*/
	foo = mode->y.blankend + 1;
	setup.CRT[0x16] |= foo & 0xFF;

	/* Vertical total
	*/
	foo = mode->y.total - 2;
	setup.CRT[0x06] |= foo & 0xFF;
	setup.CRT[0x07] |= (foo & 0x100) ? CR07_VTOTAL_B8  : 0;
	setup.CRT[0x07] |= (foo & 0x200) ? CR07_VTOTAL_B9  : 0;
	setup.CRT[0x5E] |= (foo & 0x400) ? CR5E_VTOTAL_B10 : 0;

	/* FIXME: Support interlaced modes */
	setup.CRT[CRT_INTERLACESTART]|= mode->y.total >> 1;
	if(mode->flags & FL_MON_INTERLACE) {
 
		NOTICE2("Interlaced modes not yet supported.");
		setup.CRT[0x42] |= CR42_INTERLACED;
		INTERNAL_ERROR;
		return;
	}

	setup.MISC |= (mode->x.polarity > 0) ? 0 : MISC_NEG_HSYNC;
	setup.MISC |= (mode->y.polarity > 0) ? 0 : MISC_NEG_VSYNC;

	/* Set up the FIFO control parameters. This needs some tuning! 
	*/
	foo = ((mode->x.total + mode->x.syncstart) / 2) / DCLKdiv;
	setup.CRT[CRT_FIFOSTART] |= foo & 0xFF;	/* Start display FIFO fetch */
	setup.CRT[0x5D] |= (foo & 0x100) ? CR5D_FIFOSTART_B8 : 0;

	if (mode->request.graphtype == GT_TEXT16) 
	{

		setup.CRT[CRT_MEMCONTROL2] = 8;
		setup.CRT[CRT_MEMCONTROL3] = 7;
		setup.CRT[0x62] = 84;

	} else 
	{

		setup.CRT[CRT_MEMCONTROL2] = mode->chipset; /* M parameter */
		setup.CRT[CRT_MEMCONTROL3] = 0xFF; /* N parameter */

		foo = mode->request.visible.x;

		switch (mode->request.graphtype) 
		{

		case GT_4BIT:
			foo /= 2;
			break;
			
		case GT_8BIT:
			break;
			
		case GT_15BIT:
		case GT_16BIT:
			foo *= 2;
			break;
			
		case GT_32BIT:
			foo *= 4;
			break;

		default:
		}

		foo /= 4;
		foo++;

		/* foo /= (chipinfo.ramsize < 2048*1024) ? 4 : 8; */
		setup.CRT[CRT_MEMCONTROL4] |= (foo>>8) | 0x80;
		setup.CRT[0x62] = foo & 0xFF; 
	}

	TRACE2(dump_state(&setup));

	/* Ok.  We'll do the actual hardware programming now. 
	*/
	ENTER_CRITICAL;

	/* Wait for the accelerator to finish and disable it.
	*/
	if (dpy->mode.request.graphtype != GT_TEXT16) {

		foo = 10000;
		while ((ACCEL_in16(dpy, ECR_GP_STATUS) & 0xF2FF) && foo) 
		{

			foo--;
		}
		ACCEL_out16(dpy, 0, ECR_ADVFUNC_CNTL);
	}

	/* Switch sync signals off 
	*/
	CRT_out8(dpy, CRT_in8(dpy, CRT_EXTSYNC1) | 0x20, CRT_EXTSYNC1);

	/* Prepare subsystems 
	*/
	s3_unlock(dpy);
	SEQ_out8(dpy, SEQ_in8(dpy, 0x01) | SR01_DISPLAY_OFF, 0x01);

	kgim_monitor_set_mode(dpy, mode);
	NOTICE2("monitor set");
	
	kgim_ramdac_set_mode(dpy, mode);

	/* FIXME: The DAC and clock hardware programming code belongs in
	** the ramdac and clockchip drivers, not here! 
	*/
	
	/* Program the DAC 
	*/
	SEQ_out8(dpy, 0x06, 0x08);
	sr8  = SEQ_in8(dpy, 0x08);
	sr15 = SEQ_in8(dpy, 0x15) & ~0x50;
	sr18 = SEQ_in8(dpy, 0x18) & ~0x80;

	if(mode->request.graphtype == GT_8BIT && mode->dclk >= 67500000) 
	{

		sr15 |= 0x50;
		sr18 |= 0x80;
	}

	SEQ_out8(dpy, sr15, 0x15);
	SEQ_out8(dpy, sr18, 0x18);
	SEQ_out8(dpy, sr8, 0x08);

	NOTICE2("ramdac set");
	
	kgim_clock_set_mode(dpy, mode);

	restore_state(dpy, &setup);

	LEAVE_CRITICAL;

	memcpy(&(dpy->mode), mode, sizeof(dpy->mode));

	private->split = -1;
	private->org_xdot = private->org_ydot = private->org_offs = -1;

	if (mode->request.graphtype == GT_TEXT16) 
	{

		dpy->set_font	= s3_text_set_font;
		dpy->set_split	= s3_set_split;
		dpy->set_frame	= s3_set_frame;
		dpy->set_origin	= s3_text_set_origin;

		/* Provide hardware cursor in text mode */
		if (HARDWARE_TEXT_CURSOR) 
		{

			dpy->show_cursor = s3_show_hw_text_cursor;
			dpy->hide_cursor = s3_hide_hw_text_cursor;
			dpy->undo_cursor = NULL;

		} else 
		{

			dpy->show_cursor = s3_show_sw_text_cursor;
			dpy->hide_cursor = s3_hide_sw_text_cursor;
			dpy->undo_cursor = s3_undo_sw_text_cursor;
		}

		/* Provide software cursor too in case DAC didn't already */
		if (!dpy->show_pointer) 
		{

			dpy->show_pointer = s3_show_sw_text_pointer;
			dpy->hide_pointer = s3_hide_sw_text_pointer;
			dpy->undo_pointer = s3_undo_sw_text_pointer;
		}

		dpy->fb.size.x = mode->request.virt.x / mode->request.text.x;
		dpy->fb.size.y = mode->request.virt.y / mode->request.text.y;
		dpy->fb.bpp    = MAP_16_BPP;
		dpy->fb.pbuf   = MEM_LAW;
		dpy->fb.zbuf   = NULL;
		dpy->fb.ops    = &s3_text16;

		private->fb_size = dpy->fb.size.x * dpy->fb.size.y;

	} else 
	{

		dpy->fb.ops = NULL;
		dpy->set_font = NULL;
		dpy->set_split = s3_set_split;
		dpy->set_origin = s3_set_origin;
		dpy->set_frame = NULL;

		/* FIXME: clean this stuff up and move it to its own file 
		*/
		
		/* Reset the accelerator engine 
		*/
#if 0
		outw(((mode->request.graphtype == GT_4BIT)
			? EC01_4BPP : EC01_ENH_ENB) |
			EC01_LAW_ENB | EC01_MMIO_ENB, ECR_ADVFUNC_CNTL);
		outw(EC00_RESET, ECR_SUBSYS_CNTL);
		inw_p(ECR_SUBSYS_STAT);	/* wait a bit */
		outw(EC00_ENABLE, ECR_SUBSYS_CNTL);
		inw_p(ECR_SUBSYS_STAT); /* wait a bit */
		outw(MF_MULT_MISC, ECR_MULTIFUNC); /* 16 bit I/O registers */

		/* Now set some basic accelerator registers */
		wait_fifo(dpy, 6);
		outw(SRC_FOREGROUND | MIX_SRC, ECR_FOREGR_MIX);
		outw(SRC_BACKGROUND | MIX_SRC, ECR_BACKGR_MIX);/* direct color*/
		outw(MF_T_CLIP | 0, ECR_MULTIFUNC);	/* clip virtual area  */
		outw(MF_L_CLIP | 0, ECR_MULTIFUNC);
		outw(MF_R_CLIP | (mode->request.virt.x - 1), ECR_MULTIFUNC);
		outw(MF_B_CLIP | (mode->request.virt.y - 1), ECR_MULTIFUNC);
		wait_fifo(dpy, 5);
		outw(~0, ECR_WRITE_MASK);	/* Enable all planes */
		outw(~0, ECR_WRITE_MASK);	/* Enable all planes */
		outw(MF_PIX_CONTROL | MFA_SRC_FOREGR_MIX, ECR_MULTIFUNC);
		/* !!! reset other registers too. */
#endif
	}

	NOTICE1("mode set");
}

int kgim_chipset_command(struct kgi_graphic *graph, ggi_uint cmd, ggi_uint arg)
{
	return -E(CHIPSET, UNKNOWN);
}
