/* --------------------------------------------------------------------------
**      S3 86c928 chipset driver
**  --------------------------------------------------------------------------
**
**	Copyright (C)	1995-1997	Steffen Seeger
**			1997		Jochen Hoenicke
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2, or (at your option)
**	any later version.
** 
**	You should have received a copy of the GNU General Public License
**	along with this program; see the file COPYING.  If not, write to
**	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
** 
**  ---------------------------------------------------------------------------
** 
**	This code was derived from the following source of information:
**	[1]	S3 928 Data Book, (c) July 1993 S3 Incorporated
*/

#define	HARDWARE_TEXT_CURSOR	0	/* use software/hardware cursor */

#include <ggi/maintainers.h>
#define	MAINTAINER	Jochen_Hoenicke
#define	DRIVER_NAME	"S3 86c928 chipset driver"
#define	DRIVER_REV	"$Revision: 1.5 $"

#define	DEBUG_LEVEL		0
#define	HARDCODED_DEFAULTS	FALSE


#include <ggi/system.h>
#include <ggi/module.h>

#define	S3_86C928
#include "S3/928.h"
#define	IO_VGA_BASE		0x03C0
#define	IO_VGA_SIZE		0x20
#define	IO_ACCEL_BASE		0x02E8
#define	IO_ACCEL_SIZE		4

#define	TEXT_MEMORY_SIZE	(64 KB)

/*	We keep the driver specific state per chipset handled in a
**	chipset_private struct. As s3_mmio.inc and s3_text16.inc rely on it,
**	it must be declared early.
*/
struct chipset_private 
{
	ggi_uint		frame;
	ggi_uint		split;
	ggi_uint		origin, org_xdot, org_ydot, org_offs;
	ggi_uint		fb_size;

	uint32		ptr_old, cur_old;
	uint16		ptr_and, cur_and;
	uint16		ptr_xor, cur_xor;

	/*	exported
	*/
	struct kgi_mmio_region	graphfb;	/* framebuffer	*/
	struct kgi_mmio_region	textfb;
	struct kgi_mmio_region	pdt;	/* pixel data transfer */

	/*	administrative
	*/
	struct io_region	io_vga_region;
	struct io_region	io_accel_region;
	struct mem_region	mem_accel_region;
	struct mem_region	mem_law_region;

	struct kgi_chipset	chipset_info;
};
#define	GET_PRIVATE \
	struct chipset_private *private = KGIM_PRIVATE(dpy, chipset)

#define	IO_VGA			(private->io_vga_region.baseptr-IO_VGA_BASE)
#define	IO_ACCEL		(private->io_accel_region.baseptr-IO_ACCEL_BASE)
#define	MEM_ACCEL		(private->mem_accel_region.baseptr)
#define	MEM_LAW			(private->mem_law_region.baseptr)
#define	MEM_VGA			MEM_NULL

#define	IO_VGA_REGION		(private->io_vga_region)
#define	IO_ACCEL_REGION		(private->io_accel_region)
#define	MEM_ACCEL_REGION	(private->mem_accel_region)
#define	MEM_LAW_REGION		(private->mem_law_region)

#define	CHIPSET_INFO		(private->chipset_info)


#define	NoSEQRegs	0x05
#define	NoCRTRegs	0x80
#define	NoGRCRegs	0x09
#define	NoATCRegs	0x15
#include "s3_mmio.inc"
#include "s3_text16.inc"

static const struct kgi_chipset s3_928_info =
{
	"S3 Incorporated",		/* manufacturer         */
	"86C928 ",			/* model (detected)     */
	GM_TEXT16 |/* GM_4BIT |	*/	/* supported modes      */
	GM_8BIT | GM_15BIT |
	GM_16BIT | GM_32BIT,
	0,
	{2048, 2048},			/* maximum resolution   */

	0,				/* RAM size (detected)  */
	{0, 110000000}			/* DCLK limits            */
};

/*	This table holds the default bits for text & graphics mode. 
**	A copy of it will be filled in completely during mode_init().
*/
static struct statetable default_state =
{
	0x23, 0x00,
	{
		0x03, 0x00, 0x0F, 0x20, 0x02	/* SEQ */
	},
	{
		0x00, 0x00, 0x00, 0x80, 0x00, 0x00, 0x00, 0x10,	/* CRT */
		0x00, 0x40, 0x00, 0x1F, 0x00, 0x00, 0xFF, 0xFF,
		0x00, 0x00, 0x00, 0x00, 0x1F, 0x00, 0x00, 0xE3,	/* 0x1x */
		0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x2x */
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x01, 0x00, 0x02, 0x10, 0x00, 0x00, 0x00,	/* 0x3x */
		0x48, 0xA5, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x4x */
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x38, 0x00, 0x00, 0x00,	/* 0x5x */
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x6x */
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x7x */
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
	},
	{
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* GRC */
		0xFF
	},
	{
		0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,	/* ATC */
		0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,
		0x84, 0x00, 0x0F, 0x00, 0x00
	}
};

/*
**	ressources used by the S3 80C928
*/
static const struct io_region s3_928_io_vga =
{
	IO_NULL,
	IO_VGA_BASE,
	IO_VGA_SIZE,
	0xFFFF,
	"S3 86c928 chipset"
};

static const struct io_region s3_928_io_accel =
{
	IO_NULL,
	IO_ACCEL_BASE,
	IO_ACCEL_SIZE,
	0x03FF,
	"S3 86c928 accelerator"
};

static const struct mem_region s3_928_mem_accel =
{
	MEM_NULL,
	0x000A0000,
	64 KB,
	-1,
	"S3 86c928 text buffer/MMIO accelerator"
};

static const struct mem_region s3_928_mem_law =
{
	MEM_NULL,
	0x00000000,
	0,
	-1,
	"S3 86c928 frame buffer"
};

/*
**	mmio regions.
*/
static const struct kgi_mmio_region s3_928_pdt =
{
	NULL,
	"S3 86c928 pixel data transfer",
	0,
	MMIO_RDWR | MMIO_SHARED,

	32 KB,		/* total size		*/
	32 KB,		/* physical size	*/
	0xA0000,	/* physical start	*/
	0,		/* offset		*/

	NULL,		/* set_offset		*/
	NULL		/* private		*/
};

/*	normal graphics frame buffer.
*/
static const struct kgi_mmio_region s3_928_graphfb =
{
	NULL,
	"S3 86c928 graphic frame buffer",
	MMIO_FRAMEBUFFER,
	MMIO_RDWR | MMIO_SHARED,

	0,		/* total size		*/
	0,		/* physiscal size	*/
	0,		/* physical start	*/
	0,		/* offset		*/

	NULL,		/* set_offset		*/
	NULL		/* private		*/
};

/*	normal text frame buffer
*/
static void s3_set_text_offset(struct kgi_mmio_region *r, ggi_uint offset) 
{
	#define	dpy	r->private

	ASSERT(dpy);

	/*	first 64KB is text16 frame buffer, second 64 KB is font buffer
	*/
	if (offset < 64 KB) {

		/*	setup for text buffer access
		*/
		SEQ_out8(dpy, SR02_PLANE0 | SR02_PLANE1, 0x02);
		SEQ_out8(dpy, SR04_256K_ACCESS, 0x04);

		GRC_out8(dpy, 0, 0x04);
		GRC_out8(dpy, GR5_ODD_EVEN, 0x05);
		GRC_out8(dpy, GR6_MAP_A0_64 | GR6_CHAIN_OE, 0x06);

		CRT_out8(dpy, CRT_in8(dpy, 0x3A) & ~CR3A_HISPEEDTXT, 0x3A);

	} else {

		/*	setup for font buffer access
		*/
		SEQ_out8(dpy, SR02_PLANE2, 0x02);
		SEQ_out8(dpy, SR04_256K_ACCESS | SR04_NO_ODDEVEN, 0x04);

		GRC_out8(dpy, 2, 0x04);
		GRC_out8(dpy, GR5_WRITEMODE0, 0x05);
		GRC_out8(dpy, GR6_MAP_A0_64 | GR6_GRAPHMODE, 0x06);

		CRT_out8(dpy, CRT_in8(dpy, 0x3A) | CR3A_HISPEEDTXT, 0x3A);

	}
	#undef dpy
}

static const struct kgi_mmio_region s3_928_textfb =
{
	NULL,
	"S3 86c928 text frame buffer",
	MMIO_FRAMEBUFFER,
	MMIO_RDWR | MMIO_SHARED,

	128 KB,			/* total size			*/
	64 KB,			/* physiscal size of window	*/
	0x000A0000,		/* physical start of window	*/
	0,			/* offset			*/

	s3_set_text_offset,	/* set_offset			*/
	NULL			/* private			*/
};

/*
**	helper functions
*/
static void s3_examine_chipset(struct kgi_display *dpy)
{
	#define COPYDEFAULT(X)	(default_state.CRT[X] = CRT_in8(dpy, X))
	GET_PRIVATE;
	uint8	foo;

	ASSERT(private);

	/*	copy default values
	*/
	#if HARDCODED_DEFAULTS

		#error need default register values!
	#else
		COPYDEFAULT(CRT_CHIPID);
		COPYDEFAULT(CRT_CONFIG2);

		default_state.CRT[0x3A] |= CRT_in8(dpy,0x3A) & CR3A_REFRESHMASK;
		default_state.CRT[0x58] |= CRT_in8(dpy,0x58) & CR58_VL_T1_LATCH;

		#if DEBUG_LEVEL > 1

			#define	DEFAULT(x)	default_state.CRT[x] 
			DEBUG2("default values: %.2x  %.2x  %.2x ", \
				DEFAULT(CRT_CONFIG2), \
				DEFAULT(0x3A), \
				DEFAULT(0x58) \
			);

			#undef DEFAULT
		#endif
	#endif

	/*	detect amount, type of installed memory and bus type
	*/
	foo = COPYDEFAULT(0x36);
	switch (foo & CR36_MEMSIZE_MASK) {

	case CR36_05MB:
		CHIPSET_INFO.ramsize = 512 KB;
		MEM_LAW_REGION.size  = 1024 KB;
		default_state.CRT[0x58] |= CR58_LAW_1MB;
		break;

	case CR36_1MB:
		CHIPSET_INFO.ramsize = 1024 KB;
		MEM_LAW_REGION.size  = 1024 KB;
		default_state.CRT[0x58] |= CR58_LAW_1MB;
		break;

	case CR36_2MB:
		CHIPSET_INFO.ramsize = 2048 KB;
		MEM_LAW_REGION.size  = 2048 KB;
		default_state.CRT[0x58] |= CR58_LAW_2MB;
		break;

	case CR36_3MB:
		CHIPSET_INFO.ramsize = 3072 KB;
		MEM_LAW_REGION.size  = 4096 KB;
		default_state.CRT[0x58] |= CR58_LAW_4MB;
		break;

	case CR36_4MB:
		CHIPSET_INFO.ramsize = 4096 KB;
		MEM_LAW_REGION.size  = 4096 KB;
		default_state.CRT[0x58] |= CR58_LAW_4MB;
		break;

	default:
		CHIPSET_INFO.ramsize = 512 KB;
		MEM_LAW_REGION.size  = 1024 KB;
		default_state.CRT[0x58] |= CR58_LAW_1MB;
		NOTICE("unknown VRAM size, assume 512 KB");
	}
	DEBUG1("%i KB frame buffer RAM detected.", CHIPSET_INFO.ramsize/1024);

	#if DEBUG_LEVEL > 0

		switch (foo & CR36_BUSMASK) {

		case CR36_EISA:
			NOTICE2("EISA bus\n");
			break;

		case CR36_VESALOCAL:
			NOTICE2("Vesa Local bus\n");
			break;

		case CR36_PCI:
			NOTICE2("PCI bus\n");
			break;

		case CR36_ISA:
			NOTICE2("ISA bus\n");
			break;
		}
	#endif

	/*	setup MMIO region info
	*/
	private->graphfb = s3_928_graphfb;
	private->graphfb.size = \
	private->graphfb.psize = CHIPSET_INFO.ramsize;
	private->graphfb.next = &(private->pdt);
	private->graphfb.private = dpy;

	private->pdt = s3_928_pdt;
	private->pdt.private = dpy;

	private->textfb = s3_928_textfb;
	private->textfb.private = dpy;

	/*
	**	set miscellaneous stuff
	*/
	private->ptr_and = 0x00FF;
	private->ptr_xor = 0x6F00;
	private->cur_and = 0x00FF;
	private->cur_xor = 0x7F00;

	#undef COPYDEFAULT
}

/*
**	KGI module interface
*/

int kgim_chipset_init(struct kgi_card *card, struct kgi_display *dpy)
{
	ggi_uint	foo;
	ggi_sint	err;
	uint16	cmd = 0;
	mem_addr base0 = KGIM_OPTIONS(dpy, misc)->law_base;
	struct chipset_private *private;
	struct kgim_options_pci *pci = KGIM_OPTIONS(dpy, pci);
	struct kgim_options_misc *misc = KGIM_OPTIONS(dpy, misc);

	NOTICE(DRIVER_NAME " rev " DRIVER_REV);

	DEBUG2("testing for S3 86c928 (dev = %.8x)", pci->dev);

	/*	PCI verification is easiest. However, we may have to enable it.
	*/
	if (pci->dev != PCICFG_NULL) {

		if ((pcicfg_in16(pci->dev + PCI_VENDOR_ID) != PCI_VENDOR_ID_S3)
			|| (pcicfg_in16(pci->dev + PCI_DEVICE_ID) !=
			PCI_DEVICE_ID_S3_928)) {

			NOTICE2("PCI device is no S3 86c928.");
			return -E(CHIPSET, UNKNOWN);
		}

		/*	enable PCI device
		*/
		cmd = pcicfg_in16(pci->dev + PCI_COMMAND);
		pcicfg_out16(cmd | PCI_COMMAND_IO | PCI_COMMAND_MEMORY,
			pci->dev + PCI_COMMAND);
	}

	/*	allocate and initialize private data.
	*/
	if (! (private = kgi_malloc_small(sizeof(struct chipset_private),
		GFP_KERNEL))) {

		NOTICE("  no memory for chipset private data");
		return -ENOMEM;
	}
	memset(private, 0, sizeof(*private));
	IO_VGA_REGION      = s3_928_io_vga;
	IO_ACCEL_REGION    = s3_928_io_accel;
	MEM_ACCEL_REGION   = s3_928_mem_accel;
	MEM_LAW_REGION     = s3_928_mem_law;
	CHIPSET_INFO       = s3_928_info;
	KGIM_PRIVATE(dpy, chipset) = private;

	/*	make sure noone else is serving this card.
	*/
	if (io_check_region(& IO_VGA_REGION) ||
		io_check_region(& IO_ACCEL_REGION)) {

		ERROR("io_accel or io_vga ressource busy.");
		kgi_free_small(private);
		KGIM_PRIVATE(dpy, chipset) = NULL;
		return -E(CHIPSET, UNKNOWN);
	}
	io_claim_region(&IO_VGA_REGION);
	io_claim_region(&IO_ACCEL_REGION);

	/* NOTE	we assume the chip is woken up already. If not, add wakeup!
	*/
	s3_unlock(dpy);

	foo = CRT_in8(dpy, CRT_CHIPID); 
	switch (foo & CR30_CHIPID) {

	case CR30_928_P_ID:
		CHIPSET_INFO.model[6] = 'P';
		/* fall through */
	case CR30_928_ID:
		foo &= CR30_REVISION;

		s3_examine_chipset(dpy);

		/*	determine frame buffer window address and claim region
		*/
		base0 = (CRT_in8(dpy, CRT_LAW_POS_H) << 24) | 
			(CRT_in8(dpy, CRT_LAW_POS_L) << 16);
		MEM_LAW_REGION.base = (misc->law_base ? misc->law_base:base0) &
			~(MEM_LAW_REGION.size - 1);

		if (! MEM_LAW_REGION.base) {

			ERROR("invalid law_base");
			err = -EINVAL;
			break;
		}

		private->graphfb.pstart = MEM_LAW_REGION.base;

		if (mem_check_region(& MEM_LAW_REGION) ||
			mem_check_region(& MEM_ACCEL_REGION)) {

			ERROR("mem_law or mem_accel region busy");
			err = -EBUSY;
			break;
		}
		mem_claim_region(& MEM_LAW_REGION);
		mem_claim_region(& MEM_ACCEL_REGION);

		default_state.CRT[0x59] = (MEM_LAW_REGION.base >> 24) & 0xFF;
		default_state.CRT[0x5A] = (MEM_LAW_REGION.base >> 16) & 0xFF;

		/*	enable LAW and accelerator MMIO
		*/
		CRT_out8(dpy, default_state.CRT[0x58], 0x58);
		CRT_out8(dpy, default_state.CRT[0x59], 0x59);
		CRT_out8(dpy, default_state.CRT[0x5A], 0x5A);

		/*	select CR42 to driver the ClockSelect Lines
		*/
		foo = MISC_in8(dpy);
		if ((foo & MISC_CLOCK_MASK) != MISC_EXT_CLOCK) {

			CRT_out8(dpy, (foo & MISC_CLOCK_MASK) >> 2, 0x42);
			MISC_out8(dpy, (foo & ~MISC_CLOCK_MASK)|MISC_EXT_CLOCK);
		}

		err = EOK;

		break;

	default:
		err = -E(CHIPSET, UNKNOWN);
	}

	if (err) {

		s3_lock(dpy);

		if (pci->dev != PCICFG_NULL) {

		#warning remove this!
		/*	pcicfg_out16(cmd, pci->dev + PCI_COMMAND);
		*/
		}

		io_free_region(& IO_VGA_REGION);
		io_free_region(& IO_ACCEL_REGION);
		kgi_free_small(private);
		KGIM_PRIVATE(dpy, chipset) = NULL;
		NOTICE("no S3 86c928 detected");
		return err;
	}

	DEBUG1("%s rev 0x%x detected.\n", CHIPSET_INFO.model, foo);

	dpy->card->chipset = & CHIPSET_INFO;
	return EOK;
}


void kgim_chipset_done(struct kgi_display *dpy)
{
	GET_PRIVATE;

	ASSERT(private);

	s3_lock(dpy);

	mem_free_region(& MEM_ACCEL_REGION);
	mem_free_region(& MEM_LAW_REGION);
	io_free_region(& IO_ACCEL_REGION);
	io_free_region(& IO_VGA_REGION);
	kgi_free_small(private);

	NOTICE(DRIVER_NAME " removed.");
}


static void adjust_timing(struct kgi_mode *mode)
{
	register int foo = (mode->request.graphtype == GT_TEXT16)
		? mode->request.text.x : 8;

	#define ADJUST(X) mode->x.X -= (mode->x.X % foo)

	ADJUST(width);
	ADJUST(blankstart);
	ADJUST(syncstart);
	ADJUST(syncend);
	ADJUST(blankend);
	if (mode->x.total % foo) {

		mode->x.total += foo - (mode->x.total % foo);
	}

	if (((mode->x.syncend / foo) & 0x3F) ==
		((mode->x.syncstart / foo) & 0x3F)) {

		mode->x.syncend -= foo;
	}

	if (((mode->x.blankend / foo) & 0x7F) ==
		((mode->x.blankstart / foo) & 0x7F)) {

		mode->x.blankend -= foo;
	}

	if ((mode->y.syncend & 0x0F) == (mode->y.syncstart & 0x0F)) {

		mode->y.syncend--;
	}

	if ((mode->y.blankend & 0xFF) == (mode->y.blankstart & 0xFF)) {

		mode->y.blankend--;
	}

	#undef ADJUST
}

int kgim_chipset_check_mode(struct kgi_display *dpy, struct kgi_mode *mode,
	enum kgi_tm_cmd cmd)
{
	#define XVIRTUAL (mode->request.virt.x)
	#define YVIRTUAL (mode->request.virt.y)
	GET_PRIVATE;
	int Bpp = 1;
	int textx = mode->request.text.x;
	int texty = mode->request.text.y;

	ASSERT(private);

	if (mode->flags & FL_MON_INTERLACE) {

		NOTICE1("interlaced modes not supported yet.");
		return -E(CHIPSET, UNKNOWN);
	}

	switch (mode->request.graphtype) {

	case GT_32BIT:
		Bpp += 4;	       /* (4+2+1+1)/2 = 4 Bpp  */
	case GT_15BIT:
	case GT_16BIT:
		Bpp += 2;	       /* (2+1+1)/2 =   2 Bpp  */
	case GT_8BIT:
		Bpp += 1;	       /* (1+1)/2 =   1 Bpp    */
	case GT_4BIT:		       /* 1/2 = 0.5 Bpp        */

	case GT_TEXT16:
		break;
	default:
		DEBUG1("graphtype %i not supported", mode->request.graphtype);
		return -E(CHIPSET, UNKNOWN);
	}

	switch (cmd) {

	case CMD_PROPOSE:

		if (mode->request.graphtype == GT_TEXT16) {

			mode->magnify.x = mode->magnify.y = 1;

			if (textx < 7 || 9 < textx) {

				DEBUG1("invalid char width (%i)", textx);
				return -EINVAL;
			}
			if (!(XVIRTUAL || YVIRTUAL)) {

				NOTICE1("need x or y virtual!\n");
				return -EINVAL;
			}
			if (!XVIRTUAL) {

				XVIRTUAL = TEXT_MEMORY_SIZE*textx /
					(2 * YVIRTUAL / texty);
				DEBUG1("setting XVIRTUAL = %i", XVIRTUAL);
			}
			if (!YVIRTUAL) {

				YVIRTUAL = TEXT_MEMORY_SIZE*texty /
					(2 * XVIRTUAL / textx);
				DEBUG1("setting yvirtual = %i", YVIRTUAL);
			}
			if ((2*(XVIRTUAL/textx)*(YVIRTUAL/texty)) > 64 KB) {

				NOTICE1("not enough text memory");
				return -ENOMEM;
			}

			mode->request.visible.x -=
				(mode->request.visible.x % textx);
			mode->request.visible.y -=
				(mode->request.visible.y % texty);

		} else {	/* mode->request.graphtype == GT_TEXT */

			mode->magnify.x = mode->magnify.y = 1;

			if      (XVIRTUAL <= 640)	XVIRTUAL = 640;
			else if (XVIRTUAL <= 800)	XVIRTUAL = 800;
			else if (XVIRTUAL <= 1024)	XVIRTUAL = 1024;
			else if (XVIRTUAL <= 1280)	XVIRTUAL = 1280;
			else if (XVIRTUAL <= 2048)	XVIRTUAL = 2048;
			else {

				DEBUG1("%i hpixels are too many!", XVIRTUAL);
				return -E(CHIPSET, UNKNOWN);
			}

			if ((XVIRTUAL*YVIRTUAL*Bpp/2) > CHIPSET_INFO.ramsize) {

				DEBUG1("no memory for virtual %ix%i@%ibpp", \
					XVIRTUAL, YVIRTUAL, Bpp*4);
				return -ENOMEM;
			}
			mode->request.visible.x &= 0x7FFFFFF8;
			mode->request.visible.y &= 0x7FFFFFF8;

			switch (mode->request.graphtype) {

			case GT_TEXT16:
			case GT_4BIT:
			case GT_8BIT:
				mode->lclk.mul = mode->clk.mul = 1;
				mode->lclk.div = mode->clk.div = 1;
				break;

			case GT_15BIT:
			case GT_16BIT:
				mode->lclk.mul = mode->clk.mul = 2;
				mode->lclk.div = mode->clk.div = 1;
				break;

			case GT_24BIT:
			case GT_32BIT:
				mode->lclk.mul = mode->clk.mul = 3;
				mode->lclk.div = mode->clk.div = 1;
				break;

			default:
				INTERNAL_ERROR;
			}
		}

		mode->x.width = mode->request.visible.x;
		mode->y.width = mode->request.visible.y;

		return EOK;

	case CMD_LOWER:

		if (mode->dclk < CHIPSET_INFO.dclk.min) {

			DEBUG1("%i Hz DCLK is too low.", mode->dclk);
			return -E(CHIPSET, UNKNOWN);
		}

		mode->dclk = (mode->dclk > CHIPSET_INFO.dclk.max)
			? CHIPSET_INFO.dclk.max : mode->dclk;

		adjust_timing(mode);
		return EOK;

	case CMD_RAISE:

		if (mode->dclk > CHIPSET_INFO.dclk.max) {

			DEBUG1("%i Hz DCLK is too high.", mode->dclk);
			return -E(CHIPSET, UNKNOWN);
		}

		mode->dclk = (mode->dclk < CHIPSET_INFO.dclk.min)
			? CHIPSET_INFO.dclk.min : mode->dclk;

		adjust_timing(mode);
		return EOK;

	case CMD_CHECK:

		if (mode->request.graphtype == GT_TEXT16) {

			if ((mode->magnify.x != 1) || (mode->magnify.y != 1)	/* driver */
			    ||(textx < 7)	/* hardware */
			    ||(9 < textx)	/* hardware */
			    ||(texty < 1)	/* hardware */
			    ||(texty > 32)	/* hardware */
			    ||((2 * (XVIRTUAL / textx)
				* (YVIRTUAL / texty)) > TEXT_MEMORY_SIZE)
			    || (mode->dclk > CHIPSET_INFO.dclk.max)		/* hardware */
			    ||(mode->dclk < CHIPSET_INFO.dclk.min)) {	/* hardware */
				return -E(CHIPSET, UNKNOWN);
			}
			mode->flags |= FL_CHIP_COLORTEXT;
			mode->fb = MEM_ACCEL;
			mode->fb_size = TEXT_MEMORY_SIZE;
			mode->fsize = (2 * XVIRTUAL * YVIRTUAL) /
				(textx * texty);
			mode->io = &text16_color_io;
			mode->mem = &text16_color_mem;
			mode->mmio = &(private->textfb);

		} else {

			switch (XVIRTUAL) {
			case 640:
			case 800:
			case 1024:
			case 1280:
			case 2048:
				break;

			default:
				return -E(CHIPSET, UNKNOWN);
			}

			if (((XVIRTUAL * YVIRTUAL * Bpp / 2) > CHIPSET_INFO.ramsize)
			    || (mode->magnify.x != 1)	/* !!! temp */
			    || (mode->magnify.y != 1)	/* !!! temp */
			    || (mode->dclk < CHIPSET_INFO.dclk.min)
			    || (CHIPSET_INFO.dclk.max < mode->dclk)) {

				return -E(CHIPSET, UNKNOWN);
			}
			mode->fb = MEM_LAW;
			mode->fb_size = CHIPSET_INFO.ramsize;
			mode->fsize = mode->request.virt.y *
				mode->request.virt.y * Bpp / 2;
			mode->io = NULL;
			mode->mem = NULL;
			mode->mmio = &(private->graphfb);
		}

		#define TM(X) (((unsigned) CHIPSET_INFO.maxdots.x < mode->x.X)\
			|| ((unsigned) CHIPSET_INFO.maxdots.y < mode->y.X))

		if (TM(width) || TM(blankstart) || TM(syncstart) ||
			TM(syncend) || TM(blankend) || TM(total)) {

			return -E(CHIPSET, UNKNOWN);
		}
		#undef TM

		adjust_timing(mode);
		return EOK;

	default:
		INTERNALERROR;
		return -E(CHIPSET, UNKNOWN);
	}

	#undef XVIRTUAL
	#undef YVIRTUAL
}

void kgim_chipset_set_mode(struct kgi_display *dpy, struct kgi_mode *mode)
{
	GET_PRIVATE;
	IS_CRITICAL;
	struct statetable setup = default_state;
	register int foo, mul = mode->clk.mul;
	int DCLKdiv;
	
	DEBUG1("setting mode %i, %ix%i.%ix%i.%ix%i", \
		mode->request.graphtype,\
		mode->request.visible.x, mode->request.visible.y, \
		mode->request.virt.x, mode->request.virt.y, \
		mode->request.text.x, mode->request.text.y);

	/* Setup bits per pixel and logical screen width */

	switch (mode->request.graphtype) {

	case GT_TEXT16:
		foo = 2 * mode->request.virt.x / mode->request.text.x;
		foo /= 4;	/* due to the addressing mode used. */
		break;

	case GT_4BIT:
		setup.CRT[0x50] |= CR50_4BPP;
		foo = mode->request.virt.x >> 4;
		break;

	case GT_8BIT:
		setup.CRT[0x3A] |= CR3A_NO4BPPMODE;
		setup.CRT[0x50] |= CR50_8BPP;
		foo = mode->request.virt.x >> 3;
		break;

	case GT_15BIT:
	case GT_16BIT:
		setup.CRT[0x3A] |= CR3A_NO4BPPMODE;
		setup.CRT[0x50] |= CR50_16BPP;
		foo = mode->request.virt.x >> 2;
		break;

	case GT_32BIT:
		setup.CRT[0x3A] |= CR3A_NO4BPPMODE;
		setup.CRT[0x50] |= CR50_32BPP;
		foo = mode->request.virt.x >> 1;
		break;

	default:
		INTERNAL_ERROR;
		return;
	}

	setup.CRT[CRT_LOGICALWIDTH] |= foo & 0xFF;
	setup.CRT[0x51] |= (foo & 0x100) ? CR51_LOGICALWIDTH_B8 : 0;
	setup.CRT[0x51] |= (foo & 0x200) ? CR51_LOGICALWIDTH_B9 : 0;

	if (mode->request.graphtype == GT_TEXT16) {

		setup.CRT[0x14] |= CR14_UNDERLINE_MASK;

		setup.CRT[0x17] = CR17_CGA_ADDR_WRAP | CR17_ENABLE_SYNC |
			CR17_CGA_BANKING | CR17_HGC_BANKING; 

		setup.CRT[0x31] |= CR31_HISPEEDTXT;

		setup.SEQ[0x02] = SR02_PLANE0 | SR02_PLANE1;

		setup.GRC[0x05] |= GR5_ODD_EVEN;
		setup.GRC[0x06] |= GR6_MAP_A0_64 | GR6_CHAIN_OE;

		/*	??? can we leave this enabled?  
		**	Miro20SV, V7 Mercury P-64 can.  
		**	setup.CRT[0x34] &= ~CR34_ENB_FIFO;  
		*/

		switch ((DCLKdiv = mode->request.text.x)) {

		case 7:
			setup.CRT[0x32] |= CR32_7DOT_CHCLK;
			break;
		case 8:
			setup.SEQ[0x01] |= SR01_8DOT_CHCLK;
			break;
		case 9:
			break;
		default:
			INTERNALERROR;
			return;
		}

		setup.CRT[0x09] |= mode->request.text.y - 1;
		setup.ATC[0x13] = (mode->request.text.x == 9) ? 0x08 : 0x00;

	} else {	/* mode->request.graphtype == GT_TEXT16 */

		DCLKdiv = 8;

		for (foo = 0; foo < 16; foo++)
			setup.ATC[foo] = foo;

		setup.SEQ[0x01] |= SR01_8DOT_CHCLK;
		setup.SEQ[0x02] |= SR02_PLANE_MASK;
			/* ??? enable all planes-neccessary ? */

		setup.ATC[0x10] |= AR10_GRAPHICS;
		setup.GRC[0x05] |= GR5_8BIT_COLOR;/* ??? neccessary ? */
		
		setup.GRC[0x06] |= GR6_GRAPHMODE |
			GR6_MAP_A0_64 /* ??? neccessary ? */ ;

		setup.CRT[0x31] |= CR31_ENHANCED;
		setup.CRT[0x33] |= CR33_NOBORDER;
		setup.CRT[0x53] |= MEM_ACCEL ? CR53_ENB_MMIO : 0;
		setup.CRT[0x58] |= CR58_LAW_ENABLE;

		setup.CRT[0x09] |=     /* !!! TM->ymagnify-1 &
				        * CR09_CHARHEIGHT_MASK; */ 0;
		setup.CRT[0x09] |= (mode->flags & FL_CHIP_DOUBLESCAN)
			? CR09_DOUBLESCAN : 0;

		switch (mode->request.virt.x) {

		case 640:
			setup.CRT[0x50] |= CR50_640;
			break;

		case 800:
			setup.CRT[0x50] |= CR50_800;
			break;

		case 1024:
			setup.CRT[0x50] |= CR50_1024;
			break;

		case 1280:
			setup.CRT[0x50] |= CR50_1280;
			break;

		case 2048:
			setup.CRT[0x50] |= CR50_2048;
			setup.CRT[0x31] |= CR31_2PAGE_SCRN;
			break;

		default:
			INTERNAL_ERROR;
			return;
		}
	}

	/*
	**	setup horizontal timing
	*/
	if ((mode->request.graphtype == GT_16BIT) ||
		(mode->request.graphtype == GT_15BIT)) {

		mul = 1;
		setup.CRT[0x43] |= CR43_DOUBLE_HTIMING;
	}

	/*	horizontal display end
	*/
	foo = (mode->x.width * mul / DCLKdiv) - 1;
	setup.CRT[0x01] |= foo & 0xFF;
	setup.CRT[0x5D] |= (foo & 0x100) ? CR5D_HDISPLAYEND_B8 : 0;

	/*	horizontal blank start
	*/
	foo = (mode->x.blankstart * mul / DCLKdiv);
	setup.CRT[0x02] |= foo & 0xFF;
	setup.CRT[0x5D] |= (foo & 0x100) ? CR5D_HBLANKSTART_B8 : 0;

	/*	horizontal sync start
	*/
	foo = (mode->x.syncstart * mul / DCLKdiv);
	setup.CRT[0x04] |= foo & 0xFF;
	setup.CRT[0x5D] |= (foo & 0x100) ? CR5D_HSYNCSTART_B8 : 0;

	/*	horizontal sync end
	*/
	foo = (mode->x.syncend * mul / DCLKdiv);
	setup.CRT[0x05] |= foo & CR05_HSE_MASK;

	/*	horizontal blank end
	*/
	foo = (mode->x.blankend * mul / DCLKdiv) - 1;
	setup.CRT[0x03] |= foo & CR03_HBE_MASK;
	setup.CRT[0x05] |= (foo & 0x20) ? CR05_HBLANKEND_B5 : 0;

	/*	horizontal total
	*/
	foo = (mode->x.total * mul / DCLKdiv) - 5;
	setup.CRT[0x00] |= foo & 0xFF;
	setup.CRT[0x5D] |= (foo & 0x100) ? CR5D_HTOTAL_B8 : 0;

	/*
	**	setup vertical timing
	*/

	/*	vertical display end
	*/
	foo = mode->y.width - 1;
	setup.CRT[0x12] |= foo & 0xFF;
	setup.CRT[0x07] |= (foo & 0x100) ? CR07_VDISPLAYEND_B8 : 0;
	setup.CRT[0x07] |= (foo & 0x200) ? CR07_VDISPLAYEND_B9 : 0;
	setup.CRT[0x5E] |= (foo & 0x400) ? CR5E_VDISPLAYEND_B10 : 0;

	/*	vertical blank start
	*/
	foo = mode->y.blankstart - 1;
	setup.CRT[0x15] |= foo & 0xFF;
	setup.CRT[0x07] |= (foo & 0x100) ? CR07_VBLANKSTART_B8 : 0;
	setup.CRT[0x09] |= (foo & 0x200) ? CR09_VBLANKSTART_B9 : 0;
	setup.CRT[0x5E] |= (foo & 0x400) ? CR5E_VBLANKSTART_B10 : 0;

	/*	vertical sync start
	*/
	foo = mode->y.syncstart;
	setup.CRT[0x10] |= foo & 0xFF;
	setup.CRT[0x07] |= (foo & 0x100) ? CR07_VSYNCSTART_B8 : 0;
	setup.CRT[0x07] |= (foo & 0x200) ? CR07_VSYNCSTART_B9 : 0;
	setup.CRT[0x5E] |= (foo & 0x400) ? CR5E_VSYNCSTART_B10 : 0;

	/*	vertical sync end
	*/
	foo = mode->y.syncend;
	setup.CRT[0x11] |= foo & CR11_VSYNCEND_MASK;

	/*	end vertical blank
	*/
	foo = mode->y.blankend - 1;
	setup.CRT[0x16] |= foo & 0xFF;

	/*	vertical total
	*/
	foo = mode->y.total - 2;
	setup.CRT[0x06] |= foo & 0xFF;
	setup.CRT[0x07] |= (foo & 0x100) ? CR07_VTOTAL_B8 : 0;
	setup.CRT[0x07] |= (foo & 0x200) ? CR07_VTOTAL_B9 : 0;
	setup.CRT[0x5E] |= (foo & 0x400) ? CR5E_VTOTAL_B10 : 0;


	/* !!!	no interlaced modes yet.
	** 	according to vgadoc4 this should be half horizontal total
	*/
	setup.CRT[CRT_INTERLACESTART] |= mode->x.total >> 1;

	if (mode->flags & FL_MON_INTERLACE) {

		NOTICE1("no interlaced modes yet.");
		setup.CRT[0x42] |= CR42_INTERLACED;
		INTERNAL_ERROR;
		return;
	}


	setup.MISC |= (mode->x.polarity > 0) ? 0 : MISC_NEG_HSYNC;
	setup.MISC |= (mode->y.polarity > 0) ? 0 : MISC_NEG_VSYNC;


	/*	Setup the FIFO control parameters.
	*/
	foo = (mode->x.total + mode->x.syncstart) / 2 / DCLKdiv;
	setup.CRT[0x3B] |= foo & 0xFF;
	setup.CRT[0x5D] |= (foo & 0x100) ? CR5D_FIFOSTART_B8 : 0;


	/*	Ok. We'll do the programming now.
	*/
	ENTER_CRITICAL;

		/*	wait for accelerator to finish and disable it.
		*/ 
		if (dpy->mode.request.graphtype != GT_TEXT16) {

			foo = 10000;
			while ((ACCEL_in16(dpy, ECR_GP_STATUS) & 0xF2FF) &&
				foo) {

				foo--;
			}
			AC_out16(dpy, 0, AC_ADVFUNC_CNTL);
		}

		/*	switch sync signals off
		*/
		CRT_out8(dpy, CRT_in8(dpy, 0x56) | CR56_SYNC_OFF, 0x56);

		/*	prepare subsystems
		*/
		s3_unlock(dpy);
		SEQ_out8(dpy, SEQ_in8(dpy, 0x01) | SR01_DISPLAY_OFF, 0x01);

		kgim_monitor_set_mode(dpy, mode);
		kgim_ramdac_set_mode(dpy, mode);
		kgim_clock_set_mode(dpy, mode);
		s3_unlock(dpy);
		restore_state(dpy, &setup);

	LEAVE_CRITICAL;

	memcpy(&dpy->mode, mode, sizeof (dpy->mode));

	NOTICE2("setting up display...");

	private->split = -1;
	private->org_xdot = private->org_ydot = private->org_offs = -1;

	if (mode->request.graphtype == GT_TEXT16) {

		dpy->set_font = s3_text_set_font;
		dpy->set_split = s3_set_split;
		dpy->set_frame = s3_set_frame;
		dpy->set_origin = s3_text_set_origin;

		/*	provide cursor in text mode
		*/
		if (HARDWARE_TEXT_CURSOR) {

			dpy->show_cursor = s3_show_hw_text_cursor;
			dpy->hide_cursor = s3_hide_hw_text_cursor;
			dpy->undo_cursor = NULL;

		} else {

			dpy->show_cursor = s3_show_sw_text_cursor;
			dpy->hide_cursor = s3_hide_sw_text_cursor;
			dpy->undo_cursor = s3_undo_sw_text_cursor;
		}

		/*	provide pointer too if DAC didn't already.
		*/
		if (!dpy->show_pointer) {

			dpy->show_pointer = s3_show_sw_text_pointer;
			dpy->hide_pointer = s3_hide_sw_text_pointer;
			dpy->undo_pointer = s3_undo_sw_text_pointer;
		}

		dpy->fb.size.x = mode->request.virt.x / mode->request.text.x;
		dpy->fb.size.y = mode->request.virt.y / mode->request.text.y;
		dpy->fb.bpp  = MAP_16_BPP;
		dpy->fb.pbuf = MEM_ACCEL;
		dpy->fb.zbuf = NULL;
		dpy->fb.ops  = &text16_color_io;

		private->fb_size = dpy->fb.size.x * dpy->fb.size.y;

	} else {

		dpy->fb.size.x	= mode->request.virt.x;
		dpy->fb.size.y	= mode->request.virt.y;
		#warning set dpy->fb.bpp!
		dpy->fb.pbuf	= MEM_LAW;
		dpy->fb.zbuf	= NULL;
		dpy->fb.ops	= NULL;

		private->fb_size = dpy->fb.size.x * dpy->fb.size.y;
		
		dpy->show_pointer = NULL;
		dpy->hide_pointer = NULL;
		dpy->undo_pointer = NULL;

		dpy->show_cursor = NULL;
		dpy->hide_cursor = NULL;
		dpy->undo_cursor = NULL;

		dpy->set_font = NULL;
		dpy->set_split = s3_set_split;
		dpy->set_frame = s3_set_frame;
		dpy->set_origin = s3_set_origin;


		/* !!! temporary, this will be moved to an extra source file.
		*/

		/* Reset the Accelerator Engine */
		AC_out16(dpy, ((mode->request.graphtype == GT_4BIT) 
			? EC01_4BPP
			: EC01_ENH_ENB) | EC01_LAW_ENB, AC_ADVFUNC_CNTL);

		AC_out16(dpy, EC00_RESET, AC_SUBSYS_CNTL);
		AC_out16(dpy, EC00_ENABLE, AC_SUBSYS_CNTL);

		ACCEL_out16(dpy, MF_MULT_MISC, ECR_MULTIFUNC);

		/* Now set some basic accelerator registers */
		wait_fifo(dpy, 6);
		ACCEL_out16(dpy, SRC_FOREGROUND | MIX_SRC, ECR_FOREGR_MIX);
		ACCEL_out16(dpy, SRC_BACKGROUND | MIX_SRC, ECR_BACKGR_MIX);
		/* direct color */

		ACCEL_out16(dpy, MF_T_CLIP | 0, ECR_MULTIFUNC);	/* clip virtual area  */
		ACCEL_out16(dpy, MF_L_CLIP | 0, ECR_MULTIFUNC);
		ACCEL_out16(dpy, MF_R_CLIP | (mode->request.virt.x - 1), ECR_MULTIFUNC);
		ACCEL_out16(dpy, MF_B_CLIP | (mode->request.virt.y - 1), ECR_MULTIFUNC);

		wait_fifo(dpy, 5);
		ACCEL_out16(dpy, ~0, ECR_WRITE_MASK);	/* Enable all planes */
		ACCEL_out16(dpy, ~0, ECR_WRITE_MASK);	/* Enable all planes */
		ACCEL_out16(dpy, MF_PIX_CONTROL | MFA_SRC_FOREGR_MIX,
			ECR_MULTIFUNC);

		/* !!! reset other registers too. */
		/* memset(fb_graphics, 0, fb_graph_len); */
	}

	TRACE3({

		dump_state(&setup);
		printk("\n");
	});
	NOTICE1("mode_set() done");
}

int kgim_chipset_command(struct kgi_graphic *graph, ggi_uint cmd, ggi_uint arg)
{
	return -E(CHIPSET, UNKNOWN);
}
