/* --------------------------------------------------------------------------
**	S3 I/O routines
** --------------------------------------------------------------------------
**
**	Copyright (C) 1995-1997 by Steffen Seeger
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2, or (at your option)
**	any later version.
**
**	You should have received a copy of the GNU General Public License
**	along with this program; see the file COPYING.  If not, write to
**	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
**
** ---------------------------------------------------------------------------
**	This code was derived from the following source of information:
**	[1]	S3 Vision964 Data Book, (c) 1994 S3 Incorporated
*/

#define	IO_OUT16VAL(v, r)	(((v) << 8) | (r))

#define	CRTI	CRTIc
#define	CRTD	CRTDc
#define	INS1r	INS1rc
#define	FCTRLw	FCTRLwc

static void MISC_out8(struct kgi_display *dpy, uint8 val)
{
	GET_PRIVATE;

	if (MEM_VGA) {

		mem_out8(val | MISC_COLOR_IO, MEM_VGA + MISCw);
		DEBUG3("MISC mem <- %.2x", val | MISC_COLOR_IO);

	} else {	

		io_out8(val | MISC_COLOR_IO, IO_VGA + MISCw);
		DEBUG3("MISC io <- %.2x", val | MISC_COLOR_IO);
	}
}

static uint8 MISC_in8(struct kgi_display *dpy)
{
	register GET_PRIVATE;

	if (MEM_VGA) {

		register uint8 val = mem_in8(MEM_VGA + MISCr);
		DEBUG3("MISC mem -> %.2x", val);
		return val;

	} else {

		register uint8 val = io_in8(IO_VGA + MISCr);
		DEBUG3("MISC io -> %.2x", val);
		return val;
	}
}

static void CRT_out8(struct kgi_display *dpy, uint8 val, uint8 reg)
{
	GET_PRIVATE;

	if (MEM_VGA) {

		DEBUG3("CRT mem %.2x <- %.2x", reg, val);
		mem_out8(reg, MEM_VGA + CRTI);
		mem_out8(val, MEM_VGA + CRTD);

	} else {

		DEBUG3("CRT io %.2x <- %.2x", reg, val);
		io_out16(IO_OUT16VAL(val, reg), IO_VGA + CRTI);
	}
}

static uint8 CRT_in8(struct kgi_display *dpy, uint8 reg)
{
	GET_PRIVATE;

	if (MEM_VGA) {

		uint8 val;

		mem_out8(reg, MEM_VGA + CRTI);
		val = mem_in8(MEM_VGA + CRTD);
		DEBUG3("CRT mem %.2x -> %.2x", reg, val);
		return val;

	} else {

		uint8 val;

		io_out8(reg, IO_VGA + CRTI);
		val = io_in8(IO_VGA + CRTD);
		DEBUG3("CRT io %.2x -> %.2x", reg, val);
		return val;
	}
}

static void ATC_out8(struct kgi_display *dpy, uint8 val, uint8 reg)
{
	GET_PRIVATE;

	if (MEM_VGA) {

		mem_in8(MEM_VGA + INS1r);	/* reset AFF	*/
		mem_out8(reg, MEM_VGA + ATCI);	/* set index	*/
		mem_out8(val, MEM_VGA + ATCDw);	/* write value	*/
		mem_in8(MEM_VGA + INS1r);	/* reset AFF	*/
		mem_out8(ATCI_ENABLEDISPLAY, MEM_VGA + ATCI);

	} else {

		io_in8(IO_VGA + INS1r);		/* reset AFF	*/
		io_out8(reg, IO_VGA + ATCI);	/* set index	*/
		io_out8(val, IO_VGA + ATCDw);	/* write value	*/
		io_in8(IO_VGA + INS1r);		/* reset AFF	*/
		io_out8(ATCI_ENABLEDISPLAY, IO_VGA + ATCI);
	}
}

static inline void SEQ_out8(struct kgi_display *dpy, uint8 val, uint8 reg)
{
	GET_PRIVATE;

	if (MEM_VGA) {

		mem_out8(reg, MEM_VGA + SEQI);
		mem_out8(val, MEM_VGA + SEQD);

	} else {

		io_out16(IO_OUT16VAL(val, reg), IO_VGA + SEQI);
	}
}

static inline uint8 SEQ_in8(struct kgi_display *dpy, uint8 reg)
{
	GET_PRIVATE;
	
	if (MEM_VGA) {

		mem_out8(reg, MEM_VGA + SEQI);
		return mem_in8(MEM_VGA + SEQD);

	} else {

		io_out8(reg, IO_VGA + SEQI);
		return io_in8(IO_VGA + SEQD);
	}
}

static inline void GRC_out8(struct kgi_display *dpy, uint8 val, uint8 reg)
{
	GET_PRIVATE;
	
	if (MEM_VGA) {

		mem_out8(reg, MEM_VGA + GRCI);
		mem_out8(val, MEM_VGA + GRCD);

	} else {

		io_out16(IO_OUT16VAL(val, reg), IO_VGA + GRCI);
	}
}

static inline uint16 GRC_in8(struct kgi_display *dpy, uint8 reg)
{
	GET_PRIVATE;

	if (MEM_VGA) {

		mem_out8(reg, MEM_VGA + GRCI);
		return mem_in8(MEM_VGA + GRCD);

	} else {

		io_out8(reg, IO_VGA + GRCI);
		return io_in8(IO_VGA + GRCD);
	}
}

static void s3_unlock(struct kgi_display *dpy)
{
	register GET_PRIVATE;

	if (MEM_VGA) {

		#if defined(S3_86C968)

			/*	MMIO only wake-up
			*/
			mem_out8(0x10, MEM_LAW + 0x1008508);
			mem_out8(0x01, MEM_LAW + 0x1008502);
			mem_out8(0x08, MEM_LAW + 0x1008508);

			NOTICE1("wakeup");

		#elif defined(S3_86C765) || defined(S3_86C325)

			mem_out8(0x01, MEM_LAW + 0x10083c3);

		#elif !(defined(S3_86C928))

			#warning check the code above!
		#endif

	} else {

		#if defined(S3_86C964) || defined(S3_86C968) || \
			defined(S3_86C928)

			/*	928 has I/O only, but may need explicit wakeup.
			*/
			#warning 0x102 not claimed!
			io_out8(0x10, IO_VGA + S3_SUBSYSTEM);
			io_out8(0x01, IO_VGA + S3_SETUP);
			io_out8(0x08, IO_VGA + S3_SUBSYSTEM);

		#elif defined(S3_86C765) || defined(S3_86C325)

			io_out8(0x01, IO_VGA + 0x3C3);
		#endif
	}

	MISC_out8(dpy, MISC_in8(dpy) | MISC_COLOR_IO);

	CRT_out8(dpy, CR38_MAGIC, 0x38);		/* unlock CR30-CR3F */
	CRT_out8(dpy, CR39_MAGIC, 0x39);		/* unlock CR40-CRFF */
	CRT_out8(dpy, CRT_in8(dpy, 0x33) & ~(CR33_UNLOCK_VDE | CR33_LOCK_DAC |
		CR33_LOCK_PALETTE), 0x33);

	#if defined(S3_86C968) || defined(S3_86C765) || defined(S3_86C325)

		SEQ_out8(dpy, SR08_MAGIC, 0x08);
	#endif

	/*	there are more locks, but this should work after reboot.
	*/
}

static void s3_lock(struct kgi_display *dpy)
{
	CRT_out8(dpy, 0, 0x38);			/* lock CR30-CR3F */
	CRT_out8(dpy, 0, 0x39);			/* lock CR40-CRFF */
}

/*
**	generalized DAC output functions
*/

#if defined(S3_86C928)

	#define	VRAM_DAC
	#undef	BUGGY_MMIO
#endif

#if defined(S3_86C968) || defined(S3_86C964)

	#define	VRAM_DAC
	#define	BUGGY_MMIO
#endif

#if defined(S3_86C765) || defined(S3_86C325)

	#define	INTERNAL_DAC
#endif

static uint16 dac_io[4]  = { 0x3C8, 0x3C9, 0x3C6, 0x3C7 };
#ifdef	BUGGY_MMIO

	static uint16 dac_mem[4] = { 0x3C8, 0x3C9, 0x3CA, 0x3CB };
#else
	static uint16 dac_mem[4] = { 0x3C8, 0x3C9, 0x3C6, 0x3C7 };
#endif

#ifdef VRAM_DAC

	#define	DAC_REGISTERS	0x10	/* RS[0:3] are possible	*/
#endif

#ifdef INTERNAL_DAC

	#define	DAC_REGISTERS	0x08	/* RS[0:2] are possible */
#endif


void DAC_out8(struct kgi_display *dpy, uint8 val, uint8 reg)
{
	GET_PRIVATE;

	ASSERT(reg < DAC_REGISTERS);

	if (MEM_VGA) {

		#if DAC_REGISTERS > 8

			mem_out8(0x55, MEM_VGA + CRTI);
			mem_out8((mem_in8(MEM_VGA + CRTD) & ~CR55_DAC_MASK) |
				(reg >> 2), MEM_VGA + CRTD);
		#endif
		mem_out8(val, MEM_VGA + dac_mem[reg & 0x03]);

		DEBUG3("DAC mem %.2x <- %.2x", reg, val);

	} else {

		#if DAC_REGISTERS > 8

			io_out8(0x55, IO_VGA + CRTI);
			io_out8((io_in8(IO_VGA + CRTD) & ~CR55_DAC_MASK) |
				(reg >> 2), IO_VGA + CRTD);
		#endif

		io_out8(val, IO_VGA + dac_io[reg & 0x03]);

		DEBUG3("DAC io %.2x <- %.2x", reg, val);
	}
}

uint8 DAC_in8(struct kgi_display *dpy, uint8 reg)
{
	#if !defined(S3_86C928)
		GET_PRIVATE;
	#endif
	register uint8 foo;
  
	ASSERT(reg < DAC_REGISTERS);

	if (MEM_VGA) {

		#if DAC_REGISTERS > 8

			mem_out8(0x55, MEM_VGA + CRTI);
			mem_out8((mem_in8(MEM_VGA + CRTD) & ~CR55_DAC_MASK) |
				(reg >> 2), MEM_VGA + CRTD);
			DEBUG2("CR52 = %.2x,", mem_in8(MEM_VGA + CRTD));
		#endif

		foo = mem_in8(MEM_VGA + dac_mem[reg & 0x03]);
		DEBUG3("DAC mem %.2x -> %.2x", reg, foo);
		return foo;

	} else {

		#if DAC_REGISTERS > 8

			io_out8(0x55, CRTI);
			io_out8((io_in8(CRTD) & ~CR55_DAC_MASK) | (reg >> 2),
				CRTD);
		#endif
	
		foo = io_in8(dac_io[reg & 0x03]);
		DEBUG3("DAC io %.2x -> %.2x", reg, foo);
		return foo;
	}
}

void DAC_ins8(struct kgi_display *dpy, uint8 reg, void *buf, ggi_uint count)
{
	GET_PRIVATE;

	ASSERT(reg < DAC_REGISTERS);

	if (MEM_VGA) {

		#if DAC_REGISTERS > 8

			mem_out8(0x55, MEM_VGA + CRTI);
			mem_out8((mem_in8(MEM_VGA + CRTD) & ~CR55_DAC_MASK) |
				(reg >> 2), MEM_VGA + CRTD);
		#endif

		mem_ins8(MEM_VGA + dac_mem[reg & 0x03], buf, count);

	} else {


		#if DAC_REGISTERS > 8

			io_out8(0x55, IO_VGA + CRTI);
			io_out8((io_in8(IO_VGA + CRTD) & ~CR55_DAC_MASK) |
				(reg >> 2), IO_VGA + CRTD);
		#endif

		io_ins8(IO_VGA + dac_mem[reg & 0x03], buf, count);
	}
}

void DAC_outs8(struct kgi_display *dpy, uint8 reg, void *buf, ggi_uint count)
{
	GET_PRIVATE;

	ASSERT(reg < DAC_REGISTERS);

	if (MEM_VGA) {

		#if DAC_REGISTERS > 8

			mem_out8(0x55, MEM_VGA + CRTI);
			mem_out8((mem_in8(MEM_VGA+CRTD) & ~CR55_DAC_MASK) |
				(reg >> 2), MEM_VGA + CRTD);
		#endif

		mem_outs8(MEM_VGA + dac_mem[reg & 0x03], buf, count);

	} else {

		#if DAC_REGISTERS > 8

			io_out8(0x55, IO_VGA + CRTI);
			io_out8((io_in8(IO_VGA + CRTD) & ~CR55_DAC_MASK) |
				(reg >> 2), IO_VGA + CRTD);
		#endif

		io_outs8(IO_VGA + dac_io[reg & 0x03], buf, count);
	}
}

/*
**	clock I/O functions
*/

#if defined(S3_86C964) || defined(S3_86C968) || defined(S3_86C928)


#define	CLK_DECODE	(CR42_DCLK_MASK+1)

void CLK_out8(struct kgi_display *dpy, uint8 value, uint8 reg)
{
	ASSERT(value < CLK_DECODE);

	CRT_out8(dpy, (CRT_in8(dpy, 0x42) & ~CR42_DCLK_MASK) | value, 0x42);
} 

uint8 CLK_in8(struct kgi_display *dpy, uint8 reg)
{
	return CRT_in8(dpy, 0x5C) & CR5C_CLOCK_MASK;
}

#endif

#if defined(S3_86C765) || defined(S3_86C325)

void CLK_out8(struct kgi_display *dpy, uint8 val, uint8 reg)
{
	if (reg < 7) {

		switch (reg) {

		case 0:
			val = (val & 3) << 2;
			MISC_out8(dpy, (MISC_in8(dpy) & ~MISC_CLOCK_MASK) |
				val);
			return;

		case 6:
			SEQ_out8(dpy, (SEQ_in8(dpy, 0x15) & ~CLK6_CLOCK_MASK) |
				(val & CLK6_CLOCK_MASK), 0x15);
			return;

		default:
			SEQ_out8(dpy, val, reg + 0x0F);
		}

	} else {

		DEBUG0("invalid CLK register index (%i)", reg);
	}
}

uint8 CLK_in8(struct kgi_display *dpy, uint8 reg)
{
	if (reg < 7) {

		switch (reg) {

		case 0:
			return (MISC_in8(dpy) >> 2) & 3;

		case 6:
			return SEQ_in8(dpy, 0x15) & ~CLK6_CLOCK_MASK;

		default:
			return SEQ_in8(dpy, reg + 0x0F);
		}

	} else {

		DEBUG0("invaluid CLK register index (%i)", reg);
		return 0xFF;
	}
}

#endif

/*
**	accelerator control
*/

static inline void AC_out16(struct kgi_display *dpy, uint16 val, uint16 reg)
{
	GET_PRIVATE;

	ASSERT((reg & 0xF3FF) == 0x42E8);

	/*
	**	if MEM_VGA is valid, MEM_LAW points to the new MMIO region.
	*/
	if (MEM_VGA) {

		switch (reg) {

		case 0x42e8:
			mem_out16(val, MEM_LAW + 0x1008504);
			return;

		case 0x46e8:
			mem_out16(val, MEM_LAW + 0x1008508);
			return;

		case 0x4ae8:
			mem_out16(val, MEM_LAW + 0x100850c);
			return;
		}

	} else {

		io_out16(val, IO_ACCEL + reg);
	}
}

static inline uint16 AC_in16(struct kgi_display *dpy, uint16 reg)
{
	GET_PRIVATE;

	ASSERT((reg & 0xF3FF) == 0x42E8);

	/*
	**	if MEM_VGA is valid, MEM_LAW will point to a new-MMIO region.
	*/
	if (MEM_VGA) {

		switch (reg) {

		case AC_SUBSYS_STAT:
			return mem_in16(MEM_LAW + 0x1008504);
		case 0x46e8:
			return mem_in16(MEM_LAW + 0x1008508);
		case AC_ADVFUNC_CNTL:
			return mem_in16(MEM_LAW + 0x100850c);
		}
		return 0xFFFF;

	} else {

		return io_in16(IO_ACCEL + reg);
	}
}

/*
**	accelerator setup
*/

static inline void ACCEL_out16(struct kgi_display *dpy, uint16 val, uint16 reg)
{
	GET_PRIVATE;

	if (MEM_ACCEL) {

		mem_out16(val, MEM_ACCEL + reg);

	} else {

		io_out16(val, IO_ACCEL + reg);
	}
}


static inline uint16 ACCEL_in16(struct kgi_display *dpy, uint16 reg)
{
	GET_PRIVATE;

	/*	if MEM_VGA is valid, we have a new MMIO region and can
	**	read from the MEM_ACCEL region.
	*/
	if (MEM_VGA) {

		ASSERT(MEM_ACCEL);
		return mem_in16(MEM_ACCEL + reg);

	} else {

		return io_in16(IO_ACCEL + reg);
	}
}


/*
**	State setting routines.
*/

struct statetable
{
	uint8	MISC;
	uint8	FCTRL;
	uint8	SEQ[NoSEQRegs];
	uint8	CRT[NoCRTRegs];
	uint8	GRC[NoGRCRegs];
	uint8	ATC[NoATCRegs];
};

static void restore_SEQ(struct kgi_display *dpy, uint8 *buf)
{
	register uint8 i;
	GET_PRIVATE;

	if (MEM_VGA) {

		for (i = 0; i < NoSEQRegs; i++) {

			mem_out8(i, MEM_VGA + SEQI);
			mem_out8(*(buf++), MEM_VGA + SEQD);
		}

	} else {
	
		for (i = 0; i < NoSEQRegs; i++) {

			io_out16(IO_OUT16VAL(*(buf++), i), IO_VGA + SEQI);
		}
	}
}

static void restore_CRT(struct kgi_display *dpy, uint8 *buf)
{
	register uint8 i;

	CRT_out8(dpy, CRT_in8(dpy, CRT_VSYNCEND) & ~CR11_LOCKTIMING,
		CRT_VSYNCEND);

	for(i = 0x00; i < 0x19; i++) {

		CRT_out8(dpy, buf[i], i);
	}

	for(i = 0x30; i < 0x3D; i++) {

		CRT_out8(dpy, buf[i], i);
	}

	for(i = 0x40; i < 0x6E; i++) {

		switch(i) {

		#if defined(S3_86C964) || defined(S3_86C968) || \
			defined(S3_86C928)

		case 0x42:
			CRT_out8(dpy, (CRT_in8(dpy, i) & CR42_DCLK_MASK) |
				(buf[i] & ~CR42_DCLK_MASK), i); 
			break;

		case 0x55:
			CRT_out8(dpy, (CRT_in8(dpy, i) & CR55_DAC_MASK) |
				(buf[i] & ~CR55_DAC_MASK), i);
			break;
		#endif

		#if defined(S3_86C968)
		case 0x59:
		case 0x5A:
			/*	The S3 Vision 86c968 has some trouble when 
			**	accessing these registers in new MMIO mode
			**	(clears the LAW address to 0 when writing CRTD).
			**	This 'locks' the first 64 MB of memory space.
			**	Thus we better avoid it ...
			*/
			{
				register GET_PRIVATE;

				if (MEM_VGA == NULL) {

					CRT_out8(dpy, buf[i], i);
				}
			}
			break;
		#endif

		default: 
			CRT_out8(dpy, buf[i], i);
		}
	}
}

static void restore_GRC(struct kgi_display *dpy, uint8 *buf)
{
	GET_PRIVATE;
	register uint8 i;

	if (MEM_VGA) {

		for (i = 0; i < NoGRCRegs; i++) {

			mem_out8(i, MEM_VGA + GRCI);
			mem_out8(*(buf++), MEM_VGA + GRCD);
		}

	} else {

		for (i = 0; i < NoGRCRegs; i++) {

			io_out16((*(buf++) << 8) | i, IO_VGA + GRCI);
		}
	}
}

static void restore_ATC(struct kgi_display *dpy, uint8 *buf)
{
	register uint8 i;
	GET_PRIVATE;

	if (MEM_VGA) {

		for (i = 0; i < NoATCRegs; i++) {

			mem_in8(MEM_VGA + INS1r);
			mem_out8(i, MEM_VGA + ATCI);
			mem_out8(*(buf++), MEM_VGA + ATCDw);
		}

		mem_in8(MEM_VGA + INS1r);
		mem_out8(ATCI_ENABLEDISPLAY, MEM_VGA + ATCI);

	} else {

		for (i = 0; i < NoATCRegs; i++) { 

			io_in8(IO_VGA + INS1r);
			io_out8(i, IO_VGA + ATCI);
			io_out8(*(buf++), IO_VGA + ATCDw);
		}

		io_in8(IO_VGA + INS1r);
		io_out8(ATCI_ENABLEDISPLAY, IO_VGA + ATCI);
	}
}

static void restore_state(struct kgi_display *dpy, struct statetable *state)
{
	register uint8 old;
	GET_PRIVATE;

	old = MEM_VGA ? mem_in8(MEM_VGA + MISCr) : io_in8(IO_VGA + MISCr);

	MISC_out8(dpy, (old & MISC_CLOCK_MASK) | state->MISC);

	if (MEM_VGA) {

		mem_out8(state->FCTRL, MEM_VGA + FCTRLw);

	} else {

		io_out8(state->FCTRL, IO_VGA + FCTRLw);
	}
	NOTICE2("FCTRL set");

	restore_SEQ(dpy, state->SEQ);
	NOTICE2("sequencer set");
	restore_CRT(dpy, state->CRT);
	NOTICE2("CRT set");
	restore_GRC(dpy, state->GRC);
	NOTICE2("GRC set");
	restore_ATC(dpy, state->ATC);
	NOTICE2("ATC set");
}

static unsigned short fifo_masks[13] =
{
	0x0080, 0x0040, 0x0020, 0x0010, 
	0x0008, 0x0004, 0x0002, 0x0001,
	0x8000, 0x4000, 0x2000, 0x1000,
	0x0800
/*	These are somehow broken for the Vision964

	0x0800, 0x1000, 0x2000, 0x4000, 
	0x8000, 0x0001, 0x0002, 0x0004, 
	0x0008, 0x0010, 0x0020, 0x0040 
*/
};

static inline void wait_fifo(struct kgi_display *dpy, ggi_uint i)
{
	GET_PRIVATE;
	register uint16 mask = fifo_masks[(i < 13) ? i : 12];

	/*	if MEM_VGA is valid, we have a new MMIO region and can
	**	read from the MEM_ACCEL region.
	*/
	if (MEM_VGA) {

		ASSERT(MEM_ACCEL);
		while (mask & mem_in16(MEM_ACCEL + ECR_GP_STATUS));

	} else {

		while (mask & io_in16(IO_ACCEL + ECR_GP_STATUS));
	}
}

static inline void wait_display_off(struct kgi_display *dpy)
{
	GET_PRIVATE;

	if (MEM_VGA) {

		while (~mem_in8(MEM_VGA + INS1r) & INS1_DISP_OFF);

	} else {

		while (~io_in8(IO_VGA + INS1r) & INS1_DISP_OFF);
	}
}

#if DEBUG_LEVEL > 1
static void dump_state(struct statetable *state)
{
	int i;

	printk("%.2x\n",state->MISC);
	printk("%.2x\n",state->FCTRL);

	for (i = 0; i < NoSEQRegs; i++) {

		if(! (i & 0xF)) {

			printk("\n%.2x: ", i); 
		}
		printk("%.2x, ", state->SEQ[i]);
	}

	for (i = 0; i < NoCRTRegs; i++) {

		if (! (i & 0xf)) {

			printk("\n%.2x: ",i);
		}
		printk("%.2x, ", state->CRT[i]);
	}

	for (i = 0; i < NoGRCRegs; i++) {

		if (! (i & 0xF)) {

			printk("\n%.2x: ",i);
		}
		printk("%.2x, ", state->GRC[i]);
	}

	for (i = 0; i < NoATCRegs; i++) {

		if (! (i & 0xF)) {

			printk("\n%.2x: ",i);
		}
		printk("%.2x, ", state->ATC[i]);
	}
}
#endif

