/* 
** Cirrus Logic CL-GD546x acceleration driver
**
** Copyright (C) 1997 Rodolphe Ortalo, [ortalo@laas.fr]
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2, or (at your option)
** any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#define	MAINTAINER	"Rodolphe Ortalo, [ortalo@laas.fr]"
#define	DRIVER_VERSION	"$Revision: 1.3 $"

#define	DEBUG	2
#ifdef DEBUG
	#define DEBUG_LEVEL DEBUG
#endif

/*\
|*|	Kernel includes
\*/

#include <asm/io.h>
#include <linux/ioport.h>
#include <linux/malloc.h>
#include <linux/mm.h>
#define	request_io_region	request_region
#define	check_io_region		check_region  

/*\
|*|     KGI includes
\*/

#include <ggi/module.h>
#include <ggi/ggi_commands.h>

#include "Cirrus/546x.h"

/* Ugly ? -- ortalo */
extern int CLIO_in8(int);
extern void CLIO_out8(int,int);
extern int CLIO_in16(int);
extern void CLIO_out16(int,int);
extern int CLIO_in32(int);
extern void CLIO_out32(int,int);
extern struct mem_region linFrameBuffer;

#include "suggest_546x.inc"

/* Test variables */
int fillcolor = 0;
/*
**	Functions calling the 2D hardware accelerator
*/
inline void do_colorfill_bitblt(int x, int y, int w, int h, int bgcolor)
{
	CLIO_out32(D2_OP_BGCOLOR, bgcolor);
	/* Test example */
	CLIO_out16(D2_BLTDEF, D2_RES_IN_RDRAM | D2_OP1_COLORFILL);
	CLIO_out16(D2_DRAWDEF, D2_ROP_OP1_COPY);
/*
	CLIO_out16(D2_BLTDEF, 0x1070);
	CLIO_out16(D2_DRAWDEF, 0x00CC);
*/
	CLIO_out32(D2_OP0_RDRAM, ( x & D2_X)
			| ( (y << 16) & D2_Y ) );
	CLIO_out32(D2_BLTEXT_EX, ( w & D2_X_EXTENT)
			| ( (h << 16) & D2_Y_EXTENT ) );
	return;
}
/*
**	Hardware accelerated drawing functions
**	provided by the driver.
*/
/*
**	IOCTL interface
*/
/* 2D functions
*/
static inline int cl546x_drawbox(struct kgi_graphic *gr, void *iobuf)
{	
	/* provides x,y,w,h */ /* temporary !?! */
	struct box *boxp = (struct box *)iobuf;
	int color = 0;

	if (gr->dev.mode.graphtype == GT_4BIT || 
	    gr->dev.mode.graphtype == GT_TEXT16)
	{
		
		return -E(GRAPHICS,NOSUP_MODE_MMAP);
	}

	if (boxp->w && boxp->h) {	/* 0 width is not OK! */
		/* TEST PURPOSE */
		fillcolor++;
		color = (fillcolor & 0xFF) | ( (fillcolor << 8) & 0xFF);
		color |= color << 16;
		do_colorfill_bitblt(boxp->x, boxp->y, boxp->w, boxp->h,
			color);
	}
	return 0;
}

/*
**	KGI module interface
*/

int kgim_graphics_init(struct kgi_card *card, struct kgi_display *dpy)
{
	TRACE(printk("CL-GD546x Graphics Acceleration Driver loaded\n"));
	/* !!! Check for Chipset and Version here !!! */

	return EOK;
}

void kgim_graphics_done(struct kgi_display *dpy)
{
	TRACE(printk("CL-GD546x Graphics Acceleration Driver removed\n"));
}

int kgim_graphics_check_mode (struct kgi_display *dpy, struct kgi_mode *mode,
			    enum kgi_tm_cmd cmd)
{
	/* !!! We should probably check for accel-friendly dimensions here */
	return EOK;

}

void kgim_graphics_set_mode (struct kgi_display *dpy, struct kgi_mode *mode)
{
	int ctrl_bpp = 0;
	int lncntl_bpp = 0;

	if (mode->request.graphtype != GT_TEXT16) {

	  	#warning reset the 2D & 3D acceleration engines -- ortalo
	switch (mode->request.graphtype) {
		case GT_1BIT:
		case GT_4BIT:
		case GT_TEXT16:
		case GT_TEXT32:
			NOTICE1("No acceleration in this mode: skipping init");
			return;
			break;
		case GT_8BIT:
			ctrl_bpp = D2_8BPP;
			lncntl_bpp = D2_G_FORMAT_8BPP;
			break;
		case GT_15BIT: /* Check this !!*/
			ctrl_bpp = D2_16BPP; 
			lncntl_bpp = D2_G_FORMAT_15BPP;
			break;
		case GT_16BIT:
			ctrl_bpp = D2_16BPP; 
			lncntl_bpp = D2_G_FORMAT_16BPP;
			break;
		case GT_24BIT:
			ctrl_bpp = D2_24BPP; 
			lncntl_bpp = D2_G_FORMAT_24BPP;
			break;
		case GT_32BIT:
			ctrl_bpp = D2_32BPP; 
			lncntl_bpp = D2_G_FORMAT_32BPP;
			break;
		default:
			INTERNALERROR;
			return;
		}
		/*
		**	Reset the Accelerator Engine
		*/
		NOTICE2("Resetting 2D accelerator");
		/* Reset bit mask */
		CLIO_out32(D2_BITMASK, 0xFFFFFFFF);
		/* control register */
		CLIO_out16(D2_CONTROL, 
			#warning try to remove reading (SWIZ) -- ortalo
			( CLIO_in16(D2_CONTROL) & ~D2_AUTO_BLT_EN)
			|  ctrl_bpp | D2_TILE_DISABLED );
		CLIO_out16(D2_LNCNTL, lncntl_bpp);
		/* Tile control */
		CLIO_out8(D2_TILE_CTRL, D2_NO_INTERLEAVE | D2_5TILES);
		/* No offset, framebuffer starts at 0,0 */
		CLIO_out8(D2_OFFSET_2D, 0x00);
		/* Don't know how to use this:
		CLIO_out16(D2_TAG_MASK, 0x00);
		*/
		/* bus timeout on fifo full */
		CLIO_out8(D2_TIMEOUT, D2_TIMEOUT_DISABLE);
	}
	return;
}

int kgim_graphics_command (struct kgi_graphic *gr, ggi_uint cmd,ggi_uint arg)
{ 
	ggi_suggest *sugg;
	const char *name;

	if (! (gr->dev.state & KGI_MAPPED_ON_DISPLAY) ) {

		return -E(GRAPHICS,NOSUP_MODE_MMAP);
		/* Don't accel inactive screen ! */
	}

	switch(cmd) { 
		case GRAPHICS_GETSUGGEST:
			NOTICE2("Doing GRAPHICS_GETSUGGEST");
			sugg=(ggi_suggest *)(gr->io_buf);
			gr->io_res=gr->io_buf;

			if (sugg->id<0 || sugg->id > 4)
				return -E(GRAPHICS,BOGUS);
			name=sugg_default[gr->dev.mode.graphtype].text[sugg->id];
			if (!name)
				return -E(GRAPHICS,BOGUS);
			strcpy(sugg->name,name);
			sugg->args[0]=0;
			sugg->id++;	/* Next... */
			if (sugg->id >= 4 || sugg_default[gr->dev.mode.graphtype].text[sugg->id]==NULL )
			  sugg->id=0;
			return 0;

		/* Really accelerated functions. */

		case GRAPHICS_DRAWBOX:
			return(cl546x_drawbox(gr, gr->io_buf));
		
		case GRAPHICS_COPYBOX:
			return -E(GRAPHICS,NOSUP_ALWAYS_MMAP);
/*			return(s3_copybox(gr, gr->io_buf));
*/		
		case GRAPHICS_DRAWLINE:
			return -E(GRAPHICS,NOSUP_ALWAYS_MMAP);
/*			return(s3_drawline(gr, gr->io_buf));
*/
			/* Functions that will probably benefit 
			** from lower-lever accel 
			*/
		
		case GRAPHICS_FILLSCREEN:
			return -E(GRAPHICS,NOSUP_ALWAYS_MMAP);

		/* Functions that should probably use MMAP
		*/
		default: 
			return -E(GRAPHICS,NOSUP_ALWAYS_MMAP);

/* These functions cannot be supported via standard interfaces due to
** pointers to userspace in the args, but are efficiently emulated via MMAP:
** GRAPHICS_PUTHLINE	GRAPHICS_GETHLINE	GRAPHICS_PUTVLINE	GRAPHICS_GETVLINE	
** GRAPHICS_GETBOX	GRAPHICS_PUTBOX
*/

/* These are not implemented due to efficient support via MMAP.
** GRAPHICS_PUTPIXEL	GRAPHICS_GETPIXEL	GRAPHICS_DRAWPIXEL
** GRAPHICS_DRAWCIRCLE	GRAPHICS_DRAWHLINE	GRAPHICS_DRAWVLINE
*/
		}
	return -E(GRAPHICS,NOSUP_ALWAYS_CANT);	/* This driver can't do this. */
}


