/* --------------------------------------------------------------------------
**	Cirrus Logic Laguna CL-GD546x register definitions
** --------------------------------------------------------------------------
**
**	Copyright (C) 1997 by Chris Lattner
**                            Rodolphe Ortalo
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2, or (at your option)
**	any later version.
**
**	You should have received a copy of the GNU General Public License
**	along with this program; see the file COPYING.  If not, write to
**	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
**
** --------------------------------------------------------------------------
**	MAINTAINER	Rodolphe_Ortalo
*/

#ifndef _CL_546X_H
#define _CL_546X_H

#define IOAllOrValues   0xFFF000
#define IOMask            0x0FFF

#define IOMMOrValue       0x1000
	#define IOMMIOBaseRegister		0x1000
#define IOSROrValue       0x2000
	#define IOSRBaseRegister		0x2000
#define IOGROrValue       0x4000
	#define IOGRBaseRegister		0x4000
#define IOATCOrValue      0x8000
	#define IOATCBaseRegister		0x8000

/* --------------------------------------------------------------------------
**	PCI Stuff
** --------------------------------------------------------------------------
*/
#define PCI_CL_GD5462 0xD0
#define PCI_CL_GD5464 0xD4

struct PCIDevice_s
{	unsigned short id;
	const char *string;
};

/* --------------------------------------------------------------------------
**	Miscellaneous Register
** --------------------------------------------------------------------------
*/
#define STAT_INPUT_STATUS_1	0x3DA
	#define STAT_DISPLAY_ENABLE	0x01
	#define STAT_VERTICAL_RETRACE	0x08

#define MISC_INPUT		0x1080
#define MISC_OUTPUT		0x1080
	#define MISC_NEG_V_SYNC		0x80
	#define MISC_NEG_H_SYNC		0x40
	#define MISC_PAGE_SELECT	0x20
	#define MISC_ENABLE_DISPLAY_MEMORY 0x02
	#define MISC_IOADDRESS_COLOR	0x01
	#define MISC_IOADRESS_MONO	0x00
	#define MISC_VCLK_MASK		0x0C

	#define MISC_VCLK0		0x00
		/* usually 25 MHz */
	#define MISC_VCLK1		0x04
		/* usually 28 MHz */
	#define MISC_VCLK2		0x08
		/* usually 41 MHz */
	#define MISC_VCLK3		0x0C
		/* usually 36 MHz, but also the	used programmable clock */

#define	PCIFC_VENDOR_SPECIFIC_CONTROL	(0x03FC | IOMMOrValue)
	#define PCIFC_EXT_DISP	0x10000000

/* --------------------------------------------------------------------------
**	Sequencer Registers
** --------------------------------------------------------------------------
*/
#define SRX_INDEX		0x3C4
#define SRD_DATA		0x3C5

#define SR0_RESET		0x2000
	#define SR0_SYNC_RESET		0x02
	#define SR0_ASYNC_RESET		0x01

#define SR1_CLOCKING_MODE	0x2001
	#define SR1_8_DOT_CLOCK		0x01
#define SR2_PLANE_MASK		0x2002
	#define SR2_MAP_ENABLE_MASK	0x0F
	#define SR2_PLANE_0		0x01
	#define SR2_PLANE_1		0x02
	#define SR2_PLANE_2		0x04
	#define SR2_PLANE_3		0x08

#define SR3_CHAR_MAP_SELECT	0x2003
#define SR4_MEMORY_MODE		0x2004
	#define SR4_EXTENDED_MEMORY	0x02
	#define SR4_ODD_EVEN		0x04
	#define SR4_CHAIN4		0x08
#define SR6_UNLOCK_IO		0x2006		/* Obsolete  */
#define SR7_EXTENDED_SEQ_MODE	0x2007
	#define SR7_8BPP_PACKED_PIXEL	0x01
#define SR9_SCRATCH_0		0x2009
#define SRA_SCRATCH_1		0x200A
#define SR14_SCRATCH_2		0x2014
#define SR15_SCRATCH_3		0x2015
	#define VCLK_D_MASK		0xFE
	#define VCLK_P_MASK		0x01
#define SR0B_VCLK0_D_P		0x200B
#define SR0C_VCLK1_D_P		0x200C
#define SR0D_VCLK2_D_P		0x200D
#define SR0E_VCLK3_D_P		0x1084	/* MMIO */
#define SR18_SIGNATURE_CONTROL	0x1090 /* Used for board level testing */
#define SR19_SIGNATURE_RES_L	0x1094 /* Not used, we assume that the */
#define SR1A_SIGNATURE_RES_H	0x1098 /* board works... ;-)           */
	#define VCLK_N_MASK		0x7F
#define SR1B_VCLK3_N		0x201B
#define SR1C_VCLK3_N		0x201C
#define SR1D_VCLK3_N		0x201D
#define SR1E_VCLK3_N		0x1088	/* MMIO */

/* --------------------------------------------------------------------------
**	Graphics Controller Registers
** --------------------------------------------------------------------------
*/
#define GRX_INDEX		0x3CE
#define GRD_DATA		0x3CF

#define GR0_SET_RESET		0x4000
#define GR1_SET_RESET_ENABLE	0x4001
#define GR2_COLOR_COMPARE	0x4002
#define GR3_DATA_ROTATE		0x4003
#define GR4_READ_MAP_SELECT	0x4004
	#define GR4_PLANE0		0x00
	#define GR4_PLANE1		0x01
	#define GR4_PLANE2		0x02
	#define GR4_PLANE3		0x03
#define GR5_MODE		0x4005
	#define GR5_ODD_EVEN		0x10
	#define GR5_256COLOR		0x40
	#define GR5_WRITE_MODE_MASK	0x03
	#define GR5_WRITE_MODE_0	0x00
	#define GR5_WRITE_MODE_1	0x01
	#define GR5_WRITE_MODE_2	0x02
	#define GR5_WRITE_MODE_3	0x03
#define GR6_MISCELLANEOUS	0x4006
	#define GR6_GRAPHICS_MODE	0x01
	#define GR6_CHAIN_ODD_TO_EVEN	0x02
	#define GR6_MEMORY_MAP_MASK	0x0C
	#define GR6_MEMORY_MAP_EXTENDED	0x00
	#define GR6_MEMORY_MAP_VGA	0x04
	#define GR6_MEMORY_MAP_HERCULE	0x08
	#define GR6_MEMORY_MAP_CGA	0x0C
#define GR7_COLOR_DONT_CARE	0x4007
#define GR8_BIT_MASK		0x4008
#define GR9_OFFSET_0		0x4009		/* Obsolete (why? -- RO) */
#define GRA_OFFSET_1		0x400A		/* Obsolete (why? -- RO) */
#define GRB_MODE_EXTENSIONS	0x400B		/* Obsolete (why? -- RO) */

/* --------------------------------------------------------------------------
**	Attribute Controller Registers
** --------------------------------------------------------------------------
*/
#define ARX_INDEX		0x3C0		/* TODO:Enable MMIO x7C */
	#define ARX_VIDEO_ENABLE	0x20
#define ARD_DATA_READ		0x3C1
#define ARD_DATA_WRITE		0x3C0

#define AR0_PALETTE		0x8000 /* ranges from 0 to 0xF */
#define AR10_MODE		0x8010
	#define AR10_BLINK_ENABLE	0x08
	#define AR10_LINE_GRAPHICS_ENABLE 0x04
	#define AR10_GRAPHICS_MODE	0x01
#define AR11_OVERSCAN_COLOR	0x8011
#define AR12_COLOR_PLANE_ENABLE	0x8012
#define AR13_PIXEL_PANNING	0x8013
	#define AR13_PIXEL_PANNING_MASK	0x0F
	#define AR13_8BIT_NO_SHIFT	0x00
	#define AR13_9BIT_NO_SHIFT	0x08
#define AR14_COLOR_SELECT	0x8014

/* --------------------------------------------------------------------------
**	CRT Controller Registers
** --------------------------------------------------------------------------
*/

#define CRX_INDEX		0x3D4
#define CRD_DATA		0x3D5

#define CR0_HORIZONTAL_TOTAL	0x1000
#define CR1_HORIZONTAL_DISPLAY_END 0x1004
#define CR2_HORIZONTAL_BLANKING_START 0x1008
#define CR3_HORIZONTAL_BLANKING_END 0x100C
	#define CR3_COMPATIBLE_READ	0x80
	#define CR3_HOR_BLANK_END_MASK	0x1F
#define CR4_HORIZONTAL_SYNC_START 0x1010
#define CR5_HORIZONTAL_SYNC_END	0x1014
	#define CR5_HOR_BLANK_END_5	0x80
	#define CR5_HOR_SYNC_END_MASK	0x1F
#define CR6_VERTICAL_TOTAL	0x1018
#define CR7_OVERFLOW		0x101C
	#define CR7_VER_SYNC_START_9	0x80
	#define CR7_VER_DISPLAY_END_9	0x40
	#define CR7_VER_TOTAL_9		0x20
	#define CR7_LINE_COMPARE_8	0x10
	#define CR7_VER_BLANK_START_8	0x08
	#define CR7_VER_SYNC_START_8	0x04
	#define CR7_VER_DISPLAY_END_8	0x02
	#define CR7_VER_TOTAL_8		0x01
#define CR8_SCREENA_PRESET_ROW_SCAN 0x1020
	#define CR8_ROW_SCAN_MASK	0x1F
#define CR9_CHARACTER_CELL_HEIGHT 0x1024
	#define CR9_SCAN_DOUBLE		0x80
	#define CR9_LINE_COMPARE_9	0x40
	#define CR9_VER_BLANK_START_9	0x20
	#define CR9_CELL_HEIGHT_MASK	0x1F
#define CRA_TEXT_CURSOR_START	0x1028
#define CRB_TEXT_CURSOR_END	0x102C
#define CRC_SCREEN_START_ADDRESS_HIGH 0x1030
#define CRD_SCREEN_START_ADDRESS_LOW 0x1034
#define CRE_TEXT_CURSOR_LOCATION_HIGH 0x1038
#define CRF_TEXT_CURSOR_LOCATION_LOW 0x103C
#define CR10_VERTICAL_SYNC_START 0x1040
#define CR11_VERTICAL_SYNC_END	0x1044
	#define CR11_WRITE_PROTECT_CR0_7 0x80
	#define CR11_VER_SYNC_END_MASK	0x0F
#define CR12_VERTICAL_DISPLAY_END 0x1048
#define CR13_OFFSET		0x104C
#define CR14_UNDERLINE_ROW_SCANLINE 0x1050
	#define CR14_UNDERLINE_MASK	0x1F
	#define CR14_DOUBLEWORD_MODE	0x40
	#define CR14_COUNT_BY_FOUR	0x20
#define CR15_VERTICAL_BLANK_START 0x1054
#define CR16_VERTICAL_BLANK_END	0x1058
#define CR17_MODE_CONTROL	0x105C
	#define CR17_TIMING_ENABLE	0x80
	#define CR17_BYTE_MODE		0x40
	#define CR17_ADDRESS_WRAP	0x20
	#define CR17_SELECT_ROW_SCAN_COUNTER 0x02
	#define CR17_CGA_SUPPORT	0x01
#define CR18_LINE_COMPARE	0x1060
#define CR19_INTERLACE_END	0x1064
#define CR1A_MISCELLANEOUS_CONTROL 0x1068
	#define CR1A_HOR_BLANK_END_6	0x10
	#define CR1A_HOR_BLANK_END_7	0x20
	#define CR1A_VER_BLANK_END_8	0x40
	#define CR1A_VER_BLANK_END_9	0x80
	#define CR1A_ENABLE_DOUBLE_BUFFERED_START_ADDRESS 0x02
	#define CR1A_VER_BLANK_END_89_MASK 0xC0
	#define CR1A_HOR_BLANK_END_67_MASK 0x30
#define CR1B_EXTENDED_DISPLAY_CONTROL 0x106C
	#define CR1B_BLANKING_CONTROL	0x20
	#define CR1B_ENABLE_BLANK_END_EXTENSIONS 0x80
	#define CR1B_OFFSET_8		0x10
	#define CR1B_SCREENA_ADDRESS_1718_MASK 0x0C
	#define CR1B_ENABLE_EXTENDED_ADDRESS_WRAP 0x02
#define CR1D_SCREENA_START_EXTENSION 0x1074
	#define CR1D_OFFSET_9		0x01
	#define CR1D_SCREENA_ADDRESS_1920_MASK 0x18
#define CR1E_TIMING_OVERFLOW	0x1078
	#define CR1E_HOR_TOTAL_8	0x80
	#define CR1E_HOR_DISPLAY_END_8	0x40
	#define CR1E_HOR_BLANK_START_8	0x20
	#define CR1E_HOR_SYNC_START_8	0x10
	#define CR1E_VER_TOTAL_10	0x08
	#define CR1E_VER_DISPLAY_END_10	0x04
	#define CR1E_VER_BLANK_START_10	0x02
	#define CR1E_VER_SYNC_START_10	0x01

/* --------------------------------------------------------------------------
**	DAC Registers
** --------------------------------------------------------------------------
*/
#define DAC_PIXEL_MASK		0x10A0
#define DAC_PALETTE_ADDRESS_READ 0x10A4
#define DAC_STATE		0x10A4
#define DAC_PALETTE_ADDRESS_WRITE 0x10A8
#define DAC_PIXEL_DATA		0x10AC

/* --------------------------------------------------------------------------
**	Video Pipeline Registers
** --------------------------------------------------------------------------
*/
#define GRAPHICS_VIDEO_FORMAT	0x10C0
	#define VCLK_MULTIPLY_BY_2	0x4000
	/* Graphics depth */
	#define GRAPHICS_DEPTH_8BITS	0x0000
	#define GRAPHICS_DEPTH_16BITS	0x1000
	#define GRAPHICS_DEPTH_24BITS	0x2000
	#define GRAPHICS_DEPTH_32BITS	0x3000
	#define GRAPHICS_DEPTH_MASK	0x3000
	/* Graphics format */
	#define GRAPHICS_FORMAT_PALETTIZED 0x0000
	#define GRAPHICS_FORMAT_GREYSCALE 0x0200
	#define GRAPHICS_FORMAT_RGB	0x0400
	#define GRAPHICS_FORMAT_RGB15	0x0600
	#define GRAPHICS_FORMAT_MASK	0x0E00
	/* Gamma correction */
	#define GRAPHICS_GAMMA_CORRECTION_ENABLE 0x0100
	#define GRAPHICS_8BITS_DAC_PALETTE 0x0100

#define	PALETTE_STATE		0x10B0
	#define ACCESS_CURSOR_COLORS	0x04
#define	CURSOR_X		0x10E0
	#define CURSOR_X_MASK		0x0FFF
#define	CURSOR_Y		0x10E2
	#define CURSOR_Y_MASK		0x0FFF
#define	CURSOR_PRESET		0x10E4
	#define	CURSOR_X_PRESET		0x7F00
	#define	CURSOR_Y_PRESET		0x007F
#define TILING_CONTROL		0x10EA
	#define ONE_WAY_INTERLEAVE	0x0000
	#define TILING_ENABLE		0x0080
	#define WIDE_TILES		0x0040
	#define TILES_PER_LINE_MASK	0x3F00
	#define VIDEO_5TILES		0x0500
	#define VIDEO_8TILES		0x0800
	#define FIFO_THRESHOLD_MASK	0x003F
#define	MISCELLANEOUS_CONTROL	0x10E6
	#define CURSOR_ENABLE	0x0001
#define	CURSOR_ADDRESS		0x10E8
	#define CURSOR_ADDRESS_MASK	0x7FFC

/* --------------------------------------------------------------------------
**	2D Accelerator Engine Registers
** --------------------------------------------------------------------------
*/
#define D2_BITMASK		0x15E8
#define D2_CONTROL		0x1402
	#define D2_8BPP			0x0000
	#define D2_16BPP		0x2000
	#define D2_24BPP		0x4000
	#define D2_32BPP		0x6000
	#define D2_TILE_DISABLED	0x1000 /* disabled & 128 in fact */
	#define D2_AUTO_BLT_EN		0x0100
#define D2_LNCNTL		0x150E
	#define D2_G_FORMAT_8BPP	0x0000	
	#define D2_G_FORMAT_15BPP	0x0100	
	#define D2_G_FORMAT_16BPP	0x0200	
	#define D2_G_FORMAT_24BPP	0x0500	
	#define D2_G_FORMAT_32BPP	0x0400	
#define D2_OFFSET_2D		0x1405
#define D2_TAG_MASK		0x15EC
#define D2_TILE_CTRL		0x1407
	#define D2_NO_INTERLEAVE	0x00
	#define D2_5TILES		0x05
	#define D2_8TILES		0x08
#define D2_TIMEOUT		0x1406
	#define D2_TIMEOUT_DISABLE	0x00

#define D2_OP_BGCOLOR		0x15E4
#define D2_BLTDEF		0x1586
	#define D2_Y_DIR		0x8000
	#define D2_RES_IN_RDRAM		0x1000
	#define D2_RES_IN_HOST		0x2000
	#define D2_RES_IN_SRAM0		0x4000
	#define D2_RES_IN_SRAM1		0x5000
	#define D2_RES_IN_SRAM2		0x6000
	#define D2_RES_IN_SRAM12	0x7000
	#define D2_SAME			0x0800
	#define D2_OP0_IN_SDRAM		0x0000
	#define D2_OP0_IN_RDRAM		0x0100
	#define D2_OP1_PATTERN		0x0080
	#define D2_OP1_IN_SRAM		0x0000
	#define D2_OP1_IN_RDRAM		0x0010
	#define D2_OP1_IN_HOST		0x0020
	#define D2_OP1_IN_SRAM_MONO	0x0040
	#define D2_OP1_IN_RDRAM_MONO	0x0050
	#define D2_OP1_IN_HOST_MONO	0x0060
	#define D2_OP1_COLORFILL	0x0070
	#define D2_OP2_PATTERN		0x0008
	#define D2_OP2_IN_SRAM		0x0000
	#define D2_OP2_IN_RDRAM		0x0001
	#define D2_OP2_IN_HOST		0x0002
	#define D2_OP2_IN_SRAM_MONO	0x0004
	#define D2_OP2_IN_RDRAM_MONO	0x0005
	#define D2_OP2_IN_HOST_MONO	0x0006
	#define D2_OP2_COLORFILL	0x0007
#define D2_DRAWDEF		0x1584
	#define D2_SAT1			0x8000
	#define D2_SAT2			0x4000
	#define D2_BITMASK_EN		0x2000
	#define D2_PTAG			0x0400
	#define D2_TRANSOP_NOTEQUAL	0x0200
	#define D2_TRANS		0x0100
	#define D2_ROP_MASK		0x00FF
	/* ROP operations: search them ! */
	#define D2_ROP_OP1_COPY		0x00CC

#define D2_OP0_RDRAM		0x1520
#define D2_OP1_RDRAM		0x1540
#define D2_OP2_RDRAM		0x1560
	#define D2_X			0x00000FFF
	#define D2_Y			0x0FFF0000
#define D2_BLTEXT_EX		0x1700
	#define D2_X_EXTENT		0x00000FFF
	#define D2_Y_EXTENT		0x0FFF0000
	#define D2_X_EXTENT16		0x0FFF
	#define D2_Y_EXTENT16		0x0FFF

#endif /* #ifdef _CL_546X_H */
