/*
** Driver stuff
*/
#ifndef _MGA_COMMON_H_
#define _MGA_COMMON_H_

#include <Matrox/mga_common_lib.h>

#ifndef PCI_DEVICE_ID_MATROX_MIL_2
#  define PCI_DEVICE_ID_MATROX_MIL_2 0x051B
#endif

#ifndef PCI_DEVICE_ID_MATROX_MIL_2_AGP
#  define PCI_DEVICE_ID_MATROX_MIL_2_AGP 0x051F
#endif

/*
** Structs
*/
struct common_private { /* Vars common to both Millenium and Mystique */
	struct kgi_chipset chipinfo;
	
	struct mem_region framebuffer;
	struct mem_region aperture;
	struct kgi_mmio_region graphic_mmio_linear_fb;

	struct wait_queue *vsync_wait;
	
	uint8   irq;
	uint8   rev;
	uint8   bus;
	uint8   dev_fn;
	uint8   display_running;
	uint32  chip_id;
	struct  kgi_input irq_report;
	
	ggi_uint origin, org_xdot, org_ydot, org_offs;
        ggi_uint frame;
	ggi_uint split;
	ggi_uint fb_size; /* # of pixels NOT bytes */
	ggi_uint bpp;

        uint32  ptr_old, cur_old;
        uint16  ptr_and, cur_and;
        uint16  ptr_xor, cur_xor;
};

/* This is ugly, but is there something better? */
#ifdef MGA_CHIPSET_DRIVER
#define GET_PRIVATE \
        struct chipset_private *private = KGIM_PRIVATE(dpy, chipset)
#else
struct common_private_alone {
	struct common_private cp;
};
#define GET_PRIVATE \
        struct common_private_alone *private = KGIM_PRIVATE(dpy, chipset)
#endif

/* 
** Convenience (?) macros 
*/

#define MGA_APERTURE    (private->cp.aperture)
#define MGA_FRAMEBUFFER (private->cp.framebuffer)
#define MGA_BASE1       (MGA_APERTURE.baseptr)     /* control aperture */
#define MGA_BASE2       (MGA_FRAMEBUFFER.baseptr)  /* frame buffer     */
#define MGA_IRQ         (private->cp.irq)
#define MGA_PCI_BUS     (private->cp.bus)
#define MGA_PCI_DEV_FN  (private->cp.dev_fn)
#define MGA_CHIP_ID     (private->cp.chip_id)
#define MGA_CP          (private->cp)

/*
** Imported symbols. These are supposed to be defined by the *_main files.
*/

extern const struct kgi_chipset chipinfo;

extern const struct mem_region mga_aperture;
extern const struct mem_region mga_framebuffer;
extern const struct kgi_mmio_region graphic_mmio_linear_fb;

extern const char chipset_name_init[]; /* Print on klog on driver init */

void adjust_timing(struct kgi_mode *mode);
int adjust_virtual(struct kgi_mode *mode);
void mga_interrupt_handler(int irq, void *d, struct pt_regs *regs);

/*
** Prototypes
*/
int mga_chipset_init(struct kgi_card *card, struct kgi_display *dpy,
		     int privsize, int productID);
/* Please have a look at the comment in mga_common.c */
void mga_chipset_done(struct kgi_display *dpy);

#endif /* _MGA_COMMON_H_ */
