/*\
|*| ----------------------------------------------------------------------------
|*|     GGI device file definitions
|*| ----------------------------------------------------------------------------
|*|
|*|	Copyright (C) 1995 by Andreas Beck, [becka@hp.rz.uni-duesseldorf.de]
|*|			  and Steffen Seeger, [seeger@physik.tu-chemnitz.de]
|*|
|*|	This program is free software; you can redistribute it and/or modify
|*|	it under the terms of the GNU General Public License as published by
|*|	the Free Software Foundation; either version 2, or (at your option)
|*|	any later version.
|*|
|*|	You should have received a copy of the GNU General Public License
|*|	along with this program; see the file COPYING.  If not, write to
|*|	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
|*|
|*| ---------------------------------------------------------------------------
|*|	MAINTAINER	"Steffen Seeger, [seeger@physik.tu-chemnitz.de]"
\*/

#ifndef _GGI_ERRNO_H
#define _GGI_ERRNO_H

#if defined(linux) || defined(__linux) || defined(__linux__)
#include <linux/errno.h>
#else
#include <errno.h>
#endif

/*\
|*|	Error codes
\*/

typedef int	ggi_error;

enum kgi_subsystems
{
	DRIVER,		MONITOR,	RAMDAC,		CLOCK,
	CHIPSET,	GRAPHICS,	LIB,
	KGI_NR_SUBSYSTEMS = 0xF
};

enum kgi_nosup_codes { 	/* Methods */
			AVM_IOCTL =0x00,	/* ioctl is ok - use it */
			AVM_MMAP  =0x01,	/* use mmap */
			AVM_LOWER =0x02,	/* use lower level functions */
			AVM_CANT  =0x03,	/* no support (like setpal in 16M) */
			AVM_MASK  =0x0f,
			/* Expire-times */
			AVE_TEST  =0x00,	/* Test it ... */
			AVE_NOW   =0x00,	/* Retry next time */
			AVE_MODE  =0x10,	/* reset when mode changes */
			AVE_ALWAYS=0x20,	/* Always do it this way   */
			AVE_MASK  =0xf0 };

enum kgi_error_code
{
	OK		= 0x00000000,	/* request done			*/

	UNKNOWN,		/* unspecified error			*/
	BOGUS,			/* request doesn't make sense		*/
	FAILED,			/* operation requested not possible	*/

	MODE_CHECK	= 0x00000100,	/* mode check failed,		*/

	CLK_LOW_LIMIT,		/* clock rate out of lower bound	*/
	CLK_HIGH_LIMIT,		/* clock rate out of upper bound	*/
	CLK_LIMIT,		/* clock rate out of range		*/
	RESOLUTION_LOW,		/* resolution too bad			*/
	RESOLUTION_HIGH,	/* resolution too high			*/

	NOSUP		= 0x00000200,			/* "no support" codes	*/
	NOSUP_NOW_MMAP	= NOSUP|AVE_NOW|AVM_MMAP,	/* can't do this right now. Use MMAP instead */
	NOSUP_NOW_LOWER	= NOSUP|AVE_NOW|AVM_LOWER,	/* can't do this right now. Use simpler commands */
	NOSUP_NOW_CANT	= NOSUP|AVE_NOW|AVM_CANT,	/* can't do it now. */
	NOSUP_MODE_MMAP	= NOSUP|AVE_MODE|AVM_MMAP,	/* can't do this in this mode. Use MMAP */
	NOSUP_MODE_LOWER= NOSUP|AVE_MODE|AVM_LOWER,	/* can't do this in this mode. Use simpler commands */
	NOSUP_MODE_CANT	= NOSUP|AVE_MODE|AVM_CANT,	/* impossible in this mode. */
	NOSUP_ALWAYS_MMAP = NOSUP|AVE_ALWAYS|AVM_MMAP,	/* not supported. Use MMAP */
	NOSUP_ALWAYS_LOWER= NOSUP|AVE_ALWAYS|AVM_LOWER,	/* not supported. Use simpler commands */
	NOSUP_ALWAYS_CANT = NOSUP|AVE_ALWAYS|AVM_CANT,	/* not supported at all. */

	NOSUP_MASK       = 0x00ffff00	/* these bits must match NOSUP,if it is a NOSUP-type error */
};

#define	E(subsystem, code)	((subsystem << 24) | code)
#define	ESUBSYSTEM(err)		(err >> 24)
#define	ECODE(err)		(err & 0x00FFFFFF)

#define	EOK	E(DRIVER, OK)

#endif	/* #ifdef _GGI_ERRNO_H */
