/* -----------------------------------------------------------------------------
**	GGI focus definitions
** -----------------------------------------------------------------------------
**
**	Copyright (C) 1996-1997, Steffen Seeger
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2, or (at your option)
**	any later version.
**
**	You should have received a copy of the GNU General Public License
**	along with this program; see the file COPYING.  If not, write to
**	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
**
** -----------------------------------------------------------------------------
*/

#ifndef _GGI_FOCUS_H
#define _GGI_FOCUS_H

#include <ggi/keyboard.h>

#ifdef __linux__
#include <linux/string.h>
#include <linux/kd.h>
#endif

/*	Symbolic names for the keyboard LEDs. A keyboard driver should just
**	set the ones that are possible and ignore the rest.
*/
#define	KBD_LED_SCROLLOCK	0
#define	KBD_LED_NUMLOCK		1
#define	KBD_LED_CAPSLOCK	2

/*	We allow for up to eight modifier keys, the names below have no
**	meaning at all, but it is recommended to map them according to
**	their names. They are unlikely to be extended, just because
**	you normally have only ten fingers.
*/
typedef enum ggi_modifier_keys {

	mkShift,	mkCtrl,		mkAlt,		mkAltGr,
	mkShiftL,	mkShiftR,	mkCtrlL,	mkCtrlR

} ggi_modifier_keys;

#define	MAR_NR_KEYMAPS		(1 << MAX_NR_MODIFIER_KEYS)
#define	MAX_NR_USER_KEYMAPS	MAX_NR_KEYMAPS

struct focus_keymap
{
	ggi_uint	fn_buf_size;	/* size of function string buffer	*/
	char	*fn_buf;	/* pointer to funtion string buffer	*/

	ggi_uint	fn_str_size;	/* size of function key string buffer	*/
	char	**fn_str;	/* ptr to arr of ptrs to strings	*/

	ggi_uint	keymap_size;	/* number of valid entries in keymap[]	*/
	ggi_uint	keymin, keymax;	/* min/max key values accepted		*/
	unicode *keymap[MAX_NR_KEYMAPS];	/* the keymaps		*/
};

struct kgi_focus
{
	struct kgi_input	*inputs;	/* input devices	*/

	MODIFIER_DATA;				/* modifier information	*/
	ggi_uint			down_mod[8];	/* modifier counters	*/
	struct focus_keymap	*kmap;		/* key map info		*/

	ggi_uint	flags;
	#define	CAPS_SHIFT	1

	ggi_uint	leds;		/* lights to switch on		*/

	ggi_uint	x, y;		/* pointer hot spot position	*/
	ggi_uint	buttons;	/* pointer button state		*/
	ggi_uint	maxx, maxy;	/* pointer hot spot window	*/
	ggi_uint	minx, miny;	/* pointer hot spot window	*/

	ggi_uint	npadch;

	int	last_focus;
	int	want_focus;
};

#define U(x) ((x) ^ 0xf000)

extern int kbd_init(void);

#endif	/* #ifdef _GGI_FOCUS_H	*/

