/*\
|*| ----------------------------------------------------------------------------
|*|	GGI font and character handling
|*| ----------------------------------------------------------------------------
|*|
|*|	Copyright (C) 1996 by Steffen Seeger, [seeger@physik.tu-chemnitz.de]
|*|
|*|	This program is free software; you can redistribute it and/or modify
|*|	it under the terms of the GNU General Public License as published by
|*|	the Free Software Foundation; either version 2, or (at your option)
|*|	any later version.
|*|
|*|	You should have received a copy of the GNU General Public License
|*|	along with this program; see the file COPYING.  If not, write to
|*|	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
|*|
|*| ----------------------------------------------------------------------------
|*|	MAINTAINER	"Steffen Seeger, [seeger@physik.tu-chemnitz.de]"
\*/

#ifndef _GGI_FONT_H
#define _GGI_FONT_H

/*\
|*| ----------------------------------------------------------------------------
|*|	characters
|*| ----------------------------------------------------------------------------
|*|	ISO 10646 defines symbol codes as 32 bit unsinged int organized into
|*|	groups, planes, rows and cells. By now only the first plane (group=0,
|*|	plane=0) is defined, which maps directly to UNICODE. Thus isochar to
|*|	font position mapping only cares about UNICODE (but can be extended).
\*/

typedef uint32 isochar;
typedef uint16 unicode;

#define ISO_GROUP(x)	(x >> 24)
#define	ISO_PLANE(x)	((x >> 16) & 0xFF)
#define	ISO_ROW(x)	((x >> 8) & 0xFF)
#define	ISO_CELL(x)	(x & 0xFF)

#define	ISO_GROUP_MASK	0xFF000000
#define	ISO_PLANE_MASK	0x00FF0000
#define	ISO_ROW_MASK	0x0000FF00
#define	ISO_CELL_MASK	0x000000FF

typedef struct { isochar sym; unsigned int pos; }	ggi_cellinfo;

/*\
|*| ----------------------------------------------------------------------------
|*|	kernel fonts
|*| ----------------------------------------------------------------------------
|*|	The kernel fonts have (at maximum) 256 font positions as this can
|*|	(hopefully) be done or emulated on all displays. To support ISO10646 as
|*|	good and fast as possible, we use a (speed and common case) optimized
|*|	mapping not to exhaust memory:
|*|
|*|	Generally cellinfo[ISO_ROW(x)] points to data about row ISO_ROW(x):
|*|
|*|	  * NULL if the font has no positions that map to this row.
|*|	  * if bit number row of map_direct is set, cellinfo[row] is a pointer
|*|	    to an 'uchar map[256]' array, map[ISO_CELL(sym)] giving the font
|*|	    position. Use this if more than 31 font postions map to this row.
|*|	  * else cellinfo[row] points to an array ggi_cellinfo's, sorted in
|*|	    ascending order of cellinfo.sym and the first giving total number
|*|	    of cells and highest cell assigned.
\*/

  struct kgi_fontinfo
  {	ggi_uint	positions;		/* number of positions defined	*/
	ggi_uint	default_pos;		/* fontpos to use as default	*/
	ggi_uint	map_direct[8];
	void	*cellinfo[256];		/* map info for iso -> fontpos	*/
	isochar	inversemap[256];	/* map info for fontpos -> iso	*/
  };

  struct kgi_font
  {	ggi_uint	height, width;		/* distance between char's	*/
	ggi_uint	dx, dy;			/* size of bitmap per char	*/
	ggi_uint	base;			/* base line			*/

	struct kgi_fontinfo *info;	/* iso10646 <-> fontpos mapping	*/
	void	*data;			/* bitmaps			*/
  };

#ifdef NEED_KGI_PROTOTYPES

  #ifdef __KERNEL__
    #define MAX_NR_DEFAULT_FONTS 4
    extern struct kgi_font *kgi_default_font(ggi_mode *mode);
    extern ggi_uint  kgi_fontpos(struct kgi_font *font, isochar sym);
  #endif
#endif	/* #ifdef NEED_KGI_PROTOTYPES */

/*\
|*| ----------------------------------------------------------------------------
|*|	application fonts
|*| ----------------------------------------------------------------------------
|*| !!!	application fonts will have any number of font postitions and
|*| !!!	'usual' font attributes, for now we use the same as in kernel space. 
\*/

typedef struct kgi_fontinfo	ggi_fontinfo;
typedef struct kgi_font		ggi_font;

#ifndef __KERNEL__
  extern ggi_uint	ggi_fontpos(ggi_font *font, isochar sym);
#endif

#endif	/* ifdef _GGI_FONT_H */
