/* 
   Graphics library for GGI.
   Generic Gamma Map driver

   Copyright (C) 1997 Jason McMullan - jmcc@cs.cmu.edu

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free
  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdlib.h>
#include <stdio.h>
#include <sys/ioctl.h>

#ifdef _AIX
#include <sys/types.h>
#include <unistd.h>
#endif

#include "ggi-dl.h"
#include <ggi/ggi_commands.h>

int        _GGInocols=0;
ggi_color *_GGIcolormap=NULL;

/* 
 * We are guaranteed to have been checked already...
 */

int GGIdlinit(ggi_visual *visual,const char *version)
{
	int err;

	/* Gamma mapping 
	 */
	_GGInocols=256;
	_GGIcolormap=(ggi_color *)malloc(sizeof(ggi_color)*256);
	if (_GGIcolormap==NULL)
		return -1;

	err=ioctl(visual->fd,(int)RAMDAC_GETCLUT,_GGIcolormap);
	if (err<0) {
		fprintf(stderr,"ioctl-gamma: Can't get default colormap\n");
		free(_GGIcolormap);
		return -1;
	}
	
	visual->opcolor->getgamma=GGIgetgamma;	
	visual->opcolor->setgamma=GGIsetgamma;	
	visual->opcolor->getgammamap=GGIgetgammamap;
	visual->opcolor->setgammamap=GGIsetgammamap;

	return GGI_DL_OPCOLOR;
}

int GGIdlcleanup(ggi_visual *visual)
{
	free(_GGIcolormap);

	return EOK;
}
		

