/* 
   Graphics library for GGI.

   Copyright (C) 1995 Andreas Beck - becka@hp.rz.uni-duesseldorf.de
   Copyright (C) 1997 Jason McMullan - jmcc@cs.cmu.edu

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free
  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/mman.h>

#include "ggi-dl.h"
#include <ggi/ggi_commands.h>

/* 
 * We are guaranteed to have been checked already...
 */

static ggi_gc *_ggiGC=NULL;

int GGIdlinit(ggi_visual *visual,const char *version)
{
	if (VIS_GC(visual)==NULL) {
		_ggiGC=malloc(sizeof(ggi_gc));
		if (_ggiGC==NULL) {
			return GGI_DL_ERROR;
		}
		_ggiGC->version=0;
/*		_ggiLockInit(_ggiGC);*/
	}

	if (VIS_GC(visual)==NULL)
		VIS_GC(visual)=_ggiGC;

	GGIdbinit(visual);

	/* Color mapping 
	 */
	visual->opcolor->mapcolor=GGImapcolor;
	visual->opcolor->unmappixel=GGIunmappixel;
	visual->opcolor->packcolors=GGIpackcolors;
	visual->opcolor->unpackpixels=GGIunpackpixels;

	/* Positioning is not supported...
	 */

	/* Sprites aren't supported...
	 */

	/* Generic drawing
	 */
	visual->op2d->fillscreen=GGIfillscreen;

	visual->op2d->drawpixel=GGIdrawpixel;
	visual->op2d->putpixel=GGIputpixel;
	visual->op2d->getpixel=GGIgetpixel;

	visual->op2d->drawhline=GGIdrawhline;
	visual->op2d->puthline=GGIputhline;
	visual->op2d->gethline=GGIgethline;

	visual->op2d->drawvline=GGIdrawvline;
	visual->op2d->putvline=GGIputvline;
	visual->op2d->getvline=GGIgetvline;

	visual->op2d->copybox=GGIcopybox;

	return GGI_DL_OPCOLOR|GGI_DL_OP2D;
}

int GGIdlcleanup(ggi_visual *visual)
{
	if (VIS_GC(visual)==_ggiGC) {
/*		_ggiLockDestroy(VIS_GC(visual));*/
		free(VIS_GC(visual));
		VIS_GC(visual)=NULL;
		_ggiGC=NULL;
	}

	return EOK;
}
		

