/* 
   Graphics library for GGI. Pixels.

   Copyright (C) 1997 Jason McMullan - jmcc@cs.cmu.edu
   Copyright (C) 1995 Andreas Beck - becka@hp.rz.uni-duesseldorf.de

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "ggi-dl.h"

/*******************************/
/* draw/get/put a single pixel */
/*******************************/

int GGIdrawpixel(ggi_visual *vis,int x,int y)
{
	uint8 pel;
	uint8 clr;
	uint8 *fb;
	uint8 xs;
	
	CHECKXY(vis,x,y);
 	
	/* Read-modify-write. */
	
	fb = (uchar *)VIS_FB_LINEAR(vis)+y*(VIS_FB_WIDTH(vis)>>1)+(x>>1);
	pel = *fb;

	/* This does:
	 
	/* if (x & 1)
	 *    clr = (pel & 0xF0) | (VIS_GC_FGCOLOR(vis) & 0x0f);
	 * else clr = (pel & 0x0F) | ((VIS_GC_FGCOLOR(vis) & 0x0f) << 4);
	 */
	xs = (x & 1) << 2;
	*fb = (pel & (0x0F << xs)) | ((VIS_GC_FGCOLOR(vis) & 0x0f) << (xs ^ 4));

	return EOK;
}

int GGIputpixel(ggi_visual *vis,int x,int y,uint col)
{ 
	uint8 pel;
	uint8 clr;
	uint8 *fb;
	uint8 xs;
	
	CHECKXY(vis,x,y);
 	
	fb = (uchar *)VIS_FB_LINEAR(vis)+y*(VIS_FB_WIDTH(vis)>>1)+(x>>1);
	pel = *fb;

	xs = (x & 1) << 2;
	
	*fb=(pel & (0x0F << xs)) | ((col & 0x0f) << (xs ^ 4));

	return EOK;
}

int GGIgetpixel(ggi_visual *vis,int x,int y,uint *pixel)
{ 
	int pel;
	uint8 xs;
	
	if (pixel==NULL)
		return -1;
	
	CHECKXY(vis,x,y);
 	
	pel = *(uchar *)VIS_FB_LINEAR(vis)+y*(VIS_FB_WIDTH(vis)>>1)+(x>>1);

	xs = (x & 1) << 2;
	*pixel = (uint)((pel & (0x0f << xs)) >> (xs ^ 4)); 

	return EOK;
}

