/* 
   Graphics library for GGI.
   Generic RAMDAC via IOCTL driver

   Copyright (C) 1997 Jason McMullan - jmcc@cs.cmu.edu

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdlib.h>
#include <stdio.h>
#include <sys/ioctl.h>

#ifdef _AIX
#include <sys/types.h>
#include <unistd.h>
#endif

#include "ggi-dl.h"
#include <ggi/ggi_commands.h>

int	   _GGIbpc=0;
int        _GGInocols=0;
ggi_color *_GGIcolormap=NULL;

/* 
 * We are guaranteed to have been checked already...
 */

int GGIdlinit(ggi_visual *visual,const char *version)
{
	int err,bpc;

	/* Color mapping 
	 */
	bpc=VIS_FB_BPC(visual);
	if (VIS_MODE(visual)->graphtype==GT_TEXT16) bpc=8;
	if (bpc<=0) {
		fprintf(stderr,"generic-ramdac: No BPC for this visual!\n");
		return -1;
	}

	_GGIbpc=bpc;
	_GGInocols=(1<<bpc);
	DPRINT("generic-ramdac: bpc=%d nocols=%d\n",_GGIbpc,_GGInocols);
	if (_GGInocols > 256) {
		fprintf(stderr,"generic-ramdac: too many colors (%d)\n",
				_GGInocols);
		return -1;
	}

	_GGIcolormap=(ggi_color *)malloc(sizeof(ggi_color)*256);
	if (_GGIcolormap==NULL)
		return -1;

	err=ioctl(VIS_FD(visual),(int)RAMDAC_GETCLUT,_GGIcolormap);
	if (err<0) {
		fprintf(stderr,"generic-ramdac: Can't get default colormap\n");
		free(_GGIcolormap);
		return -1;
	}
		
	visual->opcolor->mapcolor=GGImapcolor;
	visual->opcolor->unmappixel=GGIunmappixel;
	visual->opcolor->setpalvec=GGIsetpalvec;
	visual->opcolor->getpalvec=GGIgetpalvec;

	return GGI_DL_OPCOLOR;
}

int GGIdlcleanup(ggi_visual *visual)
{
	free(_GGIcolormap);

	return EOK;
}
