/* 
   Graphics library for GGI. Mode management.

   Copyright (C) 1997 Jason McMullan - jmcc@cs.cmu.edu

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <dlfcn.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/mman.h>

#include "ggi-dl.h"
#include "internal.h"
#include <ggi/ggi_commands.h>

int GGIsetfocus(ggi_visual *vis)
{
	return EOK;
}

/*
 * _Attempt_ to get the default framebuffer.. 
 */
static const char *_GGIgetmmap(ggi_visual *vis,ggi_mode *mode)
{
	int size;
	const char *cp=NULL;

	if (VIS_FB_LINEAR(vis)!=NULL)
		free(VIS_FB_LINEAR(vis));

	switch (mode->graphtype) {
		case GT_TEXT16:
		case GT_TEXT32:
			if (mode->graphtype==GT_TEXT16)
				cp="generic-text-16";
			else
				cp="generic-text-32";
			break;
		case GT_1BIT:
			cp="generic-linear-1";
			break;
		case GT_4BIT:
			cp="generic-linear-4";
			break;
		case GT_8BIT:
			cp="generic-linear-8";
			break;
		case GT_15BIT:
			cp="generic-linear-15";
			break;
		case GT_16BIT:
			cp="generic-linear-16";
			break;
		case GT_24BIT:
			cp="generic-linear-24";
			break;
		case GT_32BIT:
			cp="generic-linear-32";
			break;
		default:
			fprintf(stderr,"Unknown bit size used!\n");
			VIS_FB_LINEAR(vis)=NULL;
			return NULL;
	}

	size = _ggiSetupMode(vis);
	if (size <= 0)
		return cp;

	VIS_FB_LINEAR_SIZE(vis)=size;	
	VIS_FB_LINEAR(vis)=malloc(size);

	if (VIS_FB_LINEAR(vis) == NULL) {
		VIS_FB_LINEAR_SIZE(vis)=0;
		DPRINT("Failed to mmap a %dK area\n",size/1024);
		return NULL;
	}
	
	return cp;
}

static int _GGIdomode(ggi_visual *vis,ggi_mode *mode)
{
	int err;
	ggi_suggest sug;
	const char *name;

	_ggiZapMode(vis,~GGI_DL_OPDISPLAY);

	name=_GGIgetmmap(vis,mode);
	if (name==NULL)
		return -1;

	err=(_ggiOpenDL(vis,"generic-stubs","")==NULL);
	if (err) {
		fprintf(stderr,"display-X: Can't open the \"generic-stubs\" "
				"library.\n");
	}

	sug.id=0;
	strcpy(sug.name,name);
	sug.args[0]=0;
	sug.args[1]=0;
	err=(_ggiOpenDL(vis,sug.name,sug.args)==NULL);
	if (err) {
		fprintf(stderr,"display-memory: Can't find an appropriate "
			       "library for %s (%s)\n",
				sug.name,sug.args);
		return err;
	} else {
		DPRINT("Success in loading %s (%s)\n",sug.name,sug.args);
	}

	return EOK;
}

int GGIsetmode(ggi_visual *vis,ggi_mode *tm)
{ 
	int err;

	if (vis==NULL) {
		DPRINT("Visual==NULL\n");
		return -1;
	}
	
	/* some elements of the mode setup rely on this. */
	memcpy(VIS_MODE(vis),tm,sizeof(ggi_mode));
	err=_GGIdomode(vis,tm);

	if (err)
		return err;

	memcpy(VIS_MODE(vis),tm,sizeof(ggi_mode));

	return EOK;
}

/**********************************/
/* check any mode (text/graphics) */
/**********************************/
int GGIcheckmode(ggi_visual *vis,ggi_mode *tm)
{
	if (vis==NULL)
		return -1;

	return EOK;	
}

/************************/
/* get the current mode */
/************************/
int GGIgetmode(ggi_visual *vis,ggi_mode *tm)
{
	DPRINT("In GGIgetmode(%p,%p)\n",vis,tm);
	if (vis==NULL)
		return -1;
	
	if (VIS_MODE(vis)==NULL)
		return -1;

	memcpy(tm,VIS_MODE(vis),sizeof(ggi_mode));

	return EOK;
}
