/* 
   Graphics library for GGI. Mode management.

   Copyright (C) 1997 Jason McMullan - jmcc@cs.cmu.edu

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <dlfcn.h>
#include <sys/ioctl.h>

#include "internal.h"
#include <ggi/ggi_commands.h>

ggi_mode ggiDefaultMode={
	1,              /* 1 frame [???] */
	{320,200},      /* Default size */
	{320,200},	/* Virtual */
	{300,220},	/* mm don't care */
	GT_8BIT,	/* Mode */
	{8,16},		/* Font size */
	/*0*/
};
                                                                                                                                                  
static void _ggiCheck4Defaults(ggi_mode *tm)
{
#define DOCHECK(what)	if (tm->what == GGI_AUTO) tm->what=ggiDefaultMode.what

	DOCHECK(visible.x);
	DOCHECK(visible.y);
	if (tm->virt.x == GGI_AUTO) tm->virt.x=tm->visible.x;
	if (tm->virt.y == GGI_AUTO) tm->virt.y=tm->visible.y;
	DOCHECK(text.x);
	DOCHECK(text.y);
	DOCHECK(graphtype);
}

/********************************/
/* set any mode (text/graphics) */
/********************************/
int ggiSetMode(ggi_visual *vis,ggi_mode *tm)
{ 
	int retval;
	_ggiLock(vis);

	_ggiCheck4Defaults(tm);
	DPRINT("setmode (vis %dx%d virt %dx%d) - calling\n",
	       tm->visible.x,tm->visible.y,tm->virt.x,tm->virt.y);
	retval=vis->opdisplay->setmode(vis,tm);
	if (retval<0) {
		fprintf(stderr,"libggi: Failed to set mode (%dx%d)\n",
				tm->visible.x,tm->visible.y);
		goto ender;
	}

	DPRINT("setorigin\n");
	ggiSetOrigin(vis,0,0);
	DPRINT("setsplit\n");
	ggiSetSplitline(vis,tm->visible.y);
	DPRINT("setfg\n");
	ggiSetGCForeground(vis,0);
	DPRINT("fillscreen\n");
	ggiFillscreen(vis);
	DPRINT("setmode - success\n");

ender:

	DPRINT("setmode - done\n");
	_ggiUnlock(vis);
	return retval;
}

/**********************************/
/* check any mode (text/graphics) */
/**********************************/
int ggiCheckMode(ggi_visual *vis,ggi_mode *tm)
{
	_ggiCheck4Defaults(tm);
	return vis->opdisplay->checkmode(vis,tm);
}

/************************/
/* get the current mode */
/************************/
int ggiGetMode(ggi_visual *vis,ggi_mode *tm)
{
	return vis->opdisplay->getmode(vis,tm);
}

/******************/
/* set a textmode */
/******************/
int ggiSetTextMode(ggi_visual *vis,int cols,int rows,
				   int vcols,int vrows,
				   int fontsizex,int fontsizey,
				   ggi_graphtype mode)
{
  ggi_mode tim={	1,		/* 1 frame [???] */
			{800,600},	/* Default size */
        	        {1024,1024},
        	        {300,220},
                        GT_TEXT16,
                	{8,16},
                	/*0*/           };
                        
  if (fontsizex<4||fontsizex>32) return -EINVAL;	/* Maybe some card can do it ... */
  if (fontsizey<4||fontsizey>32) return -EINVAL;

  tim.virt.x=vcols*fontsizex;
  tim.virt.y=vrows*fontsizey;
  tim.visible.x=cols*fontsizex;
  tim.visible.y=rows*fontsizey;
  tim.text.x=fontsizex;
  tim.text.y=fontsizey;
  tim.graphtype=mode;

  return ggiSetMode(vis,&tim);
}

/*************************/
/* check a text mode */
/*************************/
int ggiCheckTextMode(ggi_visual *vis,int cols,int rows,
				     int vcols,int vrows,
				     int fontsizex,int fontsizey,
				     ggi_graphtype mode,
				     ggi_mode *md,...)
{
  int rc;
  ggi_mode tim={	1,		/* 1 frame [???] */
			{800,600},	/* Default size */
        	        {1024,1024},
        	        {300,220},
                        GT_TEXT16,
                	{8,16},
                	/*0*/           };
                        
  if (fontsizex<4||fontsizex>32) return -EINVAL;	/* Maybe some card can do it ... */
  if (fontsizey<4||fontsizey>32) return -EINVAL;

  tim.virt.x=vcols*fontsizex;
  tim.virt.y=vcols*fontsizex;
  tim.visible.x=cols*fontsizex;
  tim.visible.y=rows*fontsizey;
  tim.text.x=fontsizex;
  tim.text.y=fontsizey;

  rc=ggiCheckMode(vis,&tim);
  if (md) *md=tim;	/* give back the mode if asked for. */
  return rc;
}

/***********************/
/* set a graphics mode */
/***********************/
int ggiSetGraphMode(ggi_visual *vis,int xsize,int ysize,
		    int xvirtual,int yvirtual,ggi_graphtype type)
{
  ggi_mode tim={ 1,
		 { 800, 600},
	         {1024,1024},
	         { 300, 220},
	         GT_8BIT,
	         {8,16},
	         /*0*/      };
  tim.visible.x=xsize;
  tim.visible.y=ysize;
  tim.virt.x=xvirtual;
  tim.virt.y=yvirtual;                       
  tim.graphtype=type;

  return ggiSetMode(vis,&tim);
}

/*************************/
/* check a graphics mode */
/*************************/
int ggiCheckGraphMode(ggi_visual *visual,int xsize,int ysize,
		      int xvirtual,int yvirtual,ggi_graphtype type,
		      ggi_mode *md,...)
{
  int rc;
  ggi_mode tim={ 1,
		 { 800, 600},
	         {1024,1024},
	         { 300, 220},
	         GT_8BIT,
	         {8,16},
	         /*0*/      };
  tim.visible.x=xsize;
  tim.visible.y=ysize;
  tim.virt.x=xvirtual;
  tim.virt.y=yvirtual;                       
  tim.graphtype=type;

  rc=ggiCheckMode(visual,&tim);
  if (md) *md=tim;	/* give back the mode if asked for. */
  return rc;
}

/*
 * Setup the vis->info.fb stuff, from vis->mode
 *
 * Provides initial guesses for:
 *     VIS_FB_WIDTH
 *     VIS_FB_HEIGHT
 *     VIS_FB_BPP	bits/pixel 
 *     VIS_FB_BPC	bits for color/pixel
 *     VIS_FB_BPA	bits for alpha/pixel 
 *
 * NOTE: These can be overriden by each fb driver.
 */
int _ggiSetupMode(ggi_visual *vis)
{
	int size;

	VIS_FB_WIDTH(vis)=VIS_MODE(vis)->virt.x;
	VIS_FB_HEIGHT(vis)=VIS_MODE(vis)->virt.y;

	size=VIS_FB_WIDTH(vis)*VIS_FB_HEIGHT(vis);

	VIS_FB_BPP(vis)=0;
	VIS_FB_BPC(vis)=0;
	VIS_FB_BPA(vis)=0;

	switch (VIS_MODE(vis)->graphtype) {
		case GT_TEXT16:
		case GT_TEXT32:
			size /= (VIS_MODE(vis)->text.x * VIS_MODE(vis)->text.y);
			size *= (VIS_MODE(vis)->graphtype==GT_TEXT16) ? 2 : 4;
			VIS_FB_WIDTH(vis) /= VIS_MODE(vis)->text.x;
			VIS_FB_HEIGHT(vis) /= VIS_MODE(vis)->text.y;
			VIS_FB_BPP(vis) = (VIS_MODE(vis)->graphtype == GT_TEXT16) ? 16 : 32;
			VIS_FB_BPC(vis) = VIS_FB_BPP(vis);
			break;
		case GT_1BIT:
			size /= 8;
			VIS_FB_BPP(vis)=1;
			VIS_FB_BPC(vis)=1;
			break;
		case GT_4BIT:
			size /= 2;
			VIS_FB_BPP(vis)=4;
			VIS_FB_BPC(vis)=4;
			break;
		case GT_8BIT:
			VIS_FB_BPP(vis)=8;
			VIS_FB_BPC(vis)=8;
			break;
		case GT_15BIT:
			size *= 2;
			VIS_FB_BPP(vis)=16;
			VIS_FB_BPC(vis)=15;
			VIS_FB_BPA(vis)=1;
			break;
		case GT_16BIT:
			size *= 2;
			VIS_FB_BPP(vis)=16;
			VIS_FB_BPC(vis)=16;
			break;
		case GT_24BIT:
			size *= 3;
			VIS_FB_BPP(vis)=24;
			VIS_FB_BPC(vis)=24;
			break;
		case GT_32BIT:
			size *= 4;
			VIS_FB_BPP(vis)=32;
			VIS_FB_BPC(vis)=32;
			break;
		default:
			fprintf(stderr,"Unknown bit size used!\n");
			return -1;
	}

	size = ((size >> 16) + ((size & 0xFFFF) ? 1 : 0)) << 16;

	return size;
}
