/*
 *  Input library for GGI.
 *
 *  Copyright (C) 1997 GGI Project
 *
 *  If you do any modifications, I would like you to send diffs to me
 *  to allow for collecting a more and more complete set of drivers and
 *  to improve existing ones.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <ggi/ggi.h>

/*
** This is a list of all known gii_types. Extend this as need arises.
*/

enum gii_desctype {

	DT_EMPTY,             /* Only for testing purposes. Has no features. */
	DT_KEY,
	DT_BUTTON,
	DT_SWITCH,
	DT_SELECTOR,
	DT_SLIDER,
	DT_AXIS,
	DT_MENU,
	DT_DATA,
	DT_VR1,
	DT_VR2,
	DT_LASTPREALLOCED,

	DT_PRIVATE=0x1000000, /* If you want to roll your own you can start from here */
	DT_LAST   =0x2000000  /* Enough private space ? */

};

#define FEATURE_MASK   0xff000000
#define FEATUREMASK(x) ( (x) & FEATURE_MASK )

#define FORCE_MASK   0x00ffffff
#define FORCE_MAX    FORCE_MASK
#define FORCEMASK(x) ( (x) & FORCE_MASK  )

enum gii_featuremask {

	/* Generic thing for returning "no match" */
	NONE         = 0x00000000,

	/* First for BUTTON type events */
	PRESSED      = 0x01000000,
	RELEASED     = 0x02000000,
	
	/* Now for AXIS type */
	POS_DIR      = 0x01000000,
	NEG_DIR      = 0x02000000,
	
	/* Now for MENU */
	SELECT_NEXT  = 0x01000000,
	SELECT_PREV  = 0x02000000,
	SELECT_FIRST = 0x03000000,
	SELECT_LAST  = 0x04000000,
	EXECUTE      = 0x05000000,
	SUBMENU      = 0x06000000,
	ESCAPE       = 0x06000000,
};

enum gii_controltype {
	
	KEY_PRESS,
	KEY_REPEAT,
	KEY_RELEASE,
	
	AXIS_HASVALUE,
	AXIS_RANGE,

	MAX_CTRLTYPE
	
};

struct ctd_key {

	uint8 device;   /* 0 = don't care */
	int sym,code;   /* 0 = don't care */

};

struct ctd_axis_has {

	uint8 device;		/* 0 = don't care */
	uint32 axisnum;
	uint32 value,tolerance;

};

struct gii_control;

typedef struct control_ops {

	const  char *drivername;	/* DLL responsible for this control */
	uint32 (*check)         (struct gii_control *ctrl,ggi_event *event);
	int    (*detect)        (struct gii_control *ctrl,ggi_event *event);
	int    (*write_pvtdata) (struct gii_control *ctrl,char *string,int maxlen);
	int    (*read_pvtdata)  (struct gii_control *ctrl,const char *string);
	void   (*free_pvtdata)  (struct gii_control *ctrl);
	int    (*get_name)      (struct gii_control *ctrl,char *string,int maxlen);

} control_ops;

typedef struct gii_control {

	struct gii_control *next;   /* linked list */
	control_ops	*cops;
	void *pvtdata;

} gii_control;

typedef struct gii_feature {

	struct gii_feature *next;   /* linked list */
	enum gii_featuremask type;
	char name[16];              /* What does this feature do ? */
	gii_control *ctrl;

} gii_feature;

typedef struct ggi_descriptor {

	enum gii_desctype type;
	char name[64];              /* Used for loading/saving */
	gii_feature *feat;          /* This is a double-linked list of features */

} gii_descriptor;
