/* -----------------------------------------------------------------------------
 *	Interface to the machine independent Linux/m68k keyboard driver
 * -----------------------------------------------------------------------------
 */

    /*
     *  Filled in by mach_keyb_init()
     */

unicode *key_maps[MAX_NR_KEYMAPS];

void handle_scancode(unsigned char scancode)
{
	ggi_event event;		/* struct for event reporting	*/
	uchar  keycode, up_flag, shift;
	ushort keysym, *keymap;
	struct kbd_state *mod;
	struct kgi_device *focus;

	ASSERT(KGI_VALID_FOCUS(kbd.input));

	if (KGI_IS_FOCUSED(kbd.input)) {

		focus = kgidev[kbd.input.focus];
		mod = &(focus->kbd_state);

	} else {

		focus = NULL;
		mod = &default_state;
	}

	if (focus && (focus->event_mask & emKeyScanCode)) {

		event.type = evKeyScanCode;
		event.kbd.time	= jiffies;

		event.kbd.keycode = scancode;
		event.kbd.keysym  = 0;
		event.kbd.norm	= mod->norm;
		event.kbd.lock	= mod->lock;
		event.kbd.slock	= mod->slock;

		(focus->handle_event)(focus, &event);
	}

	/*	Convert scancode to keycode, using prev_scancode.
	 */

	up_flag = (scancode & 0x80);
	scancode &= 0x7f;
	keycode = scancode;

	/*
	 *	Note: keycode must not be 0. Now we track up/down status.
	 *
	 *	++Geert: on m68k 0x00 is a valid keycode
	 */

	if (up_flag) {

		rep = 0;
		if (!clear_bit(keycode, key_down)) {
		}

	} else {

		rep = set_bit(keycode, key_down);
	}

	/*
	 *	Now convert to keysym and do the action. 
	 */

	shift  = mod->norm ^ mod->lock ^ mod->slock;
	keymap = key_maps[shift];
	keysym = keymap ? keymap[keycode] : U(K_VOID);

	/*	rarely needeed, but if you need to debug the kbd code,
	**	uncomment the printk and get a second monitor!
	**
	**  printk(KERN_ERR "kbd%i->dev%i: %.2x, %.2x, %.4x, %s, "\
	**	" (norm %.2x, lock %.2x, slock %.2x)\n",\
	**	kbd.id, kbd.input.focus, keycode, shift, keysym,\
	**	up_flag ? "up":"down", mod->norm, mod->lock, mod->slock));
	*/

	if (focus) {

		event.type = up_flag	? evKeyRelease
					: (rep ? evKeyRepeat : evKeyPress);

		if (focus->event_mask & (1 << event.type)) {

			event.kbd.time = jiffies;
			event.kbd.keycode = keycode;
			event.kbd.keysym  = keysym;
			event.kbd.norm	= mod->norm;
			event.kbd.lock	= mod->lock;
			event.kbd.slock	= mod->slock;

			(focus->handle_event)(focus, &event);
		}
	}

	do_action(keysym, up_flag);
}

/*	This routine runs with all interrupts enabled. It does console changing,
 *	led setting and copy_to_cooked, which can take a reasonably long time.
 *
 * 	Aside from timing (which isn't really that important for keyboard
 *	interrupts as they happen often), using the software interrupt routines
 *	for this thing allows us to easily mask this when we don't want any of
 *	the above to happen. Not yet used, but this allows for easy and
 *	efficient race-condition prevention later on.
 */
static void kbd_bottomhalf(void)
{
	static int ledstate = -1;
	ggi_uint leds;

	ASSERT(KGI_VALID_FOCUS(kbd.input));
	
	leds = KGI_IS_FOCUSED(kbd.input) 
		? kgidev[kbd.input.focus]->kbd_state.leds : 0xFF;

	if (leds != ledstate) {

		ledstate = leds;

		if (mach_kbd_leds)
                        mach_kbd_leds(leds);
	}

	ASSERT((want_focus < 0) || ((0 <= want_focus) && \
		(want_focus < MAX_NR_DEVICES)));

	if (   (0 <= want_focus) && (want_focus < MAX_NR_DEVICES)
	    && (kgidev[want_focus])) {

		struct kbd_state *mod = KGI_IS_FOCUSED(kbd.input)
			? &(kgidev[kbd.input.focus]->kbd_state)
			: &default_state;

		kgidev[want_focus]->kbd_state.norm = mod->norm;
		kgidev[want_focus]->kbd_state.lock = mod->lock;
		kgidev[want_focus]->kbd_state.slock = mod->slock;

		kgi_focus(kbd.id, want_focus);
	}

	want_focus = -1;

/* !!! check 2.1 bottom half! */
}

/*	Generates sound of frequency <pitch> for <duration> msec and
 *	loudness <volume>.
 *
 *	If pitch is 0, will turn off sound, else will turn it on for that time.
 *	If duration is 0, will return immediately, else sleep for msec time,
 *	then turn sound off.
 */

static void __nosound(struct kgi_keyboard *k)
{
	/* Dummy */
}

static void dummy_mksound(unsigned int say, unsigned int nothing)
{
	/* We are real quiet in here ;-) */
}

/*
 *  Since this is machine-dependent, systems with a sound-device should
 *  change this value in arch/m68k/{amiga,atari,x}/config.c.
 */
void (*kd_mksound)(unsigned int hz, unsigned int ticks) = dummy_mksound;

static void __mksound(struct kgi_keyboard *kbd, ggi_uint pitch, ggi_uint duration)
{
	int ticks;

	if (duration) {
		ticks = (HZ*duration + 500)/1000;
		if (!ticks)
			ticks = 1;
        } else
		ticks = 0;
	kd_mksound(pitch, ticks);
}

static void __setleds(struct kgi_keyboard *k)
{
	mark_bh(KEYBOARD_BH);
}

/*
 *	initialization
 */

#ifdef MODULE
int init_module(void)
#else
int kbd_init(void)
#endif
{
	mach_keyb_init();

	want_focus = -1;

	kgi_register_keyboard(&kbd, -1);

	init_bh(KEYBOARD_BH, kbd_bottomhalf);
	mark_bh(KEYBOARD_BH);

	printk(KERN_INFO "%s %s driver\n", kbd.input.manufact, kbd.input.model);
	return EOK;
}

#ifdef MODULE

void cleanup_module(void)
{
	kgi_unregister_keyboard(&kbd);

	#error need to do more here!
}

#endif
