/* ----------------------------------------------------------------------------
**	xterm console parser
** ----------------------------------------------------------------------------
**
**	Copyright (C) 1996-1997, Steffen Seeger [seeger@physik.tu-chemnitz.de]
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2, or (at your option)
**	any later version.
**
**	You should have received a copy of the GNU General Public License
**	along with this program; see the file COPYING.  If not, write to
**	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
**
** -----------------------------------------------------------------------------
*/

#define	MAINTAINER	"Steffen Seeger, [seeger@physik.tu-chemnitz.de]"

#undef	STRICT_XTERM	/* #define to restrict to original XTERM	*/
#undef	SETEC_ASTRONOMY	/* #define to force display of concealed text	*/

/* ========================================================================== *\
|*  ---===>>> THERE IS NO USER-SERVICABLE CODE BELOW THIS LINE :-) <<<===---  *|
\* ========================================================================== */

#undef	DEBUG		/* enable debugging code only on explicit request */
#undef	DEBUG_XTERM

#include <linux/config.h>
#include <linux/version.h>
#include <linux/ascii.h>
#include <linux/console.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/sched.h>
#include <linux/interrupt.h>

#define	NEED_CONSOLE_DEFS
#include <ggi/kgi.h>

#ifdef	DEBUG_XTERM

	#undef	TRACE
	#define	TRACE(x)	x
#endif

#if LINUX_VERSION_CODE >= LinuxVersionCode(2,1,0)
#include <asm/uaccess.h>
#endif

static uchar DEFAULT_COLORS[8] = 
{	COL_LIGHTGRAY,	COL_BLACK,		/* normal	*/
	COL_LIGHTRED,	COL_BLACK,		/* blink	*/
	COL_LIGHTBLUE,	COL_BLACK,		/* underline	*/
};

enum color_index
{	COL_FG,		COL_BG,
	COL_BLINK_FG,	COL_BLINK_BG,
	COL_UL_FG,	COL_UL_BG
};

static uchar colormap[16] =
{
	COL_BLACK,	COL_RED,	COL_GREEN,	COL_BROWN,
	COL_BLUE,	COL_MAGENTA,	COL_CYAN,	COL_LIGHTGRAY,
	COL_DARKGRAY,	COL_LIGHTRED,	COL_LIGHTGREEN,	COL_YELLOW,
	COL_LIGHTBLUE,	COL_LIGHTMAGENTA,COL_LIGHTCYAN,	COL_WHITE
};

#define	DEFAULT_FG	(DEFAULT_COLORS[COL_FG])
#define	DEFAULT_BG	(DEFAULT_COLORS[COL_BG])

#define CTRL_ACTION 0x0d00ff81	/* chars<32 that invoke special action	*/
#define	CTRL_ALWAYS 0x0800f581	/* can not be overridden by DISP_CTRL	*/

/*	forward declaration of the font translation tables. See the end of this
**	file for the initializer.
*/
static unicode translation[4][256];

static void xterm_set_charset(struct kgi_console *cons, ggi_uint nr, char c)
{
	switch(c) {

	case 'K':		/* ??? original console code did this?	*/
	case 'A':		/* United Kingdom/national		*/
		G[nr] = USER_MAP;
		break;

	case '1':		/* 'alternate ROM standard character'	*/
	case 'B':		/* US-ASCII (ISO-8859-1 = Latin1)	*/
		G[nr] = LAT1_MAP;
		break;

	case '2':		/* 'alternate ROM graphics'		*/
	case '0':		/* VT100 graphics			*/
		G[nr] = GRAF_MAP;
		break;


	case 'U':		/* ??? original console code did this?	*/
		G[nr] = IBMPC_MAP;
		break;
	}

	if (CHARSET == nr) {

		TRANSLATE = translation[G[nr]];
	}
}

enum {	cursor, id, status, 
	mouseup, 
	mousedown1, mousedown2, mousedown3
};

static void xterm_do_set_mode(struct kgi_console *cons, int mode, int on)
{
	if (mode == auto_repeat) {

		if (on) {

			DEV.event_mask |= emKeyRepeat;

		} else  {

			DEV.event_mask &= ~emKeyRepeat;
		}
	}

	if (mode == reverse_video) {

		if (!MODE(reverse_video) != !on) {

			#warning invert alternate screen too!

			fb_invert(FB, 0, FB_SIZE - 1);
			scroll_modified(cons, ORG, ORG + FSIZE);
			ATTRFL ^= ATTR_REVERSE;
		}
	}


#if 0	/* !!! implement alternate screen	*/
	if ((mode == alt_screen) && (!MODE(alt_screen) != !on)) {

		int cnt = FB_SIZE_X*SIZE_Y;

		if (MODE(alt_screen)) {

			if (ORG + cnt < FB_SIZE) {

				fb_memcpy_tofb(FB, ALT_SCREEN, ORG, cnt);

           		} else {

				fb_memcpy_tofb(FB, ALT_SCREEN, ORG,
					FB_SIZE-ORG);
				fb_memcpy_tofb(FB, ALT_SCREEN+FB_SIZE-ORG,
					0, ORG+cnt-FB_SIZE);
			}

		} else {

			if (ORG + cnt < FB_SIZE) {

				fb_memcpy_fromfb(FB, ALT_SCREEN, ORG, cnt);

           		} else {

				fb_memcpy_fromfb(FB, ALT_SCREEN, ORG,
					FB_SIZE - ORG);
				fb_memcpy_fromfb(FB, ALT_SCREEN + FB_SIZE-ORG,
					0, ORG+cnt-FB_SIZE);
			}
		}
	}
#endif

	/* !!!  if (mode == column132 && MODE(allow_132mode))
	 * !!!    con_resize(cons, SIZE_Y, on ? 132 : 80);
	 */

	if (on) {

		SET_MODE(mode);

	} else {
	
		CLEAR_MODE(mode);
	}

	scroll_update_attr(cons);

	if (mode == origin) {

		scroll_gotoxy(cons, 0,0);
	}
}

static int xterm_par_to_mode(register int ques, register int par)
{
	if (ques) {

		switch (par) {

			case    1:	return cursor_key;
			case    3:	return column132;
			case    4:	return smooth_scroll;
			case    5:	return reverse_video;
			case    6:	return origin;
			case    7:	return auto_wrap;
			case    8:	return auto_repeat;
			case    9:	return x10_mouse;
			case   38:	return textronixmode;
			case   40:	return allow_132mode;
			case   41:	return curses_fix;
			case   44:	return margin_bell;
			case   45:	return reverse_wrap;
			case   46:	return logging;
			case   47:	return alt_screen;
			case 1000:	return report_mouse;
			case 1001:	return track_mouse;
		}

	} else {

		switch (par) {

			case    3:	return display_ctrl;
			case    4:	return insert;
			case   20:	return newline;
		}
	}

	return -1;
}

static void xterm_set_mode(struct kgi_console *cons, int on)
{
	int i, mode;

	for (i = 0; i <= XNPAR; i++) {

		if ((mode = xterm_par_to_mode(FLAG(question), PAR[i])) >= 0) {

			xterm_do_set_mode(cons, mode, on);
		}
	}

	kgi_hide_gadgets(&DEV);
	DEV.state &= ~(KGI_CURSOR_TO_SHOW | KGI_POINTER_TO_SHOW);

	if (MODE(show_cursor)) {

		DEV.state |= KGI_CURSOR_TO_SHOW;
	}

	if (MODE(report_mouse) | MODE(x10_mouse) | MODE(track_mouse)) {

		DEV.state |= KGI_POINTER_TO_SHOW;
		DEV.event_mask |= emPtrMove | emPtrButton;

	} else {

		DEV.event_mask &= ~emPointer;
	}
}

static void xterm_backup_mode(struct kgi_console *cons, int save)
{
	int i, mode;

	for (i = 0; i <= XNPAR; i++) {

		if ((mode = xterm_par_to_mode(FLAG(question), PAR[i])) >= 0) {

			if (save) {

				SAVE_MODE(mode);

			} else {	

				if (MODE(mode) != SMODE(mode)) {

					xterm_do_set_mode(cons, mode,
						SMODE(mode));
				}
			}
		}
	}

	kgi_hide_gadgets(&DEV);
	DEV.state &= ~(KGI_CURSOR_TO_SHOW | KGI_POINTER_TO_SHOW);

	if (MODE(show_cursor)) {

		DEV.state |= KGI_CURSOR_TO_SHOW;
	}

	if (MODE(report_mouse) | MODE(x10_mouse) | MODE(track_mouse)) {

		DEV.state |= KGI_POINTER_TO_SHOW;
		DEV.event_mask |= emPtrMove | emPtrButton;

	} else {

		DEV.event_mask &= ~emPointer;
	}
}

static void xterm_save_cur(struct kgi_console *cons)
{
	int i;

	#warning save reverse_mode too!

	S_X	= X;
	S_Y	= Y;
	S_ATTRFL = ATTRFL;
	S_CHARSET = CHARSET;

	for (i = 0; i < sizeof(S_G); i++) {

		S_G[i]	= G[i];
	}

	for (i = 0; i < sizeof(S_COLORS); i++) {

		S_COLORS[i] = COLORS[i];
	}
}


static void xterm_restore_cur(struct kgi_console *cons)
{
	int i;

	X	= S_X;
	Y	= S_Y;
	ATTRFL	= S_ATTRFL;	
	CHARSET	= S_CHARSET;

	for (i = 0; i < sizeof(G); i++) {

		G[i]	= S_G[i];
	}

	for (i = 0; i < sizeof(COLORS); i++) {

		COLORS[i] = S_COLORS[i];
	}

	TRANSLATE = translation[G[CHARSET]];
	scroll_gotoxy(cons, X, Y);
	scroll_update_attr(cons);
	CLEAR_FLAG(need_wrap);
}


static void xterm_set_term(struct kgi_console *cons)
{
	switch(PAR[0]) {

	case 10:

		if (XNPAR >= 1) {

			BELL_PITCH = ((PAR[1]<20) || (PAR[1]>32767))
					 ? 0 : PAR[1];
		} else {
	
			BELL_PITCH = 750;
		}
		break;

	case 11:

		if (XNPAR >= 1) {

			BELL_DURATION = (PAR[1] < 2000) ?  PAR[1] : 0;

		} else {

			BELL_DURATION = 125;
		}
		break;
	}
}

static void csi_m(struct kgi_console *cons)
{
	int i;

	for (i = 0; i <= XNPAR; i++) {

		switch (PAR[i]) {

		case 0:
			ATTRFL =  ATTR_NORMAL
				| ATTR_COLOR(DEFAULT_FG, DEFAULT_BG);
			break;

		case 1:
			ATTRFL &= ~ATTR_INTENSITY;
			ATTRFL |= ATTR_BOLD;
			break;

#ifndef STRICT_XTERM

		case 2:
			ATTRFL &= ~ATTR_INTENSITY;
			ATTRFL |= ATTR_HALF;
			break;
#endif

		case 4:
			ATTRFL |= ATTR_UNDERLINE;
			break;

		case 5:
			ATTRFL |= ATTR_BLINK;
			break;

		case 7:
			if (MODE(reverse_video)) {

				ATTRFL &= ~ATTR_REVERSE;

			} else {

				ATTRFL |= ATTR_REVERSE;
			}
			break;

#ifndef STRICT_XTERM
	#ifndef SETEC_ASTRONOMY
		case 8:
			ATTRFL &= ~ATTR_FGCOLOR;
			ATTRFL |= (ATTRFL << 8) & ATTR_FGCOLOR;
			break;
	#endif
		case 10:
			/*	ANSI X3.64-1979 (SCO-ish?)
			**	Select primary font, don't display
			**	control chars if defined, don't set
			**	bit 8 on output.
			*/
			TRANSLATE = translation[G[CHARSET]];
			CLEAR_MODE(display_ctrl);
			CLEAR_MODE(toggle_meta);
			break;

		case 11:
			/*	ANSI X3.64-1979 (SCO-ish?)
			**	Select first alternate font, let's
			**	chars < 32 be displayed as ROM chars.
			*/
			TRANSLATE = translation[IBMPC_MAP];
			SET_MODE(display_ctrl);
			CLEAR_MODE(toggle_meta);
			break;

		case 12:
			/*
			**	ANSI X3.64-1979 (SCO-ish?)
			**	Select second alternate font, toggle
			**	high bit before displaying as ROM char.
			*/
			TRANSLATE = translation[IBMPC_MAP];
			SET_MODE(display_ctrl);
			SET_MODE(toggle_meta);
			break;
#endif

		case 21:
		case 22:
			ATTRFL &= ~ATTR_INTENSITY;
			ATTRFL |= ATTR_NORMAL;
			break;

		case 24:
			ATTRFL &= ~ATTR_UNDERLINE;
			break;

		case 25:
			ATTRFL &= ~ATTR_BLINK;
			break;

		case 27:
			if (MODE(reverse_video)) {

				ATTRFL |= ATTR_REVERSE;

			} else {

				ATTRFL &= ~ATTR_REVERSE;
			}
			break;

#ifndef STRICT_XTERM

		case 38:
			/*	ANSI X3.64-1979 (SCO-ish?)
			**	Enables underscore, white foreground
			**	with white underscore (Linux - use
			**	default foreground).
			*/
			ATTRFL |= ATTR_UNDERLINE;

			ATTRFL &= ~0xFF00;
			ATTRFL |= COL_WHITE << 8;
			break;

		case 39:
			/*
			**	ANSI X3.64-1979 (SCO-ish?)
			** ???	Disable underline option. Reset color
			** ???	to default?
			**	It did this before:
			*/
			ATTRFL &= ~ATTR_UNDERLINE;

			ATTRFL &= ~0xFF00;
			ATTRFL |= DEFAULT_FG << 8;

			break;	

		case 49:

			ATTRFL &= ~0x00FF;
			ATTRFL |= DEFAULT_BG;
			break;
#endif

		default:
			if ((PAR[i] >= 30) && (PAR[i] <= 37)) {

				ATTRFL &= ~ATTR_FGCOLOR;
				ATTRFL |= ((ulong) colormap[PAR[i]-30]) << 8; 

			} else {

				if ((PAR[i] >= 40) && (PAR[i] <= 47)) {	

					ATTRFL &= ~ATTR_BGCOLOR;
					ATTRFL |= ((ulong) colormap[PAR[i]-40]);
				}
			}
			break;
		}
	}

	scroll_update_attr(cons);
}


void xterm_do_reset(struct kgi_console *cons, int do_reset)
{
	DEV.event_mask	= emKeyPress | emKeyRepeat;
#if 0
	if (! do_reset) {

		return;
	}
#endif
	if (DEV.state & KGI_MAPPED_ON_DISPLAY) {

		kgi_undo_gadgets(&DEV);
	}

	if (TTY) {

		TTY->winsize.ws_col = SIZE_X;
		TTY->winsize.ws_row = SIZE_Y;
	}

/* !!!	kbd_reset(&KBD_STATE);
*/
	kgi_set_pointer_window(&DEV,  0,DEV.mode.visible.x,
		0,DEV.mode.visible.y);

	SCROLL.mode	= 0;
	SET_MODE(show_cursor);
	SET_MODE(auto_wrap);
	SET_MODE(meta_esc);

	S_MODE		= SCROLL.mode;

	CLEAR_FLAG(question);
	CLEAR_FLAG(caps_lock);
	CLEAR_FLAG(caps_sticky);
	CLEAR_FLAG(num_lock);

	STATE		= ESnormal;
	UTF_CHAR	= 0;
	UTF_COUNT	= 0;

	TAB_STOP[0]	= 0x01010100;
	TAB_STOP[1]	= TAB_STOP[2] = TAB_STOP[3] = TAB_STOP[4] = 0x01010101;

	ORIGIN		= 0;
	ORG		= 0;
	TOP		= 0;
	BOTTOM		= SIZE_Y;

	BELL_PITCH	= 750 /* Hz */;
	BELL_DURATION	= 125 /* msec */;

	CHARSET		= 0;
	G[0]		= LAT1_MAP;
	G[1]		= GRAF_MAP;
	G[2]		= IBMPC_MAP; 
	G[3]		= USER_MAP;

	TRANSLATE	= translation[G[CHARSET]];

	XNPAR = 0;

	ATTRFL = ATTR_NORMAL | ATTR_COLOR(COL_LIGHTGRAY, COL_BLACK);
	memcpy(COLORS, DEFAULT_COLORS, sizeof(COLORS));

	scroll_update_attr(cons);
	scroll_gotoxy(cons, 0, 0);
	scroll_erase_display(cons, 2);
	scroll_update_gadgets(cons);
	scroll_sync(cons);

	if (DEV.state & KGI_MAPPED_ON_DISPLAY) {

		kgi_show_gadgets(&(DEV));
	}

/* !!!	kbd_setleds(DEV.kbd); */
}


#define	xterm_put_char(t, c)	tty_insert_flip_char(t, c, 0)

static void xterm_application_key(struct tty_struct *tty, uchar key, int mode)
{
	xterm_put_char(tty, ASCII_ESC);
	xterm_put_char(tty, mode ? 'O' : '[');
	xterm_put_char(tty, key);
}

static void xterm_function_key(struct kgi_console *cons, ggi_uint key)
{
	char *s = kgi_function_key_string(DEV.focus, key);

	if (s) {

		while (*s) {

			xterm_put_char(TTY, *(s++));
		}
	}
}

static const char *cur = "BDCA";
static const char *pad = "0123456789+-*/\015,.?";
static const char *app = "pqrstuvwxylSRQMnn?";

static void xterm_numpad_key(struct kgi_console *cons, uchar val,
			     uchar shift)
{

	if (MODE(applic_key) && !(shift & (1 << KG_SHIFT))) {

		xterm_application_key(TTY, app[val], 1);
		return;
	}

	if (!FLAG(num_lock)) {

		switch(K(KT_PAD, val)) {

		case K_PCOMMA:
		case K_PDOT:
			xterm_function_key(cons, KVAL(K_REMOVE));
			return;

		case K_P0:
			xterm_function_key(cons, KVAL(K_INSERT));
			return;

		case K_P1:
			xterm_function_key(cons, KVAL(K_SELECT));
			return;

		case K_P2:
			xterm_application_key(TTY, cur[KVAL(K_DOWN)], 0);
			return;

		case K_P3:
			xterm_function_key(cons, KVAL(K_PGDN));
			return;

		case K_P4:
			xterm_application_key(TTY, cur[KVAL(K_LEFT)], 0);
			return;

		case K_P6:
			xterm_application_key(TTY, cur[KVAL(K_RIGHT)], 0);
			return;

		case K_P7:
			xterm_function_key(cons, KVAL(K_FIND));
			return;

		case K_P8:
			xterm_application_key(TTY, cur[KVAL(K_UP)], 0);
			return;

		case K_P9:
			xterm_function_key(cons, KVAL(K_PGUP));
			return;
		}

	}

	xterm_put_char(TTY, pad[val]);
	if ((val == KVAL(K_PENTER)) && MODE(newline)) {

		xterm_put_char(TTY, ASCII_LF);
	}
}

static void xterm_report(struct kgi_console *cons, ggi_uint what)
{
	struct tty_struct *tty = TTY;
	char buf[40], *s = buf;
	char cbutton = ' ';
	#define VT100ID	"\033[?1;2c"	/* VT100 with advanced video option */
	#define VT102ID	"\033[?6c"

	#define MOUSEX	('!' + (char)(POINTER_X/TEXT_X))
	#define MOUSEY	('!' + (char)(POINTER_Y/TEXT_Y))

	switch (what) {

		case cursor:
			sprintf(buf, "\033[%hd;%hdR",
				Y + (MODE(origin) ? TOP+1 : 1), X+1);
			break;

		case id:
			s = VT102ID;
			break;
 			
		case status:
			s = "\033[0n";
			break;

		case mouseup:
			sprintf(buf, "\033[M#%c%c", MOUSEX, MOUSEY);
			break;

		case mousedown3:
			cbutton++;
		case mousedown2:
			cbutton++;
		case mousedown1:
			sprintf(buf, "\033[M%c%c%c", cbutton, MOUSEX, MOUSEY);
			break;
	}

	#undef MOUSEX
	#undef MOUSEY

	while (*s) {

		xterm_put_char(tty, *(s++));
	}

	tty_schedule_flip(tty);
}

void xterm_handle_event(struct kgi_device *dev, ggi_event *ev)
{
	struct kgi_console *cons = (void *) dev;
	
	if ((1 << ev->any.type) & emPointer) {

		switch (ev->any.type) {

		case evPtrButtonPress:

			if (MODE(report_mouse) | MODE(x10_mouse)) {

				if (ev->pbutton.button & 2) {

					xterm_report(cons,mousedown1);
				}
				if (ev->pbutton.button & 4) {

					xterm_report(cons,mousedown2);
				}
				if (ev->pbutton.button & 1) {

					xterm_report(cons,mousedown3);
				}
			}
			break;

		case evPtrButtonRelease:

			if (MODE(report_mouse)) {

				xterm_report(cons, mouseup);
			}
			break;

		case evPtrRelative:
		case evPtrAbsolute:
			if (ev->any.type == evPtrRelative) {
				P_X += ev->pmove.x;
				P_Y += ev->pmove.y;
			} else { /* evPtrAbsolute */
				P_X = ev->pmove.x;
				P_Y = ev->pmove.y;
			}

			P_X = (P_X < 0 ) ? 0 : (P_X >= DEV.mode.visible.x ) ? DEV.mode.visible.x - 1 : P_X;
			P_Y = (P_Y < 0 ) ? 0 : (P_Y >= DEV.mode.visible.y ) ? DEV.mode.visible.y - 1 : P_Y;

			/* !!! do selection! */
			if (DEV.state & KGI_DEVICE_IS_FOCUSED) {

				kgi_undo_gadgets(dev);
				kgi_set_pointer(dev, P_X, P_Y);
				kgi_show_gadgets(dev);

			} else {

				kgi_set_pointer(dev, P_X, P_Y);
			}
			break;
		}
	}

	if ((1 << ev->any.type) & (emKeyboard)) {

		unicode	sym = ev->key.sym;
		ggi_uint type = KTYP(sym);
		ggi_uint val  = KVAL(sym);

		if (!TTY) {

			return;
		}

		if (type >= 0xF0) {

			type -= 0xF0;

			switch(type) {

			case KT_LETTER:

				if (!FLAG(caps_lock) != !FLAG(caps_sticky)) {

					if (('a' <= val) && (val <= 'z')) {

						val -= 'a' - 'A';

					} else {

						if (('A'<=val) && (val<='Z')) {

							val += 'a' - 'A';
						}
					}
				}
				/* fall through to KT_LATIN! */

			case KT_LATIN:
				xterm_put_char(TTY, val);
				break;

			case KT_META:
				if (MODE(meta_esc)) {

					xterm_put_char(TTY, ASCII_ESC);
					xterm_put_char(TTY, val);

				} else {

					xterm_put_char(TTY, val | 0x80);
				}
				break;

			case KT_FN:
				xterm_function_key(cons, val);
				break;

			case KT_SPEC:

				switch (K(KT_SPEC, val)) {

				case K_ENTER:
					xterm_put_char(TTY, ASCII_CR);
					if (MODE(newline)) {

						xterm_put_char(TTY, ASCII_LF); 
					}
					break;

				case K_BREAK:
					tty_insert_flip_char(TTY, 0, TTY_BREAK);
					break;
					
				case K_CAPS:
					TOGGLE_FLAG(caps_lock);
/* !!!					kbd_toggle_led(&KBD_STATE, 
!!!						       KBD_LED_CAPSLOCK);
!!!					kbd_setleds(DEV.kbd);
*/					return;

				case K_NUM:
					if (MODE(applic_key)) {

						xterm_application_key(TTY,
							'P', 1);

					} else {

						TOGGLE_FLAG(num_lock);
/* !!!						kbd_toggle_led(&KBD_STATE,
!!!							       KBD_LED_NUMLOCK);
!!!						kbd_setleds(DEV.kbd);
*/						return;
					}
					break;

				case K_HOLD:
					if (TTY->stopped) {

						start_tty(TTY);

					} else {

						stop_tty(TTY);
					}
					return;

				case K_SCROLLFORW:
					scroll_forward(cons, 0);
					return;

				case K_SCROLLBACK:
					scroll_backward(cons, 0);
					return;

				case K_CAPSON:
					TOGGLE_FLAG(caps_sticky);
/* !!!					kbd_toggle_led(&KBD_STATE,
!!!						       KBD_LED_CAPSLOCK);
!!!					kbd_setleds(DEV.kbd);
*/					return;

				case K_BARENUMLOCK:
					TOGGLE_FLAG(num_lock);
/* !!!					kbd_toggle_led(&KBD_STATE, 
!!!						       KBD_LED_NUMLOCK);
!!!					kbd_setleds(DEV.kbd);
*/					return;

				default:
					return;
				}
				break;

			case KT_SHIFT:

				if (val == KVAL(K_NORMAL_SHIFT) &&
					FLAG(caps_lock)) {

					CLEAR_FLAG(caps_lock);
/* !!!					kbd_toggle_led(&KBD_STATE,
!!!						       KBD_LED_CAPSLOCK);
!!!					kbd_setleds(DEV.kbd);
*/				}
				return;

			case KT_PAD:
				xterm_numpad_key(cons, val, ev->key.normal);
				break;

			case KT_CUR:
				xterm_application_key(TTY, cur[val],
					MODE(cursor_key));
				break;

			case KT_DEAD:
				printk(KERN_ERR __FILE__
					": deadkeys not implemented.\n");
				return;
			}

		} else {

			if (MODE(utf8)) {

				if (sym < 0x80) {

					xterm_put_char(TTY, sym);

				} else if (sym < 0x800) {

					xterm_put_char(TTY, 0xC0 | (sym >> 6));
					xterm_put_char(TTY, 0x80 |(sym & 0x3F));

				} else {

					xterm_put_char(TTY, 0xE0 | (sym >> 12));
					xterm_put_char(TTY, 0x80 | (sym >> 6));
					xterm_put_char(TTY, 0x80 |(sym & 0x3F));
				}
				/*	UTF-8 is defined to up to 31 bits,
				**	but as the keyboard delivers unicode
				**	chars only we only need 16 bit here.
				*/
			}
		}

		wake_up(&keypress_wait);
		tty_schedule_flip(TTY);
	}
}

int xterm_do_write(struct kgi_console *cons, int from_user,
			  const char *buf, int count)
{
	ggi_uint n = 0;
	isochar c;		/* 8-bit character to process	*/
	isochar	tc;		/* translated character		*/
	ggi_uint printable;
	ggi_uint old_pos = WPOS;

	disable_bh(KEYBOARD_BH);

	while (!TTY->stopped && count) {

		#if LINUX_VERSION_CODE < LinuxVersionCode(2,1,0)

			c = from_user ? get_user(buf) : *buf;
		#else

			if (from_user) {

				if (get_user(c, buf)) {

					return -EFAULT;
				}

			} else {

				c = *buf;
			}
		#endif
		c &= 0xff;		/* !!! make positive */
		buf++; n++; count--;

		/* ???	shouldn't toggle_meta mode *change* bit 0x80 ?? */
		tc = TRANSLATE[MODE(toggle_meta) ? (c | 0x80) : c];

		printable = (c >= 32) || ((1 << c) & (MODE(display_ctrl)
						? ~CTRL_ALWAYS : ~CTRL_ACTION));

		if (printable && (STATE == ESnormal)) {

			if (FLAG(need_wrap)) {

				scroll_modified(cons, old_pos, WPOS);
				if (MODE(auto_wrap)) {

					scroll_lf(cons);
				}
				scroll_cr(cons);
				old_pos = WPOS;
			}

			if (MODE(insert)) {

				TRACE(printk("<INS>"));
				scroll_insert_chars(cons, 1);
			}

			TRACE(printk("%.2x %.8x     ", c, SCROLL.mode));
			scroll_write(cons, tc);
			continue;
		}


		/*	Control characters can be used in the *middle* of
		**	an escape sequence.
		*/
		switch (c) {

		case ASCII_BEL:

			TRACE(printk("<BEL>"));
			if (BELL_PITCH && BELL_DURATION) {

				scroll_mksound(cons, BELL_PITCH, BELL_DURATION);
			}
			continue;
	
		case ASCII_BS:

			TRACE(printk("<BS>"));

			scroll_modified(cons, old_pos, WPOS);
			scroll_bs(cons);
			old_pos = WPOS;
			continue;

		case ASCII_HT:

			TRACE(printk("<HT>"));

			POS  -= X;
			WPOS -= X;
			while (X < SIZE_X - 1) {

				X++;
				if (TAB_STOP[X >> 5] & (1 << (X & 31))) {

					break;
				}
			}
			POS  += X;
			WPOS += X;
			continue;

		case ASCII_LF:
		case ASCII_VT:
		case ASCII_FF:

			TRACE(printk("<LF>"));
			scroll_modified(cons, old_pos, WPOS);
			scroll_lf(cons);
			old_pos = WPOS;
	
			if (!MODE(newline)) {

				continue;
			}

		case ASCII_CR:

			TRACE(printk("<CR>"));
			scroll_modified(cons, old_pos, WPOS);
			scroll_cr(cons);
			old_pos = WPOS;
			continue;

		case ASCII_SO:

			TRACE(printk("<S0>"));
			CHARSET = 1;
			TRANSLATE = translation[G[1]];
			SET_MODE(display_ctrl);
			continue;

	 	case ASCII_SI:

			TRACE(printk("<SI>"));
			CHARSET = 0;
			TRANSLATE = translation[G[0]];
			CLEAR_MODE(display_ctrl);
			continue;

		case ASCII_CAN:
		case ASCII_SUB:

			TRACE(printk("<CAN>"));
			STATE = ESnormal; /* !!! display checkerboard!	*/
			continue;

		case ASCII_ESC:

			TRACE(printk("<ESC"));
			STATE = ESesc;
			continue;

		case 127:

			TRACE(printk("<DEL>"));
			/* !!! scroll_del(cons); */
			continue;

		case 128+27:

			TRACE(printk("<CSI"));	
			STATE = ESsquare;
			continue;
		}

		switch(STATE) {

		case ESesc:

			TRACE(printk("%c", (char) c));
			STATE = ESnormal;

			/*	handle sequences that do not change cursor
			**	position first.
			*/
			switch(c) {

			case '[':	STATE = ESsquare;
					continue;

			case '>':	/* DECKPNM - numeric keypad */
					TRACE(printk("= DECKPNM>\n"));
					CLEAR_MODE(applic_key);
					continue;

			case '=':	/* DECKPAM - appl. keypad */
					TRACE(printk("= DECKPAM>\n"));
					SET_MODE(applic_key);
				 	continue;

			case '(':	/* SCS G0 - select character set 0 */
					TRACE(printk("= SCS0>\n"));
					STATE = ESset;
					PAR[0] = 0;	/* charset to set */
					continue;

			case ')':	/* SCS G1 - select character set 1 */
					TRACE(printk("= SCS1>\n"));
					STATE = ESset;
					PAR[0] = 1;
					continue;

			case '*':	/* SCS G2 - select character set 2 */
					TRACE(printk("= SCS2>\n"));
					STATE = ESset;
					PAR[0] = 2;
					continue;

			case '+':	/* SCS G3 - select character set 3 */
					TRACE(printk("= SCS3>\n"));
					STATE = ESset;
					PAR[0] = 3;
					continue;

			case '#':	STATE = EShash;
					continue;

			case 'H':	/* HTS - set horiz. tab stop */
					TRACE(printk("= HTS>\n"));
					TAB_STOP[X >> 5] |= (1 << (X & 31));
					continue;

			case 'N':	TRACE(printk("= S2>\n"));
					CHARSET = 2;
					TRANSLATE = translation[G[2]];
					CLEAR_MODE(display_ctrl);
					continue;

			case 'O':	TRACE(printk("= S3>\n"));
					CHARSET = 3;
					TRANSLATE = translation[G[3]];
					CLEAR_MODE(display_ctrl);
					continue;

			case 'Z':	/* DECID - report ID */
					TRACE(printk("= DECID>\n"));
					xterm_report(cons, id);
					continue;

			case 'n':  /*	!!! full locking shift select G2 set */
			case 'o':  /* 	!!! full locking shift select G3 set */
					continue;

			case '7':	/* DECSC - save cursor */
					TRACE(printk("= DECSC>\n"));
					xterm_save_cur(cons);
					continue;

			#ifndef STRICT_XTERM
			case '%':	STATE = ESpercent;
					continue;
			#endif
			}

			/*	Now care of sequences that (potentially) alter
			**	the cursor position. Note that *all* cases have
			**	to be terminated with 'break'.
			*/
			scroll_modified(cons, old_pos, WPOS);

			switch (c) {

			case 'E':	/* NEL - next line */
					TRACE(printk("= NEL>\n"));
					scroll_cr(cons);
					scroll_lf(cons);
					break;

			case 'M':	/* RI - reverse index */
					TRACE(printk("= RI>\n"));
					scroll_reverse_lf(cons);
					break;

			case 'D':	/* IND - index (down one line) */
					TRACE(printk("= IND>\n"));
					scroll_lf(cons);
					break;

			case '8':	/* DECRC - restore cursor */
					TRACE(printk("= DECRC>\n"));
					xterm_restore_cur(cons);
					break;

			case 'c':	/* RIS - reset to initial state */
					TRACE(printk("= RIS>\n"));
					xterm_do_reset(cons, 1);
					break;

			default:
					TRACE(printk("= unknown ESC>\n"));
			}
			old_pos = WPOS;
			continue;

		case ESsquare:

			for(XNPAR = 0; XNPAR < MAX_NPAR; XNPAR++) {

				PAR[XNPAR] = 0;
			}

			XNPAR = 0;
			STATE = ESgetpars;

			if (c == '[') {		/* Function key */

				TRACE(printk("%c", (char) c));
				STATE=ESfunckey;
				continue;
			}

			if (c == '?') {

				TRACE(printk("%c", (char) c));
				SET_FLAG(question);
				continue;
			}
			/* fall through if ((c != '[') && (c != '?')) */

		case ESgetpars:

			if ((c == ';') && (XNPAR < MAX_NPAR-1)) {

				TRACE(printk("%c", (char) c));
				XNPAR++;
				continue;

			} else {

				if ((c >= '0') && (c <= '9')) {

					TRACE(printk("%c", (char) c));
					PAR[XNPAR] *= 10;
					PAR[XNPAR] += c-'0';
					continue;

				} else {

					STATE = ESgotpars;
				}
			}
			/* fall through when parameters are complete! */

		case ESgotpars:

			STATE = ESnormal;

			TRACE(printk("%c", (char) c));

			if (FLAG(question)) {

				/*	all these sequences may alter cursor
				**	position. Terminate cases with 'break'!
				*/
				scroll_modified(cons, old_pos, WPOS);
				switch(c) {

				case 'h':	/* SM - set mode */

					TRACE(printk("= SM>\n"));
					xterm_set_mode(cons, 1);
					break;

				case 'l':	/* RM - reset mode */

					TRACE(printk("= RM>\n"));
					xterm_set_mode(cons, 0);
					break;

				case 'r':	/* DECRM - restore mode */

					TRACE(printk("= DECRM>\n"));	
					xterm_backup_mode(cons, 0);
					break;

				case 's':	/* DECSM - save mode */

					TRACE(printk("= DECSM>\n"));
					xterm_backup_mode(cons, 1);
					break;

				default:

					TRACE(printk("= unknown ESC[? >\n"));
				}
				old_pos = WPOS;

				CLEAR_FLAG(question);
				continue;
			}

			/*	first care of sequences that do not alter
			**	cursor postion.
			*/
			switch(c) {

			case 'J':	/* ED - erase display	*/

				TRACE(printk("= ED>\n"));
				scroll_erase_display(cons, PAR[0]);
				continue;

			case 'K':	/* EL - erase in line */

				TRACE(printk("= EL>\n"));
				scroll_erase_line(cons, PAR[0]);
				continue;

			case 'P':

				TRACE(printk("= ???>\n"));
				if (PAR[0] <= 0) {

					PAR[0] = 1;
				}
				scroll_delete_chars(cons, PAR[0]);
				continue;

			case 'c':	/* ??? DA - device attributes */

				TRACE(printk("= DA>\n"));
				if (!PAR[0]) {

					xterm_report(cons, id);
				}
				continue;

			case 'g':	/* TBC - tabulation clear */

				TRACE(printk("= TBC>\n"));
				if (!PAR[0]) {

					TAB_STOP[X >> 5] &= ~(1 << (X & 31));

				} else {

					if (PAR[0] == 3) {

						TAB_STOP[0] = TAB_STOP[1] =
						TAB_STOP[2] = TAB_STOP[3] =
						TAB_STOP[4] = 0;
					}
				}
				continue;

			case 'm':	/* SGR - select graphic redition */

				TRACE(printk("= SGR>\n"));
				csi_m(cons);
				continue;

			case 'n':	/* DSR - device status report */

				TRACE(printk("= DSR>\n"));

				if (PAR[0] == 5) {

					xterm_report(cons, status);

				} else {

					if (PAR[0] == 6) {

					     xterm_report(cons, cursor);
					}
				}
				continue;

			case '@':	/* ICH - insert (blank) character(s) */

				TRACE(printk("= ICH>\n"));
				if (! PAR[0]) {

					PAR[0]++;
				}
				scroll_insert_chars(cons, PAR[0]);
				continue;
		#if 0
			case 'q':	/* ??? DECLL - load LED */
					/* ??? check this! */

				if (!PAR[0]) {

					kbd_set_led_state(&KBD_STATE, 0);

				} else {

					if (PAR[0] < 5) {

						kbd_set_led(&KBD_STATE, PAR[0]-1);
						kbd_setleds(KBD);
					}
				}
				continue;
		#endif
		#ifndef STRICT_XTERM

			case 'X':

				scroll_erase_chars(cons, PAR[0]);
				continue;

			case ']':	/* setterm functions */

				xterm_set_term(cons);
				continue;
		#endif
			}

			/*	now handle those sequences that may alter
			**	cursor position. NOTE: all cases have to
			**	be terminated with 'break'.
			*/
			scroll_modified(cons, old_pos, WPOS);

			switch(c) {

			case 'A':	/* CUU - cursor up	*/

				TRACE(printk("= CUU>\n"));
				if (!PAR[0]) {

					PAR[0]++;
				}
				scroll_gotoxy(cons, X, Y - PAR[0]);
				break;

			case 'B':	/* CUD - cursor down	*/
			case 'e':

				TRACE(printk("= CUD>\n"));
				if (!PAR[0]) {

					PAR[0]++;
				}
				scroll_gotoxy(cons, X, Y + PAR[0]);
				break;

			case 'C':	/* CUF - cursor forward	*/
			case 'a':

				TRACE(printk("= CUF>\n"));
				if (!PAR[0]) {

					PAR[0]++;
				}
				scroll_gotoxy(cons, X + PAR[0], Y);
				break;

			case 'D':	/* CUB - cursor backward*/

				TRACE(printk("= CUB>\n"));
				if (!PAR[0]) {

					PAR[0]++;
				}
				scroll_gotoxy(cons, X - PAR[0], Y);
				break;

			case 'H':	/* CUP - set cursor position */
			case 'f':	/* HVP - horiz. and vert. position */

				TRACE(printk(c=='H' ? "= CUP>\n" : "= HVP>\n"));
				if (PAR[0]) {

					PAR[0]--;
				}
				if (PAR[1]) {

					PAR[1]--;
				}
				if (MODE(origin)) {

					PAR[0] += TOP;
				}
				scroll_gotoxy(cons, PAR[1], PAR[0]);
				break;

			case 'L':

				TRACE(printk("= ???>\n"));
				if (PAR[0] <= 0) {

					PAR[0] = 1;
				}
				scroll_down(cons, Y, BOTTOM, PAR[0]);
				break;

			case 'M':

				TRACE(printk("= ???>\n"));
				if (PAR[0] <= 0) {

					PAR[0] = 1;
				}
				scroll_up(cons, Y, BOTTOM, PAR[0]);
				break;

			case 'h':	/* SM - set mode */

				TRACE(printk("= SM>\n"));
				xterm_set_mode(cons, 1);
				break;

			case 'l':	/* RM - reset mode */

				TRACE(printk("= RM>\n"));
				xterm_set_mode(cons, 0);
				break;

			case 'r':	/* DECSTBM - set top & bottom margin */

				TRACE(printk("= DECSTBM>\n"));

				if (!PAR[0]) {

					PAR[0]++;
				}

				if (!PAR[1]) {

					PAR[1] = SIZE_Y;
				}

				/* Minimum allowed region is 2 lines */
				if ((PAR[0]<PAR[1]) && (PAR[1]<=SIZE_Y)) {

					TOP    = PAR[0]-1;
					BOTTOM = PAR[1];
					scroll_gotoxy(cons, 0, 0);
				}
				break;

			case 'x':	/* !!! DECREQTPARM - report parameter */

				TRACE(printk("= DECREQTPARM>\n"));
				break;

			#ifndef STRICT_XTERM
			case 's':

				TRACE(printk("= save cur>\n"));
				xterm_save_cur(cons);
				break;

			case 'u':
	
				TRACE(printk("= rest cur>\n"));
				xterm_restore_cur(cons);
				break;

			case 'E':

				if (!PAR[0]) {

					PAR[0]++;
				}
				scroll_gotoxy(cons, 0, Y + PAR[0]);
				break;

			case 'F':

				if (!PAR[0]) {

					PAR[0]++;
				}
				scroll_gotoxy(cons, 0, Y - PAR[0]);
				break;

			case 'G':
			case '`':

				if (PAR[0]) {

					PAR[0]--;
				}
				scroll_gotoxy(cons, PAR[0], Y);
				break;

			case 'd':

				if (PAR[0]) {

					PAR[0]--;
				}
				if (MODE(origin)) {

					PAR[0] += TOP;
				}
				scroll_gotoxy(cons, X, PAR[0]);
				break;
			#endif

			default:

				TRACE(printk("= unknown ESC[>\n"));
			}
			old_pos = WPOS;
			continue;

		case ESpercent:

			STATE = ESnormal;
			TRACE(printk("%c", (char) c));

			switch (c) {

			case '@':	/* defined in ISO 2022 */

				TRACE(printk("= UTF-8 off>\n"));
				CLEAR_MODE(utf8);
				continue;

			case 'G':	/* prelim official escape code */
			case '8':	/* retained for compatibility */

				TRACE(printk("= UTF-8 on>\n"));
				SET_MODE(utf8);
				continue;
			}
			TRACE(printk("= unkown ESC-percent>\n"));
			continue;

		case ESfunckey:

			STATE = ESnormal;
			TRACE(printk("fn-key %c\n", (char) c));
			continue;

		case EShash:

			if (c == '8') {	/* DEC screen alignment test */

				TRACE(printk("= DECSAT>\n"));
				ERASE = (ERASE & 0xff00) | 'E';
				scroll_erase_display(cons, 2);
				ERASE = (ERASE & 0xff00) | ' ';

			} else {

				TRACE(printk("= unkown ESC-#>\n"));
			}
			STATE = ESnormal;
			continue;

		case ESset:

			TRACE(printk("%c>\n", (char) c));
			xterm_set_charset(cons, PAR[0], c);
			STATE = ESnormal;
			continue;

		default:

			TRACE(printk("<%.2x, state %i>\n", c, STATE));
			STATE = ESnormal;
			continue;
		}
	}

	enable_bh(KEYBOARD_BH);

	if (FLAG(need_wrap) || (old_pos != WPOS)) {

		scroll_modified(cons, old_pos, FLAG(need_wrap) ? WPOS+1 : WPOS);
	}

	return n;
}


/* -----------------------------------------------------------------------------
**	font translation tables
** -----------------------------------------------------------------------------
*/

static unicode translation[][256] = {
  /* 8-bit Latin-1 mapped to Unicode -- trivial mapping */
  #define LAT1_MAP	0
  {
    0x0000, 0x0001, 0x0002, 0x0003, 0x0004, 0x0005, 0x0006, 0x0007,
    0x0008, 0x0009, 0x000a, 0x000b, 0x000c, 0x000d, 0x000e, 0x000f,
    0x0010, 0x0011, 0x0012, 0x0013, 0x0014, 0x0015, 0x0016, 0x0017,
    0x0018, 0x0019, 0x001a, 0x001b, 0x001c, 0x001d, 0x001e, 0x001f,
    0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027,
    0x0028, 0x0029, 0x002a, 0x002b, 0x002c, 0x002d, 0x002e, 0x002f,
    0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037,
    0x0038, 0x0039, 0x003a, 0x003b, 0x003c, 0x003d, 0x003e, 0x003f,
    0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047,
    0x0048, 0x0049, 0x004a, 0x004b, 0x004c, 0x004d, 0x004e, 0x004f,
    0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057,
    0x0058, 0x0059, 0x005a, 0x005b, 0x005c, 0x005d, 0x005e, 0x005f,
    0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067,
    0x0068, 0x0069, 0x006a, 0x006b, 0x006c, 0x006d, 0x006e, 0x006f,
    0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077,
    0x0078, 0x0079, 0x007a, 0x007b, 0x007c, 0x007d, 0x007e, 0x007f,
    0x0080, 0x0081, 0x0082, 0x0083, 0x0084, 0x0085, 0x0086, 0x0087,
    0x0088, 0x0089, 0x008a, 0x008b, 0x008c, 0x008d, 0x008e, 0x008f,
    0x0090, 0x0091, 0x0092, 0x0093, 0x0094, 0x0095, 0x0096, 0x0097,
    0x0098, 0x0099, 0x009a, 0x009b, 0x009c, 0x009d, 0x009e, 0x009f,
    0x00a0, 0x00a1, 0x00a2, 0x00a3, 0x00a4, 0x00a5, 0x00a6, 0x00a7,
    0x00a8, 0x00a9, 0x00aa, 0x00ab, 0x00ac, 0x00ad, 0x00ae, 0x00af,
    0x00b0, 0x00b1, 0x00b2, 0x00b3, 0x00b4, 0x00b5, 0x00b6, 0x00b7,
    0x00b8, 0x00b9, 0x00ba, 0x00bb, 0x00bc, 0x00bd, 0x00be, 0x00bf,
    0x00c0, 0x00c1, 0x00c2, 0x00c3, 0x00c4, 0x00c5, 0x00c6, 0x00c7,
    0x00c8, 0x00c9, 0x00ca, 0x00cb, 0x00cc, 0x00cd, 0x00ce, 0x00cf,
    0x00d0, 0x00d1, 0x00d2, 0x00d3, 0x00d4, 0x00d5, 0x00d6, 0x00d7,
    0x00d8, 0x00d9, 0x00da, 0x00db, 0x00dc, 0x00dd, 0x00de, 0x00df,
    0x00e0, 0x00e1, 0x00e2, 0x00e3, 0x00e4, 0x00e5, 0x00e6, 0x00e7,
    0x00e8, 0x00e9, 0x00ea, 0x00eb, 0x00ec, 0x00ed, 0x00ee, 0x00ef,
    0x00f0, 0x00f1, 0x00f2, 0x00f3, 0x00f4, 0x00f5, 0x00f6, 0x00f7,
    0x00f8, 0x00f9, 0x00fa, 0x00fb, 0x00fc, 0x00fd, 0x00fe, 0x00ff
  },
  /* VT100 graphics mapped to Unicode */
  #define GRAF_MAP	1
  {
    0x0000, 0x0001, 0x0002, 0x0003, 0x0004, 0x0005, 0x0006, 0x0007,
    0x0008, 0x0009, 0x000a, 0x000b, 0x000c, 0x000d, 0x000e, 0x000f,
    0x0010, 0x0011, 0x0012, 0x0013, 0x0014, 0x0015, 0x0016, 0x0017,
    0x0018, 0x0019, 0x001a, 0x001b, 0x001c, 0x001d, 0x001e, 0x001f,
    0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027,
    0x0028, 0x0029, 0x002a, 0x2192, 0x2190, 0x2191, 0x2193, 0x002f,
    0x2588, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037,
    0x0038, 0x0039, 0x003a, 0x003b, 0x003c, 0x003d, 0x003e, 0x003f,
    0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047,
    0x0048, 0x0049, 0x004a, 0x004b, 0x004c, 0x004d, 0x004e, 0x004f,
    0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057,
    0x0058, 0x0059, 0x005a, 0x005b, 0x005c, 0x005d, 0x005e, 0x00a0,
    0x25c6, 0x2592, 0x2409, 0x240c, 0x240d, 0x240a, 0x00b0, 0x00b1,
    0x2591, 0x240b, 0x2518, 0x2510, 0x250c, 0x2514, 0x253c, 0xf800,
    0xf801, 0x2500, 0xf803, 0xf804, 0x251c, 0x2524, 0x2534, 0x252c,
    0x2502, 0x2264, 0x2265, 0x03c0, 0x2260, 0x00a3, 0x00b7, 0x007f,
    0x0080, 0x0081, 0x0082, 0x0083, 0x0084, 0x0085, 0x0086, 0x0087,
    0x0088, 0x0089, 0x008a, 0x008b, 0x008c, 0x008d, 0x008e, 0x008f,
    0x0090, 0x0091, 0x0092, 0x0093, 0x0094, 0x0095, 0x0096, 0x0097,
    0x0098, 0x0099, 0x009a, 0x009b, 0x009c, 0x009d, 0x009e, 0x009f,
    0x00a0, 0x00a1, 0x00a2, 0x00a3, 0x00a4, 0x00a5, 0x00a6, 0x00a7,
    0x00a8, 0x00a9, 0x00aa, 0x00ab, 0x00ac, 0x00ad, 0x00ae, 0x00af,
    0x00b0, 0x00b1, 0x00b2, 0x00b3, 0x00b4, 0x00b5, 0x00b6, 0x00b7,
    0x00b8, 0x00b9, 0x00ba, 0x00bb, 0x00bc, 0x00bd, 0x00be, 0x00bf,
    0x00c0, 0x00c1, 0x00c2, 0x00c3, 0x00c4, 0x00c5, 0x00c6, 0x00c7,
    0x00c8, 0x00c9, 0x00ca, 0x00cb, 0x00cc, 0x00cd, 0x00ce, 0x00cf,
    0x00d0, 0x00d1, 0x00d2, 0x00d3, 0x00d4, 0x00d5, 0x00d6, 0x00d7,
    0x00d8, 0x00d9, 0x00da, 0x00db, 0x00dc, 0x00dd, 0x00de, 0x00df,
    0x00e0, 0x00e1, 0x00e2, 0x00e3, 0x00e4, 0x00e5, 0x00e6, 0x00e7,
    0x00e8, 0x00e9, 0x00ea, 0x00eb, 0x00ec, 0x00ed, 0x00ee, 0x00ef,
    0x00f0, 0x00f1, 0x00f2, 0x00f3, 0x00f4, 0x00f5, 0x00f6, 0x00f7,
    0x00f8, 0x00f9, 0x00fa, 0x00fb, 0x00fc, 0x00fd, 0x00fe, 0x00ff
  },
  /* IBM Codepage 437 mapped to Unicode */
  #define IBMPC_MAP	2
  {
    0x0000, 0x263a, 0x263b, 0x2665, 0x2666, 0x2663, 0x2660, 0x2022, 
    0x25d8, 0x25cb, 0x25d9, 0x2642, 0x2640, 0x266a, 0x266b, 0x263c,
    0x25b6, 0x25c0, 0x2195, 0x203c, 0x00b6, 0x00a7, 0x25ac, 0x21a8,
    0x2191, 0x2193, 0x2192, 0x2190, 0x221f, 0x2194, 0x25b2, 0x25bc,
    0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027,
    0x0028, 0x0029, 0x002a, 0x002b, 0x002c, 0x002d, 0x002e, 0x002f,
    0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037,
    0x0038, 0x0039, 0x003a, 0x003b, 0x003c, 0x003d, 0x003e, 0x003f,
    0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047,
    0x0048, 0x0049, 0x004a, 0x004b, 0x004c, 0x004d, 0x004e, 0x004f,
    0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057,
    0x0058, 0x0059, 0x005a, 0x005b, 0x005c, 0x005d, 0x005e, 0x005f,
    0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067,
    0x0068, 0x0069, 0x006a, 0x006b, 0x006c, 0x006d, 0x006e, 0x006f,
    0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077,
    0x0078, 0x0079, 0x007a, 0x007b, 0x007c, 0x007d, 0x007e, 0x2302,
    0x00c7, 0x00fc, 0x00e9, 0x00e2, 0x00e4, 0x00e0, 0x00e5, 0x00e7,
    0x00ea, 0x00eb, 0x00e8, 0x00ef, 0x00ee, 0x00ec, 0x00c4, 0x00c5,
    0x00c9, 0x00e6, 0x00c6, 0x00f4, 0x00f6, 0x00f2, 0x00fb, 0x00f9,
    0x00ff, 0x00d6, 0x00dc, 0x00a2, 0x00a3, 0x00a5, 0x20a7, 0x0192,
    0x00e1, 0x00ed, 0x00f3, 0x00fa, 0x00f1, 0x00d1, 0x00aa, 0x00ba,
    0x00bf, 0x2310, 0x00ac, 0x00bd, 0x00bc, 0x00a1, 0x00ab, 0x00bb,
    0x2591, 0x2592, 0x2593, 0x2502, 0x2524, 0x2561, 0x2562, 0x2556,
    0x2555, 0x2563, 0x2551, 0x2557, 0x255d, 0x255c, 0x255b, 0x2510,
    0x2514, 0x2534, 0x252c, 0x251c, 0x2500, 0x253c, 0x255e, 0x255f,
    0x255a, 0x2554, 0x2569, 0x2566, 0x2560, 0x2550, 0x256c, 0x2567,
    0x2568, 0x2564, 0x2565, 0x2559, 0x2558, 0x2552, 0x2553, 0x256b,
    0x256a, 0x2518, 0x250c, 0x2588, 0x2584, 0x258c, 0x2590, 0x2580,
    0x03b1, 0x00df, 0x0393, 0x03c0, 0x03a3, 0x03c3, 0x00b5, 0x03c4,
    0x03a6, 0x0398, 0x03a9, 0x03b4, 0x221e, 0x03c6, 0x03b5, 0x2229,
    0x2261, 0x00b1, 0x2265, 0x2264, 0x2320, 0x2321, 0x00f7, 0x2248,
    0x00b0, 0x2219, 0x00b7, 0x221a, 0x207f, 0x00b2, 0x25a0, 0x00a0
  }, 
  /* User mapping -- default to codes for direct font mapping */
  #define USER_MAP	3
  {
    0xf000, 0xf001, 0xf002, 0xf003, 0xf004, 0xf005, 0xf006, 0xf007,
    0xf008, 0xf009, 0xf00a, 0xf00b, 0xf00c, 0xf00d, 0xf00e, 0xf00f,
    0xf010, 0xf011, 0xf012, 0xf013, 0xf014, 0xf015, 0xf016, 0xf017,
    0xf018, 0xf019, 0xf01a, 0xf01b, 0xf01c, 0xf01d, 0xf01e, 0xf01f,
    0xf020, 0xf021, 0xf022, 0xf023, 0xf024, 0xf025, 0xf026, 0xf027,
    0xf028, 0xf029, 0xf02a, 0xf02b, 0xf02c, 0xf02d, 0xf02e, 0xf02f,
    0xf030, 0xf031, 0xf032, 0xf033, 0xf034, 0xf035, 0xf036, 0xf037,
    0xf038, 0xf039, 0xf03a, 0xf03b, 0xf03c, 0xf03d, 0xf03e, 0xf03f,
    0xf040, 0xf041, 0xf042, 0xf043, 0xf044, 0xf045, 0xf046, 0xf047,
    0xf048, 0xf049, 0xf04a, 0xf04b, 0xf04c, 0xf04d, 0xf04e, 0xf04f,
    0xf050, 0xf051, 0xf052, 0xf053, 0xf054, 0xf055, 0xf056, 0xf057,
    0xf058, 0xf059, 0xf05a, 0xf05b, 0xf05c, 0xf05d, 0xf05e, 0xf05f,
    0xf060, 0xf061, 0xf062, 0xf063, 0xf064, 0xf065, 0xf066, 0xf067,
    0xf068, 0xf069, 0xf06a, 0xf06b, 0xf06c, 0xf06d, 0xf06e, 0xf06f,
    0xf070, 0xf071, 0xf072, 0xf073, 0xf074, 0xf075, 0xf076, 0xf077,
    0xf078, 0xf079, 0xf07a, 0xf07b, 0xf07c, 0xf07d, 0xf07e, 0xf07f,
    0xf080, 0xf081, 0xf082, 0xf083, 0xf084, 0xf085, 0xf086, 0xf087,
    0xf088, 0xf089, 0xf08a, 0xf08b, 0xf08c, 0xf08d, 0xf08e, 0xf08f,
    0xf090, 0xf091, 0xf092, 0xf093, 0xf094, 0xf095, 0xf096, 0xf097,
    0xf098, 0xf099, 0xf09a, 0xf09b, 0xf09c, 0xf09d, 0xf09e, 0xf09f,
    0xf0a0, 0xf0a1, 0xf0a2, 0xf0a3, 0xf0a4, 0xf0a5, 0xf0a6, 0xf0a7,
    0xf0a8, 0xf0a9, 0xf0aa, 0xf0ab, 0xf0ac, 0xf0ad, 0xf0ae, 0xf0af,
    0xf0b0, 0xf0b1, 0xf0b2, 0xf0b3, 0xf0b4, 0xf0b5, 0xf0b6, 0xf0b7,
    0xf0b8, 0xf0b9, 0xf0ba, 0xf0bb, 0xf0bc, 0xf0bd, 0xf0be, 0xf0bf,
    0xf0c0, 0xf0c1, 0xf0c2, 0xf0c3, 0xf0c4, 0xf0c5, 0xf0c6, 0xf0c7,
    0xf0c8, 0xf0c9, 0xf0ca, 0xf0cb, 0xf0cc, 0xf0cd, 0xf0ce, 0xf0cf,
    0xf0d0, 0xf0d1, 0xf0d2, 0xf0d3, 0xf0d4, 0xf0d5, 0xf0d6, 0xf0d7,
    0xf0d8, 0xf0d9, 0xf0da, 0xf0db, 0xf0dc, 0xf0dd, 0xf0de, 0xf0df,
    0xf0e0, 0xf0e1, 0xf0e2, 0xf0e3, 0xf0e4, 0xf0e5, 0xf0e6, 0xf0e7,
    0xf0e8, 0xf0e9, 0xf0ea, 0xf0eb, 0xf0ec, 0xf0ed, 0xf0ee, 0xf0ef,
    0xf0f0, 0xf0f1, 0xf0f2, 0xf0f3, 0xf0f4, 0xf0f5, 0xf0f6, 0xf0f7,
    0xf0f8, 0xf0f9, 0xf0fa, 0xf0fb, 0xf0fc, 0xf0fd, 0xf0fe, 0xf0ff
  }
  #define LAST_MAP	4
};

/*	The standard kernel character-to-font mappings are not invertible
**	-- this is just a best effort.
*/
